//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"

using System;
using System.Collections.Generic;
using System.Globalization;
using SM = System.Math;

namespace Microsoft.Xna.Framework
{
    public enum ContainmentType
    {
        Disjoint    = 0,
        Contains    = 1,
        Intersects  = 2
    }


    public struct BoundingBox : IEquatable<BoundingBox>
    {
        public const int CornerCount = 8;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public Vector3 Min;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public Vector3 Max;

        /// <summary>
        /// Returns 8 corner position of bounding box.
        ///     Near    Far
        ///    0----1  4----5
        ///    |    |  |    |
        ///    |    |  |    |
        ///    3----2  7----6
        /// </summary>
        /// <returns></returns>
        public Vector3[] GetCorners()
        {
            return new Vector3[]{
                new Vector3( Min.X, Max.Y, Max.Z ),
                new Vector3( Max.X, Max.Y, Max.Z ),
                new Vector3( Max.X, Min.Y, Max.Z ),
                new Vector3( Min.X, Min.Y, Max.Z ),
                new Vector3( Min.X, Max.Y, Min.Z ),
                new Vector3( Max.X, Max.Y, Min.Z ),
                new Vector3( Max.X, Min.Y, Min.Z ),
                new Vector3( Min.X, Min.Y, Min.Z ),
            };
        }

        /// <summary>
        /// It copies bounding box corner points to given array.
        /// </summary>
        /// <param name="corners">vector3 array that has equal or bigger than 8 elements.</param>
        public void GetCorners(Vector3[] corners )
        {
            if (corners == null)
                throw new ArgumentNullException("corners");
            if (corners.Length < 8)
                throw new ArgumentOutOfRangeException("corners", "FrameworkResources.NotEnoughCorners");

            corners[0].X = Min.X; corners[0].Y = Max.Y; corners[0].Z = Max.Z;
            corners[1].X = Max.X; corners[1].Y = Max.Y; corners[1].Z = Max.Z;
            corners[2].X = Max.X; corners[2].Y = Min.Y; corners[2].Z = Max.Z;
            corners[3].X = Min.X; corners[3].Y = Min.Y; corners[3].Z = Max.Z;
            corners[4].X = Min.X; corners[4].Y = Max.Y; corners[4].Z = Min.Z;
            corners[5].X = Max.X; corners[5].Y = Max.Y; corners[5].Z = Min.Z;
            corners[6].X = Max.X; corners[6].Y = Min.Y; corners[6].Z = Min.Z;
            corners[7].X = Min.X; corners[7].Y = Min.Y; corners[7].Z = Min.Z;
        }

        /// <summary>
        /// Construct a bounding box from min and max position.
        /// </summary>
        /// <param name="min"></param>
        /// <param name="max"></param>
        public BoundingBox(Vector3 min, Vector3 max)
        {
            Min = min; Max = max;
        }

        #region Override methods

        // Implement IEquatable<T>
        public bool Equals(BoundingBox other)
        {
            return (Min == other.Min && Max == other.Max);
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is BoundingBox)
            {
                equal = Equals((BoundingBox)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return Min.GetHashCode() + Max.GetHashCode();
        }

        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{Min:{0} Max:{1}}}", Min.ToString(), Max.ToString());
        }
        #endregion

        /// <summary>
        /// Create a merged bounding box from two bounding boxes.
        /// </summary>
        /// <param name="original"></param>
        /// <param name="additional"></param>
        /// <returns></returns>
        public static BoundingBox CreateMerged(BoundingBox original, BoundingBox additional)
        {
            BoundingBox result;

            Vector3.Min(ref original.Min, ref additional.Min, out result.Min);
            Vector3.Max(ref original.Max, ref additional.Max, out result.Max);

            return result;
        }

        public static void CreateMerged(ref BoundingBox original, ref BoundingBox additional, out BoundingBox result)
        {
            Vector3 min;
            Vector3 max;

            Vector3.Min(ref original.Min, ref additional.Min, out min);
            Vector3.Max(ref original.Max, ref additional.Max, out max);

            result.Min = min;
            result.Max = max;
        }

        /// <summary>
        /// Create Bounding box from sphere.
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns></returns>
        public static BoundingBox CreateFromSphere(BoundingSphere sphere)
        {
            BoundingBox result;

            result.Min.X = sphere.Center.X - sphere.Radius;
            result.Min.Y = sphere.Center.Y - sphere.Radius;
            result.Min.Z = sphere.Center.Z - sphere.Radius;

            result.Max.X = sphere.Center.X + sphere.Radius;
            result.Max.Y = sphere.Center.Y + sphere.Radius;
            result.Max.Z = sphere.Center.Z + sphere.Radius;

            return result;
        }

        public static void CreateFromSphere(ref BoundingSphere sphere, out BoundingBox result)
        {
            result.Min.X = sphere.Center.X - sphere.Radius;
            result.Min.Y = sphere.Center.Y - sphere.Radius;
            result.Min.Z = sphere.Center.Z - sphere.Radius;

            result.Max.X = sphere.Center.X + sphere.Radius;
            result.Max.Y = sphere.Center.Y + sphere.Radius;
            result.Max.Z = sphere.Center.Z + sphere.Radius;
        }


        /// <summary>
        /// Create a bounding box from points.
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        public static BoundingBox CreateFromPoints(IEnumerable<Vector3> points)
        {
            if (points == null)
            {
                throw new ArgumentNullException();
            }

            bool atLeastOne = false;
            Vector3 min = new Vector3( Single.MaxValue );
            Vector3 max = new Vector3( Single.MinValue );
            foreach( Vector3 p in points )
            {
                Vector3 pp = p;

                Vector3.Min( ref min, ref pp, out min );
                Vector3.Max(ref max, ref pp, out max);

                atLeastOne = true;
            }

            if (atLeastOne == false)
            {
                throw new ArgumentException("FrameworkResources.BoundingBoxZeroPoints");
            }

            // TODO: Should we do empty test?
            return new BoundingBox( min, max );
        }

        #region Intersections
        /// <summary>
        /// Intersection test aginst another bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns true if given bounding box is intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingBox box)
        {
            if (Max.X < box.Min.X || Min.X > box.Max.X) return false;
            if (Max.Y < box.Min.Y || Min.Y > box.Max.Y) return false;
            if (Max.Z < box.Min.Z || Min.Z > box.Max.Z) return false;
            return true;
        }

        public void Intersects(ref BoundingBox box, out bool result)
        {
            result = false;
            if (Max.X < box.Min.X || Min.X > box.Max.X) return;
            if (Max.Y < box.Min.Y || Min.Y > box.Max.Y) return;
            if (Max.Z < box.Min.Z || Min.Z > box.Max.Z) return;
            result = true;
        }


        /// <summary>
        /// Intersection test aginst another bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns true if given frustum is intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingFrustum frustum)
        {
            if (null == frustum)
            {
                throw new ArgumentNullException("frustum", "FrameworkResources.NullNotAllowed");
            }

            return frustum.Intersects(this);
        }

        /// <summary>
        /// Intersection test aginst another plane.
        /// </summary>
        /// <param name="plane"></param>
        /// <returns>Returns PlaneIntersectionType that contins Inside, Outisde, or Intersecting.</returns>
        public PlaneIntersectionType Intersects(Plane plane)
        {
            // This algorithm came from Real-Time Rendering Second Edition p.587
            Vector3 vMin;
            vMin.X = (plane.Normal.X >= 0.0f) ? Min.X : Max.X;
            vMin.Y = (plane.Normal.Y >= 0.0f) ? Min.Y : Max.Y;
            vMin.Z = (plane.Normal.Z >= 0.0f) ? Min.Z : Max.Z;

            Vector3 vMax;
            vMax.X = (plane.Normal.X >= 0.0f) ? Max.X : Min.X;
            vMax.Y = (plane.Normal.Y >= 0.0f) ? Max.Y : Min.Y;
            vMax.Z = (plane.Normal.Z >= 0.0f) ? Max.Z : Min.Z;

            float dot = plane.Normal.X * vMin.X + plane.Normal.Y * vMin.Y + plane.Normal.Z * vMin.Z;
            if (dot + plane.D > 0.0f) return PlaneIntersectionType.Front;

            dot = plane.Normal.X * vMax.X + plane.Normal.Y * vMax.Y + plane.Normal.Z * vMax.Z;
            if ( dot + plane.D < 0.0f) return PlaneIntersectionType.Back;
            return PlaneIntersectionType.Intersecting;
        }

        public void Intersects(ref Plane plane, out PlaneIntersectionType result)
        {
            // This algorithm came from Real-Time Rendering Second Edition p.587
            Vector3 vMin;
            vMin.X = (plane.Normal.X >= 0.0f) ? Min.X : Max.X;
            vMin.Y = (plane.Normal.Y >= 0.0f) ? Min.Y : Max.Y;
            vMin.Z = (plane.Normal.Z >= 0.0f) ? Min.Z : Max.Z;

            Vector3 vMax;
            vMax.X = (plane.Normal.X >= 0.0f) ? Max.X : Min.X;
            vMax.Y = (plane.Normal.Y >= 0.0f) ? Max.Y : Min.Y;
            vMax.Z = (plane.Normal.Z >= 0.0f) ? Max.Z : Min.Z;

            float dot = plane.Normal.X * vMin.X + plane.Normal.Y * vMin.Y + plane.Normal.Z * vMin.Z;
            if (dot + plane.D > 0.0f)
            {
                result = PlaneIntersectionType.Front;
                return;
            }

            dot = plane.Normal.X * vMax.X + plane.Normal.Y * vMax.Y + plane.Normal.Z * vMax.Z;
            if ( dot + plane.D < 0.0f)
            {
                result = PlaneIntersectionType.Back;
                return;
            }

            result = PlaneIntersectionType.Intersecting;
        }

        /// <summary>
        /// Intersection test aginst another Ray.
        /// </summary>
        /// <param name="ray"></param>
        /// <returns></returns>
        public Nullable<float> Intersects(Ray ray)
        {
            const float epsilon = 1e-6f;

            // This algorithm based on Slabs method on Real-Time Collosion Detection p.180.
            float tmin = 0.0f;  // set to -Single.MaxValue to get first hit on line
            float tmax = Single.MaxValue;

            // X-axis slab test.
            if (SM.Abs(ray.Direction.X) < epsilon)
            {
                // Ray is parallel to slab. No hit if origin not within slab.
                if (ray.Position.X < Min.X || ray.Position.X > Max.X)
                {
                    return null;
                }
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab.
                float ood = 1.0f / ray.Direction.X;
                float t1 = (Min.X - ray.Position.X) * ood;
                float t2 = (Max.X - ray.Position.X) * ood;
                // Make t1 be intersection with near plane, t2 with far plane.
                if (t1 > t2)
                {
                    float tmp = t1;
                    t1 = t2;
                    t2 = tmp;
                }
                // Compute the intersection of slab intersection intervals.
                tmin = MathHelper.Max(t1, tmin);
                tmax = MathHelper.Min(t2, tmax);
                // Exit with no collosion as soon as slab intersection becomes empty.
                if (tmin > tmax)
                {
                    return null;
                }
            }

            // Y-axis slab test.
            if (SM.Abs(ray.Direction.Y) < epsilon)
            {
                // Ray is parallel to slab. No hit if origin not within slab.
                if (ray.Position.Y < Min.Y || ray.Position.Y > Max.Y)
                {
                    return null;
                }
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab.
                float ood = 1.0f / ray.Direction.Y;
                float t1 = (Min.Y - ray.Position.Y) * ood;
                float t2 = (Max.Y - ray.Position.Y) * ood;
                // Make t1 be intersection with near plane, t2 with far plane.
                if (t1 > t2)
                {
                    float tmp = t1;
                    t1 = t2;
                    t2 = tmp;
                }
                // Compute the intersection of slab intersection intervals.
                tmin = MathHelper.Max(t1, tmin);
                tmax = MathHelper.Min(t2, tmax);
                // Exit with no collosion as soon as slab intersection becomes empty.
                if (tmin > tmax)
                {
                    return null;
                }
            }

            // Z-axis slab test.
            if (SM.Abs(ray.Direction.Z) < epsilon)
            {
                // Ray is parallel to slab. No hit if origin not within slab.
                if (ray.Position.Z < Min.Z || ray.Position.Z > Max.Z)
                {
                    return null;
                }
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab.
                float ood = 1.0f / ray.Direction.Z;
                float t1 = (Min.Z - ray.Position.Z) * ood;
                float t2 = (Max.Z - ray.Position.Z) * ood;
                // Make t1 be intersection with near plane, t2 with far plane.
                if (t1 > t2)
                {
                    float tmp = t1;
                    t1 = t2;
                    t2 = tmp;
                }
                // Compute the intersection of slab intersection intervals.
                tmin = MathHelper.Max(t1, tmin);
                tmax = MathHelper.Min(t2, tmax);
                // Exit with no collosion as soon as slab intersection becomes empty.
                if (tmin > tmax)
                {
                    return null;
                }
            }

            return tmin;
        }

        public void Intersects(ref Ray ray, out Nullable<float> result)
        {
            const float epsilon = 1e-6f;

            result = null;

            // This algorithm based on Slabs method on Real-Time Collosion Detection p.180.
            float tmin = 0.0f;  // set to -Single.MaxValue to get first hit on line
            float tmax = Single.MaxValue;

            // X-axis slab test.
            if (SM.Abs(ray.Direction.X) < epsilon)
            {
                // Ray is parallel to slab. No hit if origin not within slab.
                if (ray.Position.X < Min.X || ray.Position.X > Max.X)
                {
                    return;
                }
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab.
                float ood = 1.0f / ray.Direction.X;
                float t1 = (Min.X - ray.Position.X) * ood;
                float t2 = (Max.X - ray.Position.X) * ood;
                // Make t1 be intersection with near plane, t2 with far plane.
                if (t1 > t2)
                {
                    float tmp = t1;
                    t1 = t2;
                    t2 = tmp;
                }
                // Compute the intersection of slab intersection intervals.
                tmin = MathHelper.Max(t1, tmin);
                tmax = MathHelper.Min(t2, tmax);
                // Exit with no collosion as soon as slab intersection becomes empty.
                if (tmin > tmax)
                {
                    return;
                }
            }

            // Y-axis slab test.
            if (SM.Abs(ray.Direction.Y) < epsilon)
            {
                // Ray is parallel to slab. No hit if origin not within slab.
                if (ray.Position.Y < Min.Y || ray.Position.Y > Max.Y)
                {
                    return;
                }
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab.
                float ood = 1.0f / ray.Direction.Y;
                float t1 = (Min.Y - ray.Position.Y) * ood;
                float t2 = (Max.Y - ray.Position.Y) * ood;
                // Make t1 be intersection with near plane, t2 with far plane.
                if (t1 > t2)
                {
                    float tmp = t1;
                    t1 = t2;
                    t2 = tmp;
                }
                // Compute the intersection of slab intersection intervals.
                tmin = MathHelper.Max(t1, tmin);
                tmax = MathHelper.Min(t2, tmax);
                // Exit with no collosion as soon as slab intersection becomes empty.
                if (tmin > tmax)
                {
                    return;
                }
            }

            // Z-axis slab test.
            if (SM.Abs(ray.Direction.Z) < epsilon)
            {
                // Ray is parallel to slab. No hit if origin not within slab.
                if (ray.Position.Z < Min.Z || ray.Position.Z > Max.Z)
                {
                    return;
                }
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab.
                float ood = 1.0f / ray.Direction.Z;
                float t1 = (Min.Z - ray.Position.Z) * ood;
                float t2 = (Max.Z - ray.Position.Z) * ood;
                // Make t1 be intersection with near plane, t2 with far plane.
                if (t1 > t2)
                {
                    float tmp = t1;
                    t1 = t2;
                    t2 = tmp;
                }
                // Compute the intersection of slab intersection intervals.
                tmin = MathHelper.Max(t1, tmin);
                tmax = MathHelper.Min(t2, tmax);
                // Exit with no collosion as soon as slab intersection becomes empty.
                if (tmin > tmax)
                {
                    return;
                }
            }

            result = tmin;
        }

        /// <summary>
        /// Intersection test against another sphere.
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns>Returns true if another sphere is intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingSphere sphere)
        {
            // Get closest point on AABB.
            Vector3 q;
            Vector3.Clamp(ref sphere.Center, ref Min, ref Max, out q);
            float ds;
            Vector3.DistanceSquared(ref sphere.Center, ref q, out ds);
            return ( ds > sphere.Radius * sphere.Radius )? false: true;
        }

        public void Intersects(ref BoundingSphere sphere, out bool result)
        {
            // Get closest point on AABB.
            Vector3 q;
            Vector3.Clamp(ref sphere.Center, ref Min, ref Max, out q);
            float ds;
            Vector3.DistanceSquared(ref sphere.Center, ref q, out ds);
            result = (ds > sphere.Radius * sphere.Radius) ? false : true;
        }
        #endregion

        #region Contains
        /// <summary>
        /// Contains test against another bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns IntersectionType that contains Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(BoundingBox box)
        {
            if (Max.X < box.Min.X || Min.X > box.Max.X) return ContainmentType.Disjoint;
            if (Max.Y < box.Min.Y || Min.Y > box.Max.Y) return ContainmentType.Disjoint;
            if (Max.Z < box.Min.Z || Min.Z > box.Max.Z) return ContainmentType.Disjoint;

            return (
                Min.X <= box.Min.X && box.Max.X <= Max.X &&
                Min.Y <= box.Min.Y && box.Max.Y <= Max.Y &&
                Min.Z <= box.Min.Z && box.Max.Z <= Max.Z )? ContainmentType.Contains: ContainmentType.Intersects;
        }

        public void Contains(ref BoundingBox box, out ContainmentType result)
        {
            result = ContainmentType.Disjoint;

            if (Max.X < box.Min.X || Min.X > box.Max.X) return;
            if (Max.Y < box.Min.Y || Min.Y > box.Max.Y) return;
            if (Max.Z < box.Min.Z || Min.Z > box.Max.Z) return;

            result = (
                Min.X <= box.Min.X && box.Max.X <= Max.X &&
                Min.Y <= box.Min.Y && box.Max.Y <= Max.Y &&
                Min.Z <= box.Min.Z && box.Max.Z <= Max.Z) ? ContainmentType.Contains : ContainmentType.Intersects;
        }

        /// <summary>
        /// Contains test against another bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns IntersectionType that contains Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(BoundingFrustum frustum)
        {
            if (null == frustum)
            {
                throw new ArgumentNullException("frustum", "FrameworkResources.NullNotAllowed");
            }

            if ( frustum.Intersects( this ) == false ) return ContainmentType.Disjoint;

            // It intersecting, make sure all points are inside box.
            foreach (Vector3 p in frustum.cornerArray)
            {
                if (Contains(p) == ContainmentType.Disjoint)
                {
                    return ContainmentType.Intersects;
                }
            }
            return ContainmentType.Contains;
        }

        /// <summary>
        /// Contains test against another point.
        /// </summary>
        /// <param name="point"></param>
        /// <returns>Returns IntersectionType that contains Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(Vector3 point)
        {
            return  ( Min.X <= point.X && point.X <= Max.X &&
                      Min.Y <= point.Y && point.Y <= Max.Y &&
                      Min.Z <= point.Z && point.Z <= Max.Z )? ContainmentType.Contains: ContainmentType.Disjoint;
        }

        public void Contains(ref Vector3 point, out ContainmentType result)
        {
            result = (Min.X <= point.X && point.X <= Max.X &&
                      Min.Y <= point.Y && point.Y <= Max.Y &&
                      Min.Z <= point.Z && point.Z <= Max.Z) ? ContainmentType.Contains : ContainmentType.Disjoint;
        }

        /// <summary>
        /// Contains test against another sphere.
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns>Returns IntersectionType that contains Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(BoundingSphere sphere)
        {
            // Get closest point on AABB.
            Vector3 q;
            Vector3.Clamp(ref sphere.Center, ref Min, ref Max, out q);
            float ds;
            Vector3.DistanceSquared(ref sphere.Center, ref q, out ds);
            float r = sphere.Radius;

            if ( ds > r * r ) return ContainmentType.Disjoint;

            return (
                Min.X + r <= sphere.Center.X && sphere.Center.X <= Max.X - r && Max.X - Min.X > r &&
                Min.Y + r <= sphere.Center.Y && sphere.Center.Y <= Max.Y - r && Max.Y - Min.Y > r &&
                Min.Z + r <= sphere.Center.Z && sphere.Center.Z <= Max.Z - r && Max.X - Min.X > r ) ? ContainmentType.Contains : ContainmentType.Intersects;
        }

        public void Contains(ref BoundingSphere sphere, out ContainmentType result)
        {
            // Get closest point on AABB.
            Vector3 q;
            Vector3.Clamp(ref sphere.Center, ref Min, ref Max, out q);
            float ds;
            Vector3.DistanceSquared(ref sphere.Center, ref q, out ds);
            float r = sphere.Radius;

            if (ds > r * r)
            {
                result = ContainmentType.Disjoint;
                return;
            }

            result = (
                Min.X + r <= sphere.Center.X && sphere.Center.X <= Max.X - r && Max.X - Min.X > r &&
                Min.Y + r <= sphere.Center.Y && sphere.Center.Y <= Max.Y - r && Max.Y - Min.Y > r &&
                Min.Z + r <= sphere.Center.Z && sphere.Center.Z <= Max.Z - r && Max.X - Min.X > r ) ? ContainmentType.Contains : ContainmentType.Intersects;
        }
        #endregion

        /// <summary>
        /// Suport mappint for Gjk
        /// </summary>
        /// <returns></returns>
        internal void SupportMapping(ref Vector3 v, out Vector3 result)
        {
            result.X = (v.X >= 0.0f) ? Max.X : Min.X;
            result.Y = (v.Y >= 0.0f) ? Max.Y : Min.Y;
            result.Z = (v.Z >= 0.0f) ? Max.Z : Min.Z;
        }

        #region Operator overloads
        public static bool operator ==(BoundingBox a, BoundingBox b)
        {
            return a.Equals(b);
        }

        public static bool operator !=(BoundingBox a, BoundingBox b)
        {
            return (a.Min != b.Min || a.Max != b.Max);
        }
        #endregion
    }
}
