//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using SM = System.Math;

namespace Microsoft.Xna.Framework
{

    public class BoundingFrustum : IEquatable<BoundingFrustum>
    {
        public const int CornerCount = 8;

        #region Properties
        /// <summary>
        /// Gets Near plane of frustum.
        /// </summary>
        public Plane Near   { get { return planes[NearPlaneIndex]; } }
        /// <summary>
        /// Gets Far plane of frustum.
        /// </summary>
        public Plane Far    { get { return planes[FarPlaneIndex]; } }
        /// <summary>
        /// Gets Left plane of frustum.
        /// </summary>
        public Plane Left   { get { return planes[LeftPlaneIndex]; } }
        /// <summary>
        /// Gets Right plane of frustum.
        /// </summary>
        public Plane Right { get { return planes[RightPlaneIndex]; } }
        /// <summary>
        /// Gets Top plane of frustum.
        /// </summary>
        public Plane Top    { get { return planes[TopPlaneIndex]; } }
        /// <summary>
        /// Gets Bottom plane of frustum.
        /// </summary>
        public Plane Bottom { get { return planes[BottomPlaneIndex]; } }

        /// <summary>
        /// Returns 8 corners position of bounding frustum.
        ///     Near    Far
        ///    0----1  4----5
        ///    |    |  |    |
        ///    |    |  |    |
        ///    3----2  7----6
        /// </summary>
        /// <returns></returns>
        public Vector3[] GetCorners()
        {
            // Copy array for avoid user modification of internal data.
			Vector3[] copy = new Vector3[cornerArray.Length];
			cornerArray.CopyTo(copy, 0);
			return copy;
        }

        /// <summary>
        /// It copies bounding box corner points to given array.
        /// </summary>
        /// <param name="corners">vector3 array that has equal or bigger than 8 elements.</param>
        public void GetCorners(Vector3[] corners)
        {
            if (corners == null)
                throw new ArgumentNullException("corners");
            if (corners.Length < CornerCount)
                throw new ArgumentOutOfRangeException("corners", "FrameworkResources.NotEnoughCorners");

            this.cornerArray.CopyTo(corners, 0);
        }

        /// <summary>
        /// Gets Matrix that used when this bounding frustum was instanciated.
        /// </summary>
        public Matrix Matrix { get { return matrix; } set { SetMatrix(ref value); } }
        #endregion

        const int NearPlaneIndex    = 0;
        const int FarPlaneIndex     = 1;
        const int LeftPlaneIndex    = 2;
        const int RightPlaneIndex   = 3;
        const int TopPlaneIndex     = 4;
        const int BottomPlaneIndex  = 5;
        const int NumPlanes         = 6;

        private Matrix      matrix;
        private Plane[]     planes  = new Plane[6];
        internal Vector3[]  cornerArray = new Vector3[CornerCount];
        private Gjk gjk;        // for Gjk based intersection.

        #region Override methods

        // Implement IEquatable<T>
        public bool Equals(BoundingFrustum other)
        {
            if (other == null) return false;
            return matrix == other.matrix;
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            BoundingFrustum other = obj as BoundingFrustum;
            if (other != null)
            {
                equal = (matrix == other.matrix);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return matrix.GetHashCode();
        }

        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{Near:{0} Far:{1} Left:{2} Right:{3} Top:{4} Bottom:{5}}}",
                Near.ToString(), Far.ToString(), Left.ToString(), Right.ToString(), Top.ToString(), Bottom.ToString() );
        }
        #endregion

        /// <summary>
        /// We are not allowed to instanciate this class by using default constructor.
        /// </summary>
        private BoundingFrustum() { }

        /// <summary>
        /// Instanciate a BoundingFrustum from given matrix.
        /// </summary>
        /// <param name="value">Combined matrix that usually takes view * projection matrix.</param>
        public BoundingFrustum(Matrix value)
        {
            SetMatrix(ref value);
        }

        private void SetMatrix(ref Matrix value)
        {
            matrix = value;

            // Compute 6 planes.
            planes[LeftPlaneIndex].Normal.X     = -value.M14 - value.M11;
            planes[LeftPlaneIndex].Normal.Y     = -value.M24 - value.M21;
            planes[LeftPlaneIndex].Normal.Z     = -value.M34 - value.M31;
            planes[LeftPlaneIndex].D            = -value.M44 - value.M41;

            planes[RightPlaneIndex].Normal.X    = -value.M14 + value.M11;
            planes[RightPlaneIndex].Normal.Y    = -value.M24 + value.M21;
            planes[RightPlaneIndex].Normal.Z    = -value.M34 + value.M31;
            planes[RightPlaneIndex].D           = -value.M44 + value.M41;

            planes[TopPlaneIndex].Normal.X      = -value.M14 + value.M12;
            planes[TopPlaneIndex].Normal.Y      = -value.M24 + value.M22;
            planes[TopPlaneIndex].Normal.Z      = -value.M34 + value.M32;
            planes[TopPlaneIndex].D             = -value.M44 + value.M42;

            planes[BottomPlaneIndex].Normal.X   = -value.M14 - value.M12;
            planes[BottomPlaneIndex].Normal.Y   = -value.M24 - value.M22;
            planes[BottomPlaneIndex].Normal.Z   = -value.M34 - value.M32;
            planes[BottomPlaneIndex].D          = -value.M44 - value.M42;

            planes[NearPlaneIndex].Normal.X     = -value.M13;
            planes[NearPlaneIndex].Normal.Y     = -value.M23;
            planes[NearPlaneIndex].Normal.Z     = -value.M33;
            planes[NearPlaneIndex].D            = -value.M43;

            planes[FarPlaneIndex].Normal.X      = -value.M14 + value.M13;
            planes[FarPlaneIndex].Normal.Y      = -value.M24 + value.M23;
            planes[FarPlaneIndex].Normal.Z      = -value.M34 + value.M33;
            planes[FarPlaneIndex].D             = -value.M44 + value.M43;

            // Normalize plnaes.
            for (int i = 0; i < 6; ++i)
            {
                float mag = planes[i].Normal.Length();
                planes[i].Normal /= mag;
                planes[i].D /= mag;
            }

            // Compute near plane corners.
            Ray ray = ComputeIntersectionLine(ref planes[NearPlaneIndex], ref planes[LeftPlaneIndex]);
            cornerArray[0]  = ComputeIntersection(ref planes[TopPlaneIndex], ref ray);
            cornerArray[3]  = ComputeIntersection(ref planes[BottomPlaneIndex], ref ray);

            ray = ComputeIntersectionLine(ref planes[RightPlaneIndex], ref planes[NearPlaneIndex]);
            cornerArray[1]  = ComputeIntersection(ref planes[TopPlaneIndex], ref ray);
            cornerArray[2]  = ComputeIntersection(ref planes[BottomPlaneIndex], ref ray);

            // Compute far plane corners.
            ray = ComputeIntersectionLine(ref planes[LeftPlaneIndex], ref planes[FarPlaneIndex]);
            cornerArray[4]  = ComputeIntersection(ref planes[TopPlaneIndex], ref ray);
            cornerArray[7]  = ComputeIntersection(ref planes[BottomPlaneIndex], ref ray);

            ray = ComputeIntersectionLine(ref planes[FarPlaneIndex], ref planes[RightPlaneIndex]);
            cornerArray[5]  = ComputeIntersection(ref planes[TopPlaneIndex], ref ray);
            cornerArray[6]  = ComputeIntersection(ref planes[BottomPlaneIndex], ref ray);
        }

        private static Ray ComputeIntersectionLine( ref Plane p1, ref Plane p2 )
        {
            Ray ray = new Ray();

            ray.Direction = Vector3.Cross(p1.Normal, p2.Normal);
            float denom = ray.Direction.LengthSquared();
            ray.Position = Vector3.Cross( -p1.D * p2.Normal + p2.D * p1.Normal, ray.Direction ) / denom;

            return ray;
        }

        private static Vector3 ComputeIntersection( ref Plane plane, ref Ray ray )
        {
            float t = (-plane.D - Vector3.Dot(plane.Normal, ray.Position)) / Vector3.Dot(plane.Normal, ray.Direction);
            return ray.Position + ray.Direction * t;
        }

        #region Intersections
        /// <summary>
        /// Intersection test against BoundingBox
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns true if this frustum and given bounding box are intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingBox box)
        {
            bool result;
            Intersects(ref box, out result);
            return result;
        }

        public void Intersects(ref BoundingBox box, out bool result)
        {
            const float etol = 4e-5f;
            const float epsilon = 1e-5f;

            if (gjk == null)
            {
                gjk = new Gjk();
            }
            gjk.Reset();
            Vector3 v;
            Vector3.Subtract(ref cornerArray[0], ref box.Min, out v);
            if (v.LengthSquared() < epsilon)
            {
                Vector3.Subtract(ref cornerArray[0], ref box.Max, out v);
            }

            float vls = Single.MaxValue;
            float err = 0.0f;
            result = false;
            do
            {
                Vector3 nv, sa, sb, w;

                nv.X = -v.X; nv.Y = -v.Y; nv.Z = -v.Z;
                SupportMapping(ref nv, out sa);
                box.SupportMapping(ref v, out sb);
                Vector3.Subtract(ref sa, ref sb, out w);

                float dotVW = v.X * w.X + v.Y * w.Y + v.Z * w.Z;
                if ( dotVW > 0.0f)
                {
                    // v is (considered) a separating axis.
                    return;
                }
                gjk.AddSupportPoint(ref w);
                v = gjk.ClosestPoint;

                // Safe exist in the case it couldn't find closer point than previous iteration.
                float prevVls = vls;
                vls = v.LengthSquared();
                if (prevVls - vls <= epsilon * prevVls)
                {
                    return;
                }

                err = etol * gjk.MaxLengthSquared;

            } while (!gjk.FullSimplex && vls >= err);

            result = true;
        }

        /// <summary>
        /// Intersection test agains Frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns true if this frustum and given frustum are intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingFrustum frustum)
        {
            if (frustum == null)
            {
                throw new ArgumentNullException( "frustum" );
            }

            const float etol = 4e-5f;
            const float epsilon = 1e-5f;

            if (gjk == null)
            {
                gjk = new Gjk();
            }
            gjk.Reset();
            Vector3 v;
            Vector3.Subtract(ref cornerArray[0], ref frustum.cornerArray[0], out v);
            if (v.LengthSquared() < epsilon )
            {
                Vector3.Subtract(ref cornerArray[0], ref frustum.cornerArray[1], out v);
            }

            float vls = Single.MaxValue;
            float err = 0.0f;
            do
            {
                Vector3 nv, sa, sb, w;

                nv.X = -v.X; nv.Y = -v.Y; nv.Z = -v.Z;
                SupportMapping(ref nv, out sa);
                frustum.SupportMapping(ref v, out sb);
                Vector3.Subtract(ref sa, ref sb, out w);

                float dotVW = v.X * w.X + v.Y * w.Y + v.Z * w.Z;
                if (dotVW > 0.0f)
                {
                    // v is (considered) a separating axis.
                    return false;
                }
                gjk.AddSupportPoint(ref w);
                v = gjk.ClosestPoint;

                float prevVls = vls;
                vls = v.LengthSquared();
                err = etol * gjk.MaxLengthSquared;

                // Safe exist in the case it couldn't find closer point than previous iteration.
                if (prevVls - vls <= epsilon * prevVls)
                {
                    return false;
                }

            } while (!gjk.FullSimplex && vls >= err );

            return true;
        }

        /// <summary>
        /// Intersection test against Plane.
        /// </summary>
        /// <param name="plane"></param>
        /// <returns>It returns PlaneIntersectionType that contains Outside, Inside, and Intersecting states.</returns>
        public PlaneIntersectionType Intersects(Plane plane)
        {
            // Check 8 corners of Frustum against plane.
            int insideOutside = 0;
            for( int i = 0; i < 8; ++i )
            {
                float d;
                Vector3.Dot(ref cornerArray[i], ref plane.Normal, out d) ;
                if (d + plane.D > 0.0f)
                {
                    insideOutside |= 0x01;  // Outside
                }
                else
                {
                    insideOutside |= 0x02;  // Inside
                }

                // Early out as soon as Frustum is intersecting against plane.
                if (insideOutside == 0x3)
                {
                    return PlaneIntersectionType.Intersecting;
                }
            }

            return ( insideOutside == 0x1 )? PlaneIntersectionType.Front: PlaneIntersectionType.Back;
        }

        public void Intersects(ref Plane plane, out PlaneIntersectionType result)
        {
            // Check 8 corners of Frustum against plane.
            int insideOutside = 0;
            for( int i = 0; i < 8; ++i )
            {
                float d;
                Vector3.Dot(ref cornerArray[i], ref plane.Normal, out d);
                if (d + plane.D > 0.0f)
                {
                    insideOutside |= 0x01;  // Outside
                }
                else
                {
                    insideOutside |= 0x02;  // Inside
                }

                // Early out as soon as Frustum is intersecting against plane.
                if (insideOutside == 0x3)
                {
                    result = PlaneIntersectionType.Intersecting;
                    return;
                }
            }

            result = ( insideOutside == 0x1 )? PlaneIntersectionType.Front: PlaneIntersectionType.Back;
        }

        /// <summary>
        /// Intersection test against Ray.
        /// </summary>
        /// <param name="ray"></param>
        /// <returns>It returns Nullable&gt;float&lt; it nearest intersection point of ray otherwise it returns null.</returns>
        public Nullable<float> Intersects(Ray ray)
        {
            Nullable<float> result;
            Intersects(ref ray, out result);
            return result;
        }

        public void Intersects(ref Ray ray, out Nullable<float> result)
        {
            // Following algorithm returns non zero distance when ray origin is inside of frustum.
            // So, we check ray origin is inside or not first.
            // Current definition of ray intersection is return 0 distance when ray origin is inside bounding volume.
            ContainmentType containmentType;
            Contains(ref ray.Position, out containmentType);
            if (containmentType == ContainmentType.Contains)
            {
                result = 0;
                return;
            }

            // This algorithm based on "Fast Ray-Convex Polyhedron Intersectin," in James Arvo, ed., Graphics Gems II pp. 247-250
            const float epsilon = 1e-5f;
            float tnear = -Single.MaxValue;
            float tfar = Single.MaxValue;

            result = null;
            foreach (Plane plane in planes)
            {
                Vector3 pn = plane.Normal;

                float vd;
                Vector3.Dot(ref ray.Direction, ref pn, out vd);
                float vn;
                Vector3.Dot(ref ray.Position, ref pn, out vn);
                vn += plane.D;
                if (SM.Abs(vd) < epsilon)
                {
                    // Ray is parallel to plane - check if ray origin is inside plane's
                    if (vn > 0.0f)
                    {
                        // Ray origin is outisde half-space.
                        return;
                    }
                }
                else
                {
                    // Ray not parallel - get distance to plane.
                    float t = -vn / vd;
                    if (vd < 0.0f)
                    {
                        // Front face - T is a near point.
                        if (t > tfar)
                        {
                            return;
                        }
                        if (t > tnear)
                        {
                            // Hit near face.
                            tnear = t;
                        }
                    }
                    else
                    {
                        // back face - T is far point.
                        if (t < tnear)
                        {
                            return;
                        }
                        if (t < tfar)
                        {
                            // Hit far face.
                            tfar = t;
                        }
                    }
                }
            }

            // Survived all tests.
            // Note: if ray originates on polyhedron, may want to change 0.0f to some
            // epsilon to avoid intersecting the originating face.
            float distance = (tnear >= 0.0f)? tnear: tfar;
            if (distance >= 0.0f)
            {
                result = distance;
            }
        }

        /// <summary>
        /// Intersection test against BoundingSphere
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns>Returns true if this frustum and given bounding sphere are intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingSphere sphere)
        {
            bool result;
            Intersects(ref sphere, out result);
            return result;
        }

        public void Intersects(ref BoundingSphere sphere, out bool result)
        {
            const float etol = 4e-5f;
            const float epsilon = 1e-5f;

            if (gjk == null)
            {
                gjk = new Gjk();
            }
            gjk.Reset();
            Vector3 v;
            Vector3.Subtract(ref cornerArray[0], ref sphere.Center, out v);
            if (v.LengthSquared() < epsilon)
            {
                v = Vector3.UnitX;
            }

            float vls = Single.MaxValue;
            float err = 0.0f;
            result = false;
            do
            {
                Vector3 nv, sa, sb, w;

                nv.X = -v.X; nv.Y = -v.Y; nv.Z = -v.Z;
                SupportMapping(ref nv, out sa);
                sphere.SupportMapping(ref v, out sb);
                Vector3.Subtract(ref sa, ref sb, out w);

                float dotVW = v.X * w.X + v.Y * w.Y + v.Z * w.Z;
                if ( dotVW > 0.0f)
                {
                    // v is (considered) a separating axis.
                    return;
                }
                gjk.AddSupportPoint(ref w);
                v = gjk.ClosestPoint;

                // Safe exist in the case it couldn't find closer point than previous iteration.
                float prevVls = vls;
                vls = v.LengthSquared();
                if (prevVls - vls <= epsilon * prevVls)
                {
                    return;
                }

                err = etol * gjk.MaxLengthSquared;

            } while (!gjk.FullSimplex && vls >= err);

            result = true;
        }
        #endregion

        #region Contains
        /// <summary>
        /// Contains test against BoundingBox
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns IntersectionType that contains Inside, Outside, and Intersects.</returns>
        public ContainmentType Contains(BoundingBox box)
        {
            // Check intersection box against 6-planes.
            bool intersecting = false;
            foreach (Plane plane in planes)
            {
                PlaneIntersectionType result = box.Intersects(plane);
                if (result == PlaneIntersectionType.Front)
                {
                    // Outside.
                    return ContainmentType.Disjoint;
                }
                else if (result == PlaneIntersectionType.Intersecting)
                {
                    intersecting = true;
                }
            }
            return (intersecting) ? ContainmentType.Intersects : ContainmentType.Contains;
        }

        public void Contains(ref BoundingBox box, out ContainmentType result)
        {
            // Check intersection box against 6-planes.
            bool intersecting = false;
            foreach (Plane plane in planes)
            {
                PlaneIntersectionType planeResult = box.Intersects(plane);
                if (planeResult == PlaneIntersectionType.Front)
                {
                    // Outside.
                    result = ContainmentType.Disjoint;
                    return;
                }
                else if (planeResult == PlaneIntersectionType.Intersecting)
                {
                    intersecting = true;
                }
            }
            result = (intersecting) ? ContainmentType.Intersects : ContainmentType.Contains;
        }

        /// <summary>
        /// Contains test against frustum
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns IntersectionType that contains Inside, Outside, and Intersects.</returns>
        public ContainmentType Contains(BoundingFrustum frustum)
        {
            if (frustum == null)
            {
                throw new ArgumentNullException( "frustum" );
            }
            ContainmentType result = ContainmentType.Disjoint;

            if (Intersects(frustum))
            {
                result = ContainmentType.Contains;

                // Make sure all corner of another frsutm in this frustum.
                for (int i = 0; i < cornerArray.Length; ++i)
                {
                    if ( Contains(frustum.cornerArray[i]) == ContainmentType.Disjoint )
                    {
                        result = ContainmentType.Intersects;
                        break;
                    }
                }
            }
            return result;
        }

        /// <summary>
        /// Contains test against point.
        /// </summary>
        /// <param name="point"></param>
        /// <returns>Returns IntersectionType that contains Inside, Outside, and Intersects.</returns>
        public ContainmentType Contains(Vector3 point)
        {
            foreach (Plane plane in planes)
            {
                float d = plane.Normal.X * point.X + plane.Normal.Y * point.Y + plane.Normal.Z * point.Z + plane.D;
                if ( d > 1e-5f )
                {
                    return ContainmentType.Disjoint;
                }
            }
            return ContainmentType.Contains;
        }

        public void Contains(ref Vector3 point, out ContainmentType result)
        {
            foreach (Plane plane in planes)
            {
                float d = plane.Normal.X * point.X + plane.Normal.Y * point.Y + plane.Normal.Z * point.Z + plane.D;
                if (d > 1e-5f)
                {
                    result = ContainmentType.Disjoint;
                    return;
                }
            }
            result = ContainmentType.Contains;
        }

        /// <summary>
        /// Contains test against bounding sphere.
        /// </summary>
        /// <returns>Returns IntersectionType that contains Inside, Outside, and Intersects.</returns>
        public ContainmentType Contains(BoundingSphere sphere)
        {
            // This algorithm based on Real-Time Rendering Second Edittion, pp.609
            // This is no acculate test that judges as intersecting when the sphere placed at corner or edge of of frustum.
            Vector3 p = sphere.Center;
            float r = sphere.Radius;
            int insideCount = 0;

            foreach (Plane plane in planes)
            {
                float dot = plane.Normal.X * p.X + plane.Normal.Y * p.Y + plane.Normal.Z * p.Z;
                float dist = dot + plane.D;
                if (dist > r)
                {
                    return ContainmentType.Disjoint;
                }

                // Sphere is completly inside of plane.
                if (dist < -r)
                {
                    insideCount++;
                }
            }

            return (insideCount == 6) ? ContainmentType.Contains : ContainmentType.Intersects;
        }

        public void Contains(ref BoundingSphere sphere, out ContainmentType result)
        {
            // This algorithm based on Real-Time Rendering Second Edittion, pp.609
            // This is no acculate test that judges as intersecting when the sphere placed at corner or edge of of frustum.
            Vector3 p = sphere.Center;
            float r = sphere.Radius;
            int insideCount = 0;

            foreach (Plane plane in planes)
            {
                float dot = plane.Normal.X * p.X + plane.Normal.Y * p.Y + plane.Normal.Z * p.Z;
                float dist = dot + plane.D;
                if (dist > r)
                {
                    result = ContainmentType.Disjoint;
                    return;
                }

                // Sphere is completly inside of plane.
                if (dist < -r)
                {
                    insideCount++;
                }
            }

            result = (insideCount == 6) ? ContainmentType.Contains : ContainmentType.Intersects;
        }
        #endregion

        /// <summary>
        /// Support mapping for GJK.
        /// </summary>
        /// <returns></returns>
        internal void SupportMapping(ref Vector3 v, out Vector3 result)
        {
            int maxIdx = 0;
            float maxDot;
            Vector3.Dot(ref cornerArray[0], ref v, out maxDot);
            for (int i = 1; i < cornerArray.Length; ++i)
            {
                Vector3 nv;
                nv.X = -v.X; nv.Y = -v.Y; nv.Z = -v.Z;

                float dot;
                Vector3.Dot(ref cornerArray[i], ref v, out dot);
                if (dot > maxDot)
                {
                    maxIdx = i;
                    maxDot = dot;
                }
            }
            result = cornerArray[maxIdx];
        }

        #region Operator overloads
        public static bool operator ==(BoundingFrustum a, BoundingFrustum b)
        {
            return Object.Equals(a, b);
        }

        public static bool operator !=(BoundingFrustum a, BoundingFrustum b)
        {
            return !Object.Equals(a, b);
        }
        #endregion
    }
}
