//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"

using System;
using System.Globalization;
using System.Collections.Generic;
using System.Text;

namespace Microsoft.Xna.Framework
{

    public struct BoundingSphere : IEquatable<BoundingSphere>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public Vector3 Center;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Radius;

        #region Constructors
        public BoundingSphere(Vector3 center, float radius)
        {
            if (radius < 0.0f)
            {
                throw new ArgumentException( "FrameworkResources.NegativeRadius" );
            }
            Center = center;
            Radius = radius;
        }
        #endregion

        #region Override methods

        // Implement IEquatable<T>
        public bool Equals(BoundingSphere other)
        {
            return (Center == other.Center && Radius == other.Radius);
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is BoundingSphere)
            {
                equal = Equals((BoundingSphere)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return Center.GetHashCode() + Radius.GetHashCode();
        }

        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{Center:{0} Radius:{1}}}", Center.ToString(), Radius.ToString(ci));
        }
        #endregion

        /// <summary>
        /// Merge two bounding spheres and create another bounding sphere.
        /// </summary>
        /// <param name="original"></param>
        /// <param name="additional"></param>
        /// <returns>Merged bounding sphere.</returns>
        public static BoundingSphere CreateMerged(BoundingSphere original, BoundingSphere additional)
        {
            Vector3 v;
            Vector3.Subtract(ref additional.Center, ref original.Center, out v);
            float d = v.Length();
            float r1 = original.Radius;
            float r2 = additional.Radius;

            if ( r1 + r2 >= d )
            {
                if (r1 - r2 >= d)
                {
                    // original sphere contains addtional sphere.
                    return original;
                }
                else if ( r2 - r1 >= d )
                {
                    // additional sphere contains original sphere.
                    return additional;
                }
            }

            Vector3 n = v * (1.0f / d);
            float t1 = MathHelper.Min( -r1, d - r2 );
            float t2 = MathHelper.Max( r1, d + r2 );
            float t_5 = ( t2 - t1 ) * 0.5f;

            BoundingSphere result;

            result.Center = original.Center + n * ( t_5 + t1 );
            result.Radius = t_5;

            return result;
        }

        public static void CreateMerged(ref BoundingSphere original, ref BoundingSphere additional, out BoundingSphere result)
        {
            Vector3 v;
            Vector3.Subtract(ref additional.Center, ref original.Center, out v);
            float d = v.Length();
            float r1 = original.Radius;
            float r2 = additional.Radius;

            if ( r1 + r2 >= d )
            {
                if (r1 - r2 >= d)
                {
                    // original sphere contains addtional sphere.
                    result = original;
                    return;
                }
                else if ( r2 - r1 >= d )
                {
                    // additional sphere contains original sphere.
                    result = additional;
                    return;
                }
            }

            Vector3 n = v * (1.0f / d);
            float t1 = MathHelper.Min( -r1, d - r2 );
            float t2 = MathHelper.Max( r1, d + r2 );
            float t_5 = ( t2 - t1 ) * 0.5f;

            result.Center = original.Center + n * (t_5 + t1);
            result.Radius = t_5;
        }

        /// <summary>
        /// Create bounding sphere from given bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns></returns>
        public static BoundingSphere CreateFromBoundingBox(BoundingBox box)
        {
            BoundingSphere result;

            Vector3.Lerp( ref box.Min, ref box.Max, 0.5f, out result.Center );

            float dist;
            Vector3.Distance(ref box.Min, ref box.Max, out dist);
            result.Radius = dist * 0.5f;

            return result;
        }

        public static void CreateFromBoundingBox(ref BoundingBox box, out BoundingSphere result)
        {
            Vector3.Lerp(ref box.Min, ref box.Max, 0.5f, out result.Center);
            float dist;
            Vector3.Distance(ref box.Min, ref box.Max, out dist);
            result.Radius =  dist * 0.5f;
        }

        /// <summary>
        /// Create bounding sphere from points.
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        public static BoundingSphere CreateFromPoints(IEnumerable<Vector3> points )
        {
            if (points == null)
                throw new ArgumentNullException("points");

            IEnumerator<Vector3> enumerator = points.GetEnumerator();
            if (enumerator.MoveNext() == false)
            {
                throw new ArgumentException("FrameworkResources.BoundingSphereZeroPoints");
            }

            // Find the approximate smallest enclosing bounding sphere for a set of 
            // points. Exact computation of the smallest enclosing bounding sphere is 
            // possible but is slower and requires a more complex algorithm.
            // The algorithm is based on  Jack Ritter, "An Efficient Bounding Sphere", 
            // Graphics Gems.

            // Find the points with minimum and maximum x, y, and z
            Vector3 minX, maxX, minY, maxY, minZ, maxZ;
            minX = maxX = minY = maxY = minZ = maxZ = enumerator.Current;

            foreach (Vector3 point in points)
            {
                if (point.X < minX.X)
                    minX = point;

                if (point.X > maxX.X)
                    maxX = point;

                if (point.Y < minY.Y)
                    minY = point;

                if (point.Y > maxY.Y)
                    maxY = point;

                if (point.Z < minZ.Z)
                    minZ = point;

                if (point.Z > maxZ.Z)
                    maxZ = point;
            }
              
            // Use the min/max pair that are farthest apart to form the initial sphere.
            float distX, distY, distZ;
            Vector3.Distance(ref maxX, ref minX, out distX);
            Vector3.Distance(ref maxY, ref minY, out distY);
            Vector3.Distance(ref maxZ, ref minZ, out distZ);
            
            Vector3 center;
            float   radius;

            if ( distX > distY )
            {
                if ( distX > distZ )
                {
                    // Use min/max x.
                    Vector3.Lerp(ref maxX, ref minX, 0.5f, out center);
                    radius = distX * 0.5f;
                }
                else
                {
                    // Use min/max z.
                    Vector3.Lerp(ref maxZ, ref minZ, 0.5f, out center);
                    radius = distZ * 0.5f;
                }
            }
            else // Y >= X
            {
                if ( distY > distZ )
                {
                    // Use min/max y.
                    Vector3.Lerp(ref maxY, ref minY, 0.5f, out center);
                    radius = distY * 0.5f;
                }
                else
                {
                    // Use min/max z.
                    Vector3.Lerp(ref maxZ, ref minZ, 0.5f, out center);
                    radius = distZ * 0.5f;
                }
            }
            
            // Add any points not inside the sphere.
            foreach (Vector3 point in points)
            {
                Vector3 delta;
                delta.X = point.X - center.X;
                delta.Y = point.Y - center.Y;
                delta.Z = point.Z - center.Z;
                float dist = delta.Length();
                if ( dist > radius)
                {
                    // Adjust sphere to include the new point.
                    radius = (radius + dist) * 0.5f;
                    center += ( 1.0f - radius / dist) * delta;
                }
            }

            BoundingSphere result;
            result.Center = center;
            result.Radius = radius;

            return result;
        }

        /// <summary>
        /// Create bounding sphere from given bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns></returns>
        public static BoundingSphere CreateFromFrustum(BoundingFrustum frustum)
        {
            if (frustum == null)
            {
                throw new ArgumentNullException( "frustum" );
            }
            return CreateFromPoints( frustum.cornerArray );
        }

        #region Intersections
        /// <summary>
        /// Intersection test against another bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns true if another bounding box is intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingBox box)
        {
            // Get closest point on AABB.
            Vector3 q;
            Vector3.Clamp(ref Center, ref box.Min, ref box.Max, out q);
            float ds;
            Vector3.DistanceSquared(ref Center, ref q, out ds);
            return (ds > Radius * Radius) ? false : true;
        }

        public void Intersects(ref BoundingBox box, out bool result)
        {
            // Get closest point on AABB.
            Vector3 q;
            Vector3.Clamp(ref Center, ref box.Min, ref box.Max, out q);
            float ds;
            Vector3.DistanceSquared(ref Center, ref q, out ds);
            result = (ds > Radius * Radius) ? false : true;
        }

        /// <summary>
        /// Intersection test against another bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns true if another bounding frustum is intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingFrustum frustum)
        {
            if (null == frustum)
            {
                throw new ArgumentNullException("frustum", "FrameworkResources.NullNotAllowed");
            }

            bool result;
            frustum.Intersects(ref this, out result);
            return result;
        }

        /// <summary>
        /// Intersection test against another plane.
        /// </summary>
        /// <param name="plane"></param>
        /// <returns>Returns PlaneIntersectionType that contins value Outside, Inside, or Intersects.</returns>
        public PlaneIntersectionType Intersects(Plane plane)
        {
            return plane.Intersects(this);
        }

        public void Intersects(ref Plane plane, out PlaneIntersectionType result)
        {
            plane.Intersects(ref this, out result);
        }

        /// <summary>
        /// Intersection test against another ray.
        /// </summary>
        /// <param name="ray"></param>
        /// <returns></returns>
        public Nullable<float> Intersects(Ray ray)
        {
            return ray.Intersects(this);
        }

        public void Intersects(ref Ray ray, out Nullable<float> result)
        {
            ray.Intersects(ref this, out result);
        }

        /// <summary>
        /// Intersection test against another bounding sphere.
        /// </summary>
        /// <returns>Returns true if another bounding sphere is intersecting, otherwise returns false.</returns>
        public bool Intersects(BoundingSphere sphere)
        {
            float ds;
            Vector3.DistanceSquared(ref Center, ref sphere.Center, out ds);
            float r1 = Radius;
            float r2 = sphere.Radius;
            return ( r1 * r1 + 2.0f * r1 * r2 + r2 * r2 > ds) ? true : false;
        }

        public void Intersects(ref BoundingSphere sphere, out bool result)
        {
            float ds;
            Vector3.DistanceSquared(ref Center, ref sphere.Center, out ds);
            float r1 = Radius;
            float r2 = sphere.Radius;
            result = (r1 * r1 + 2.0f * r1 * r2 + r2 * r2 > ds) ? true : false;
        }
        #endregion

        #region Contains
        /// <summary>
        /// Contains test against another bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns IntersectionType that contians value Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(BoundingBox box)
        {
            if (!box.Intersects(this)) return ContainmentType.Disjoint;

            // Make sure 8 corners are inside of sphere.
            float rr = Radius * Radius;

            Vector3 d;
            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return ContainmentType.Intersects;

            return ContainmentType.Contains;
        }

        public void Contains(ref BoundingBox box, out ContainmentType result)
        {
            bool intersects;
            box.Intersects(ref this, out intersects);
            if (!intersects)
            {
                result = ContainmentType.Disjoint;
                return;
            }

            // Make sure 8 corners are inside of sphere.
            float rr = Radius * Radius;

            result = ContainmentType.Intersects;

            Vector3 d;
            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Max.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Max.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Max.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return;

            d.X = Center.X - box.Min.X; d.Y = Center.Y - box.Min.Y; d.Z = Center.Z - box.Min.Z;
            if (d.LengthSquared() > rr) return;

            result = ContainmentType.Contains;
        }

        /// <summary>
        /// Contains test against another bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns IntersectionType that contians value Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(BoundingFrustum frustum)
        {
            if (null == frustum)
            {
                throw new ArgumentNullException("frustum", "FrameworkResources.NullNotAllowed");
            }

            if (frustum.Intersects(this) == false) return ContainmentType.Disjoint;

            // It intersecting, make sure all points are inside sphere.
            float rr = Radius * Radius;
            foreach (Vector3 p in frustum.cornerArray)
            {
                Vector3 d;
                d.X = p.X - Center.X;
                d.Y = p.Y - Center.Y;
                d.Z = p.Z - Center.Z;
                if (d.LengthSquared() > rr)
                {
                    return ContainmentType.Intersects;
                }
            }
            return ContainmentType.Contains;
        }

        /// <summary>
        /// Contains test against another point.
        /// </summary>
        /// <param name="point"></param>
        /// <returns>Returns IntersectionType that contians value Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(Vector3 point)
        {
            float ds = Vector3.DistanceSquared(point, Center);
            return ds < Radius * Radius ? ContainmentType.Contains : ContainmentType.Disjoint;
        }

        public void Contains(ref Vector3 point, out ContainmentType result)
        {
            float ds;
            Vector3.DistanceSquared(ref point, ref Center, out ds);
            result = ds < Radius * Radius ? ContainmentType.Contains : ContainmentType.Disjoint;
        }

        /// <summary>
        /// Contains test against another bounding sphere.
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns>Returns IntersectionType that contians value Contains, Intersects, or Disjoint.</returns>
        public ContainmentType Contains(BoundingSphere sphere)
        {
            float d;
            Vector3.Distance(ref Center, ref sphere.Center, out d);
            float r1 = Radius;
            float r2 = sphere.Radius;
            return (r1 + r2 >= d) ?
                ((r1 - r2 >= d) ? ContainmentType.Contains : ContainmentType.Intersects) : ContainmentType.Disjoint;
        }

        public void Contains(ref BoundingSphere sphere, out ContainmentType result)
        {
            float d;
            Vector3.Distance(ref Center, ref sphere.Center, out d);
            float r1 = Radius;
            float r2 = sphere.Radius;
            result = (r1 + r2 >= d) ?
                ((r1 - r2 >= d) ? ContainmentType.Contains : ContainmentType.Intersects) : ContainmentType.Disjoint;
        }
        #endregion

        /// <summary>
        /// Suport mappint for Gjk
        /// </summary>
        /// <returns></returns>
        internal void SupportMapping(ref Vector3 v, out Vector3 result)
        {
            // When length of v close to zero, Gjk intersection algorithm terminated.
            // Thus, it guranteed Gjk algorithm doens't passed zero length v value to this method.
            float len = v.Length();
            float rOverLen = Radius / len;
            result.X = Center.X + (v.X * rOverLen);
            result.Y = Center.Y + (v.Y * rOverLen);
            result.Z = Center.Z + (v.Z * rOverLen);
        }

        #region Transform
        /// <summary>
        /// Transform the bounding sphere returning a new one
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public BoundingSphere Transform(Matrix matrix)
        {
            BoundingSphere result = new BoundingSphere();
            result.Center = Vector3.Transform(this.Center, matrix);
            
            float sqScaleX = matrix.M11 * matrix.M11 + matrix.M12 * matrix.M12 + matrix.M13 * matrix.M13;
            float sqScaleY = matrix.M21 * matrix.M21 + matrix.M22 * matrix.M22 + matrix.M23 * matrix.M23;
            float sqScaleZ = matrix.M31 * matrix.M31 + matrix.M32 * matrix.M32 + matrix.M33 * matrix.M33;
            // Pick most significant scale.
            float sqScale = Math.Max( sqScaleX, Math.Max( sqScaleY, sqScaleZ ) );
            result.Radius = this.Radius * (float)Math.Sqrt( sqScale );

            return result;
        }
        public void Transform(ref Matrix matrix, out BoundingSphere result)
        {
            result.Center = Vector3.Transform(this.Center, matrix);

            float sqScaleX = matrix.M11 * matrix.M11 + matrix.M12 * matrix.M12 + matrix.M13 * matrix.M13;
            float sqScaleY = matrix.M21 * matrix.M21 + matrix.M22 * matrix.M22 + matrix.M23 * matrix.M23;
            float sqScaleZ = matrix.M31 * matrix.M31 + matrix.M32 * matrix.M32 + matrix.M33 * matrix.M33;
            // Pick most significant scale.
            float sqScale = Math.Max( sqScaleX, Math.Max( sqScaleY, sqScaleZ ) );
            
            result.Radius = this.Radius * (float)Math.Sqrt( sqScale );
        }
        #endregion

        #region Operator overloads
        public static bool operator ==(BoundingSphere a, BoundingSphere b)
        {
            return a.Equals(b);
        }

        public static bool operator !=(BoundingSphere a, BoundingSphere b)
        {
            return (a.Center != b.Center || a.Radius != b.Radius );
        }
        #endregion

    }
}
