//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Collections.Generic;
using System.Text;

namespace Microsoft.Xna.Framework
{
    delegate Vector3 SupportMapping( Object sender, Vector3 v );

    /// <summary>
    /// GJK (Gilbert-Johnson-Keerthi) Algorithm from "Collision Detection in interactive 3D environments".
    /// </summary>
    class Gjk
    {
        /// <summary>
        /// Return true if simplex is full (has 4 supporting point)
        /// </summary>
        public bool FullSimplex { get { return simplexBits == 0xf; } }

        /// <summary>
        /// Return max squared length of current simplex.
        /// </summary>
        public float MaxLengthSquared { get { return maxLengthSq; } }

        /// <summary>
        /// Return current closest point.
        /// </summary>
        public Vector3 ClosestPoint { get { return closestPoint; } }

        public Gjk()
        {
            y           = new Vector3[4];
            yLengthSq   = new float[4];

            edges           = new Vector3[4][]{ new Vector3[4], new Vector3[4], new Vector3[4], new Vector3[4] };
            edgeLengthSq    = new float[4][]{ new float[4], new float[4], new float[4], new float[4] };

            det             = new float[16][];
            for (int i = 0; i < 16; ++i)
            {
                det[i] = new float[4];
            }
        }

        /// <summary>
        /// Reset current simplex and ready for new iteration.
        /// </summary>
        public void Reset()
        {
            simplexBits = 0;
            maxLengthSq = 0.0f;
        }

#if unused
        /// Example of intersection algorithm.
        /// <summary>
        /// Generalized version of GJK separate axis algorithm.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="Sa"></param>
        /// <param name="b"></param>
        /// <param name="Sb"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public bool Intersects(object a, SupportMapping Sa, object b, SupportMapping Sb, Vector3 v)
        {
            Reset();

            do
            {
                Vector3 nv = -v;
                Vector3 w = Sa(a, nv) - Sb(b, nv);
                if (Contains(w) || Vector3.Dot(v, w) > 0.0f)
                {
                    // v is (considered) a separating axis.
                    return false;
                }
                AddSupportPoint(w);
                v = ClosestPoint;
            } while ( !FullSimplex && v.LengthSquared() >= MaxLengthSquared );

            return true;
        }

        /// <summary>
        /// returns true if given vector is part of simplex.
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public bool Contains(Vector3 v)
        {
            for (int i = BitsToIndices[simplexBits]; i != 0; i >>= 3)
            {
                int idx = (i & 0x7) - 1;
                if (y[idx] == v)
                {
                    return true;
                }
            }
            return false;
        }
#endif

        /// <summary>
        /// Add new support point and update simplex.
        /// </summary>
        /// <param name="newPoint"></param>
        /// <returns></returns>
        public bool AddSupportPoint(ref Vector3 newPoint)
        {
            int newIndex = ( BitsToIndices[simplexBits ^ 0xf] & 0x7 ) - 1;
            y[newIndex] = newPoint;
            yLengthSq[newIndex] = newPoint.LengthSquared();

            // Compute edges.
            for (int i = BitsToIndices[simplexBits]; i != 0; i >>= 3)
            {
                int idx = (i & 0x7) - 1;
                Vector3 edge = y[idx] - newPoint;
                edges[idx][newIndex] = edge;
                edges[newIndex][idx] = -edge;
                edgeLengthSq[newIndex][idx] = edgeLengthSq[idx][newIndex] = edge.LengthSquared();
            }

            UpdateDeterminant(newIndex);

            return UpdateSimplex(newIndex);
        }

        #region private methods.

        private static float Dot(ref Vector3 a, ref Vector3 b)
        {
            return a.X * b.X + a.Y * b.Y + a.Z * b.Z;
        }

        private void UpdateDeterminant( int xmIdx )
        {
            int xmBit = 1 << xmIdx;

            // delta[ {yj}, j ] == 1
            det[xmBit][xmIdx] = 1.0f;

            int indices = BitsToIndices[simplexBits];
            for (int x1 = indices, i = 0; x1 != 0; x1 >>= 3, ++i)
            {
                int x1Idx = ( x1 & 0x07 ) - 1;
                int x1Bit = 1 << x1Idx;

                // Y:{ x1, xm }
                int x1_xm = x1Bit | xmBit;

                // IX:{xm}, {yj}={x1}
                det[x1_xm][x1Idx] = Dot(ref edges[xmIdx][x1Idx], ref y[xmIdx]);
                // IX:{x1}, {yj}={xm}
                det[x1_xm][xmIdx] = Dot(ref edges[x1Idx][xmIdx], ref y[x1Idx]);

                for (int x2 = indices, j = 0; j < i; x2 >>= 3, ++j)
                {
                    int x2Idx = (x2 & 0x07) - 1;
                    int x2Bit = 1 << x2Idx;

                    // Y:{x1, x2, xm}
                    int x1_x2_xm = x1_xm | x2Bit;

                    // IX:{x1,xm}, {yj}={x2}

                    int k = ( edgeLengthSq[x1Idx][x2Idx] < edgeLengthSq[xmIdx][x2Idx] )? x1Idx: xmIdx;  // Use closest point as yk.
                    det[x1_x2_xm][x2Idx] =  det[x1_xm][x1Idx] * Dot(ref edges[k][x2Idx], ref y[x1Idx]) +
                                            det[x1_xm][xmIdx] * Dot(ref edges[k][x2Idx], ref y[xmIdx]);
                    // IX:{x2,xm}, {yj}={x1}
                    k = ( edgeLengthSq[x2Idx][x1Idx] < edgeLengthSq[xmIdx][x1Idx] )? x2Idx: xmIdx;       // Use closest point as yk.
                    det[x1_x2_xm][x1Idx] =  det[x2Bit|xmBit][x2Idx] * Dot( ref edges[k][x1Idx], ref y[x2Idx] ) +
                                            det[x2Bit|xmBit][xmIdx] * Dot( ref edges[k][x1Idx], ref y[xmIdx] );
                    // IX:{x1,x2}, {yj}={xm}
                    k = ( edgeLengthSq[x1Idx][xmIdx] < edgeLengthSq[x2Idx][xmIdx] )? x1Idx: x2Idx;      // Use closest point as yk.
                    det[x1_x2_xm][xmIdx] =  det[x1Bit|x2Bit][x2Idx] * Dot( ref edges[k][xmIdx], ref y[x2Idx] ) +
                                            det[x1Bit|x2Bit][x1Idx] * Dot( ref edges[k][xmIdx], ref y[x1Idx] );
                }
            }

            // And full-simplex case Y:{x1, x2, x3, xm}
            if ( ( simplexBits | xmBit ) == 0xf )
            {
                // IX:{ x2, x3, x4 } {yj}={x1}
                int k = ( edgeLengthSq[1][0] < edgeLengthSq[2][0] )?
                            ( ( edgeLengthSq[1][0] < edgeLengthSq[3][0] )? 1: 3 ) :
                            ( ( edgeLengthSq[2][0] < edgeLengthSq[3][0] )? 2: 3 );
                det[0xf][0] =   det[0xe][1] * Dot( ref edges[k][0], ref y[1] ) + 
                                det[0xe][2] * Dot( ref edges[k][0], ref y[2] ) + 
                                det[0xe][3] * Dot( ref edges[k][0], ref y[3] );

                // IX:{x1, x3, x4 } {yj}={x2}
                k = (edgeLengthSq[0][1] < edgeLengthSq[2][1]) ?
                            ((edgeLengthSq[0][1] < edgeLengthSq[3][1]) ? 0 : 3) :
                            ((edgeLengthSq[2][1] < edgeLengthSq[3][1]) ? 2 : 3);
                det[0xf][1] =   det[0xd][0] * Dot( ref edges[k][1], ref y[0] ) + 
                                det[0xd][2] * Dot( ref edges[k][1], ref y[2] ) + 
                                det[0xd][3] * Dot( ref edges[k][1], ref y[3] );

                // IX:{x1, x2, x4 } {yj}={x3}
                k = (edgeLengthSq[0][2] < edgeLengthSq[1][2]) ?
                            ((edgeLengthSq[0][2] < edgeLengthSq[3][2]) ? 0 : 3) :
                            ((edgeLengthSq[1][2] < edgeLengthSq[3][2]) ? 1 : 3);
                det[0xf][2] =   det[0xb][0] * Dot( ref edges[k][2], ref y[0] ) + 
                                det[0xb][1] * Dot( ref edges[k][2], ref y[1] ) + 
                                det[0xb][3] * Dot( ref edges[k][2], ref y[3] );

                // IX:{x1, x2, x3 } {yj}={x4}
                k = (edgeLengthSq[0][3] < edgeLengthSq[1][3]) ?
                            ((edgeLengthSq[0][3] < edgeLengthSq[2][3]) ? 0 : 2) :
                            ((edgeLengthSq[1][3] < edgeLengthSq[2][3]) ? 1 : 2);
                det[0xf][3] =   det[0x7][0] * Dot( ref edges[k][3], ref y[0] ) + 
                                det[0x7][1] * Dot( ref edges[k][3], ref y[1] ) + 
                                det[0x7][2] * Dot( ref edges[k][3], ref y[2] );
            }
        }

        /// <summary>
        /// Update simplex that smallest X is less than Y
        /// </summary>
        /// <param name="newIndex"></param>
        /// <returns></returns>
        private bool UpdateSimplex(int newIndex)
        {
            int yBits = simplexBits | (1 << newIndex);
            int jBit = 1 << newIndex;

            // Find new simplex set that satisfies rules.
            for (int xBits = simplexBits; xBits != 0; xBits--)
            {
                if ((xBits & yBits) == xBits)
                {
                    if (IsSatisfiesRule(xBits | jBit, yBits ))
                    {
                        simplexBits = xBits | jBit;
                        closestPoint = ComputeClosestPoint();
                        return true;
                    }
                }
            }

            // If there are only one point in simplex, new support point is closest point.
            // So, we don't need to compute closest point by using detaminant.
            bool result = false;
            if (IsSatisfiesRule(jBit, yBits))
            {
                simplexBits = jBit;
                closestPoint = y[newIndex];
                maxLengthSq = yLengthSq[newIndex];
                result = true;
            }

            return result;
        }

        /// <summary>
        /// return current closest point.
        /// </summary>
        /// <returns></returns>
        private Vector3 ComputeClosestPoint()
        {
            float sum = 0.0f;

            Vector3 v = Vector3.Zero;
            maxLengthSq = 0.0f;
            for (int i = BitsToIndices[simplexBits]; i != 0; i >>= 3)
            {
                int idx = (i & 0x7) - 1;
                float di = det[simplexBits][idx];
                sum += di;
                v += y[idx] * di;

                maxLengthSq = MathHelper.Max(maxLengthSq, yLengthSq[idx]);
            }

            return v / sum;
        }

        /// <summary>
        /// Check specfied x subsets is satisfies rules.
        /// </summary>
        /// <param name="xBits"></param>
        /// <param name="yBits"></param>
        /// <returns></returns>
        private bool IsSatisfiesRule(int xBits, int yBits)
        {
            bool result = true;

            for (int indices = BitsToIndices[yBits]; indices != 0; indices >>= 3)
            {
                int idx = (indices & 0x7) - 1;
                int bit = 1 << idx;

                if ((bit & xBits) != 0)
                {
                    // must be lamda[i] > 0
                    if (det[xBits][idx] <= 0.0f)
                    {
                        result = false;
                        break;
                    }
                }
                else
                {
                    // must be lamda[j] <= 0
                    if (det[xBits | bit][idx] > 0.0f)
                    {
                        result = false;
                        break;
                    }
                }
            }

            return result;
        }

        #endregion

        // Convert table of bits to packed indices.
        private static int[] BitsToIndices = new int[16]{
            0,                          // 0x0 (0000) {}
            (1<<0),                     // 0x1 (0001) {1}
            (2<<0),                     // 0x2 (0010) {2}
            (1<<0)|(2<<3),              // 0x3 (0011) {1, 2}
            (3<<0),                     // 0x4 (0100) {3}
            (1<<0)|(3<<3),              // 0x5 (0101) {1, 3}
            (2<<0)|(3<<3),              // 0x6 (0110) {2, 3}
            (1<<0)|(2<<3)|(3<<6),       // 0x7 (0111) {1, 2, 3}
            (4<<0),                     // 0x8 (1000) {4}
            (1<<0)|(4<<3),              // 0x9 (1001) {1, 4}
            (2<<0)|(4<<3),              // 0xa (1010) {2, 4}
            (1<<0)|(2<<3)|(4<<6),       // 0xb (1011) {1, 2, 4}
            (3<<0)|(4<<3),              // 0xc (1100) {3, 4}
            (1<<0)|(3<<3)|(4<<6),       // 0xd (1101) {1, 3, 4}
            (2<<0)|(3<<3)|(4<<6),       // 0xe (1110) {2, 3, 4}
            (1<<0)|(2<<3)|(3<<6)|(4<<9) // 0xf (1111) {1, 2, 3, 4}
        };

        private Vector3     closestPoint;   
        private Vector3[]   y;              // Support points of A - B.
        private float[]     yLengthSq;

        private Vector3[][] edges;          // Cache for y[i] - y[j]
        private float[][]   edgeLengthSq;   // Cache for legth squared of y[i] - y[j].
        private float[][]   det;            // Cache for detaminant, delta[X,i]

        private int simplexBits;
        private float maxLengthSq;
    }
}
