//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
using System;
using System.Collections.Generic;
using System.Text;
using SM = System.Math;

#if !XBOX360 && !ZUNE && !WINDOWSMOBILE    // We have one module on Xbox360 side.
[assembly: CLSCompliant(true)]
#endif

namespace Microsoft.Xna.Framework
{
    public static class MathHelper
    {
        #region Constants
        /// <summary>
        /// Nutulal number
        /// </summary>
        public const float E = 2.71828182845904523536f;

        /// <summary>
        /// log2(e)
        /// </summary>
        public const float Log2E = 1.44269504088896340736f;

        /// <summary>
        /// log10(e)
        /// </summary>
        public const float Log10E = 0.434294481903251827651f;

        /// <summary>
        /// Pi
        /// </summary>
        public const float Pi = 3.14159265358979323846f;

        /// <summary>
        /// 2 * Pi
        /// </summary>
        public const float TwoPi = 6.28318530717958647692f;

        /// <summary>
        /// Pi over 2
        /// </summary>
        public const float PiOver2 = 1.57079632679489661923f;

        /// <summary>
        /// Pi over 4
        /// </summary>
        public const float PiOver4 = 0.785398163397448309616f;
        #endregion

        /// <summary>
        /// Convert degrees to radian
        /// </summary>
        /// <param name="degrees"></param>
        /// <returns></returns>
        public static float ToRadians(float degrees)
        {
            return degrees * (Pi / 180.0f);
        }

        /// <summary>
        /// Convert radians to degrees
        /// </summary>
        /// <returns></returns>
        public static float ToDegrees(float radians)
        {
            return radians * (180.0f/Pi);
        }

        public static float Distance(float value1, float value2)
        {
            return (float)Math.Abs(value1 - value2);
        }

        public static float Min(float value1, float value2)
        {
            return System.Math.Min(value1, value2);
        }

        public static float Max(float value1, float value2)
        {
            return System.Math.Max(value1, value2);
        }

        public static float Clamp(float value, float min, float max)
        {
            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.
            value = (value > max) ? max : value;
            value = (value < min) ? min : value;
            return value;
        }

        public static float Lerp(float value1, float value2, float amount)
        {
            return value1 + (value2 - value1) * amount;
        }

        /// <summary>
        /// Returns a point in Barycentric coordinates, using the specified values.
        /// value1 + amount1( value2 - value1 ) + amount2( value3 - value1 )
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="amount1"></param>
        /// <param name="amount2"></param>
        /// <returns></returns>
        public static float Barycentric(float value1, float value2, float value3, float amount1, float amount2)
        {
            return value1 + amount1 * (value2 - value1) + amount2 * (value3 - value1);
        }

        public static float SmoothStep(float value1, float value2, float amount)
        {
            float t = Clamp(amount, 0.0f, 1.0f);
            return Lerp(value1, value2, t * t * (3.0f - 2.0f * t));
        }

        /// <summary>
        /// Perfroms a Catmull-Rom interpolation, using the specfied float values.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="value4"></param>
        /// <param name="amount">Weighting factor.</param>
        /// <returns></returns>
        public static float CatmullRom(float value1, float value2, float value3, float value4, float amount)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;
            return 0.5f * (
                        (2.0f * value2) +
                        (-value1 + value3) * amount +
                        (2.0f * value1 - 5.0f * value2 + 4.0f * value3 - value4) * t2 +
                        (-value1 + 3.0f * value2 - 3.0f * value3 + value4) * t3
                    );
        }

        /// <summary>
        /// Performs a Hermite spline interpolation, using the specfied float values.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="tangent1"></param>
        /// <param name="value2"></param>
        /// <param name="tangent2"></param>
        /// <param name="amount">Weighting factor</param>
        /// <returns></returns>
        public static float Hermite(float value1, float tangent1, float value2, float tangent2, float amount)
        {
            float t = amount;
            float t2 = t * t;
            float t3 = t * t2;
            float h1 =  2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 =         t3 - 2.0f * t2 + t;
            float h4 =         t3 -        t2;
            return value1 * h1 + value2 * h2 + tangent1 * h3 + tangent2 * h4;
        }

        /// <summary>
        /// Returns the angle wrapped to fit the range -Pi and Pi.
        /// For instance if the input is TwoPi (a full circle), this would
        /// be wrapped to zero, while TwoPi + 0.1 would be wrapped to 0.1.
        /// </summary>
        public static float WrapAngle(float angle)
        {
            angle = (float)Math.IEEERemainder(angle, TwoPi);

            if (angle <= -Pi)
                angle += TwoPi;
            else if (angle > Pi)
                angle -= TwoPi;

            return angle;
        }
    }
}
