//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Globalization;

namespace Microsoft.Xna.Framework
{
    public enum PlaneIntersectionType
    {
        Front,
        Back,
        Intersecting
    }

    public struct Plane : IEquatable<Plane>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public Vector3 Normal;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float D;

        /// <summary>
        /// Create a plane from plane equation.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="c"></param>
        /// <param name="d"></param>
        public Plane(float a, float b, float c, float d)
        {
            Normal.X = a;
            Normal.Y = b;
            Normal.Z = c;
            D = d;
        }

        public Plane(Vector3 normal, float d)
        {
            Normal = normal;
            D = d;
        }

        public Plane(Vector4 value)
        {
            Normal.X = value.X; Normal.Y = value.Y; Normal.Z = value.Z;
            D = value.W;
        }

        /// <summary>
        /// Create a plane from 3 points (Triagnle).
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <param name="point3"></param>
        public Plane(Vector3 point1, Vector3 point2, Vector3 point3)
        {
            float ax = point2.X - point1.X;
            float ay = point2.Y - point1.Y;
            float az = point2.Z - point1.Z;

            float bx = point3.X - point1.X;
            float by = point3.Y - point1.Y;
            float bz = point3.Z - point1.Z;

            // N=Cross(a,b)
            float nx = ay * bz - az * by;
            float ny = az * bx - ax * bz;
            float nz = ax * by - ay * bx;

            // Normalize(N)
            float ls = nx * nx + ny * ny + nz * nz;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            Normal.X = nx * invNorm;
            Normal.Y = ny * invNorm;
            Normal.Z = nz * invNorm;

            // D = - Dot(N, point1)
            D = -(Normal.X * point1.X + Normal.Y * point1.Y + Normal.Z * point1.Z);
        }

        #region Override methods
        // Implement IEquatable<T>
        public bool Equals(Plane other)
        {
            return (Normal.X == other.Normal.X && Normal.Y == other.Normal.Y && Normal.Z == other.Normal.Z && D == other.D);
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is Plane)
            {
                equal = Equals((Plane)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return Normal.GetHashCode() + D.GetHashCode();
        }

        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{Normal:{0} D:{1}}}", Normal.ToString(), D.ToString(ci));
        }
        #endregion

        /// <summary>
        /// Normalize the plane coefficients so that the plane normal has unit length.
        /// </summary>
        public void Normalize()
        {
            const float FLT_EPSILON = 1.192092896e-07f; /* smallest such that 1.0+FLT_EPSILON != 1.0 */

            float f = Normal.X * Normal.X + Normal.Y * Normal.Y + Normal.Z * Normal.Z;
            if (Math.Abs(f - 1.0f) < FLT_EPSILON)
            {
                return; // It already normalized, so we don't need to farther process.
            }

            float fInv = 1.0f / (float)Math.Sqrt(f);
            Normal.X *= fInv;
            Normal.Y *= fInv;
            Normal.Z *= fInv;
            D *= fInv;
        }

        public static Plane Normalize( Plane value )
        {
            const float FLT_EPSILON = 1.192092896e-07f; /* smallest such that 1.0+FLT_EPSILON != 1.0 */

            Plane result;

            float f = value.Normal.X * value.Normal.X + value.Normal.Y * value.Normal.Y + value.Normal.Z * value.Normal.Z;
            if (Math.Abs(f - 1.0f) < FLT_EPSILON)
            {
                result.Normal = value.Normal;
                result.D = value.D;
                return result; // It already normalized, so we don't need to farther process.
            }

            float fInv = 1.0f / (float)Math.Sqrt(f);
            result.Normal.X = value.Normal.X * fInv;
            result.Normal.Y = value.Normal.Y * fInv;
            result.Normal.Z = value.Normal.Z * fInv;
            result.D = value.D * fInv;

            return result;
        }

        public static void Normalize( ref Plane value, out Plane result )
        {
            const float FLT_EPSILON = 1.192092896e-07f; /* smallest such that 1.0+FLT_EPSILON != 1.0 */

            float f = value.Normal.X * value.Normal.X + value.Normal.Y * value.Normal.Y + value.Normal.Z * value.Normal.Z;
            if (Math.Abs(f - 1.0f) < FLT_EPSILON)
            {
                result.Normal = value.Normal;
                result.D = value.D;
                return; // It already normalized, so we don't need to farther process.
            }

            float fInv = 1.0f / (float)Math.Sqrt(f);
            result.Normal.X = value.Normal.X * fInv;
            result.Normal.Y = value.Normal.Y * fInv;
            result.Normal.Z = value.Normal.Z * fInv;
            result.D = value.D * fInv;
        }

        /// <summary>
        /// Transforms a plane by a matrix.
        /// </summary>
        /// <param name="plane">plane that contains vlaue that will be transformed. The Normal of this plane must be normalized before this method is called.</param>
        /// <param name="matrix">transformation matirx.</param>
        /// <returns></returns>
        public static Plane Transform(Plane plane, Matrix matrix)
        {
            Matrix m;
            Matrix.Invert(ref matrix, out m);
            
            Plane result;
            float x = plane.Normal.X, y = plane.Normal.Y, z = plane.Normal.Z, w = plane.D;
            result.Normal.X = x * m.M11 + y * m.M12 + z * m.M13 + w * m.M14;
            result.Normal.Y = x * m.M21 + y * m.M22 + z * m.M23 + w * m.M24;
            result.Normal.Z = x * m.M31 + y * m.M32 + z * m.M33 + w * m.M34;
            result.D        = x * m.M41 + y * m.M42 + z * m.M43 + w * m.M44;
            return result;
        }

        public static void Transform(ref Plane plane, ref Matrix matrix, out Plane result)
        {
            Matrix m;
            Matrix.Invert(ref matrix, out m);

            float x = plane.Normal.X, y = plane.Normal.Y, z = plane.Normal.Z, w = plane.D;
            result.Normal.X = x * m.M11 + y * m.M12 + z * m.M13 + w * m.M14;
            result.Normal.Y = x * m.M21 + y * m.M22 + z * m.M23 + w * m.M24;
            result.Normal.Z = x * m.M31 + y * m.M32 + z * m.M33 + w * m.M34;
            result.D        = x * m.M41 + y * m.M42 + z * m.M43 + w * m.M44;
        }

        /// <summary>
        /// Transforms a plane by a quaternion. This input quaternion is inverse transpose of the actual transformation.
        /// </summary>
        /// <param name="plane">plane that contains vlaue that will be transformed. The Normal of this plane must be normalized before this method is called.</param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        public static Plane Transform(Plane plane, Quaternion rotation)
        {
            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float m11 = 1.0f - yy2 - zz2;
            float m21 = xy2 - wz2;
            float m31 = xz2 + wy2;

            float m12 = xy2 + wz2;
            float m22 = 1.0f - xx2 - zz2;
            float m32 = yz2 - wx2;

            float m13 = xz2 - wy2;
            float m23 = yz2 + wx2;
            float m33 = 1.0f - xx2 - yy2;

            Plane result;
            float x = plane.Normal.X, y = plane.Normal.Y, z = plane.Normal.Z;
            result.Normal.X = x * m11 + y * m21 + z * m31;
            result.Normal.Y = x * m12 + y * m22 + z * m32;
            result.Normal.Z = x * m13 + y * m23 + z * m33;
            result.D = plane.D;

            return result;
        }

        public static void Transform(ref Plane plane, ref Quaternion rotation, out Plane result)
        {
            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float m11 = 1.0f - yy2 - zz2;
            float m21 = xy2 - wz2;
            float m31 = xz2 + wy2;

            float m12 = xy2 + wz2;
            float m22 = 1.0f - xx2 - zz2;
            float m32 = yz2 - wx2;

            float m13 = xz2 - wy2;
            float m23 = yz2 + wx2;
            float m33 = 1.0f - xx2 - yy2;

            float x = plane.Normal.X, y = plane.Normal.Y, z = plane.Normal.Z;
            result.Normal.X = x * m11 + y * m21 + z * m31;
            result.Normal.Y = x * m12 + y * m22 + z * m32;
            result.Normal.Z = x * m13 + y * m23 + z * m33;
            result.D = plane.D;
        }

        public float Dot(Vector4 value)
        {
            return Normal.X * value.X + Normal.Y * value.Y + Normal.Z * value.Z + D * value.W;
        }

        public void Dot(ref Vector4 value, out float result)
        {
            result = Normal.X * value.X + Normal.Y * value.Y + Normal.Z * value.Z + D * value.W;
        }

        public float DotCoordinate(Vector3 value)
        {
            return Normal.X * value.X + Normal.Y * value.Y + Normal.Z * value.Z + D;
        }

        public void DotCoordinate(ref Vector3 value, out float result)
        {
            result = Normal.X * value.X + Normal.Y * value.Y + Normal.Z * value.Z + D;
        }

        public float DotNormal(Vector3 value)
        {
            return Normal.X * value.X + Normal.Y * value.Y + Normal.Z * value.Z;
        }

        public void DotNormal(ref Vector3 value, out float result)
        {
            result = Normal.X * value.X + Normal.Y * value.Y + Normal.Z * value.Z;
        }

        #region Intersections
        /// <summary>
        /// Intersection test against another bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns PlaneIntersectionType that contains value Inside, Outside, or Intersects.</returns>
        public PlaneIntersectionType Intersects(BoundingBox box)
        {
            // This algorithm came from Real-Time Rendering Second Edition p.587
            Vector3 vMin;
            vMin.X = (Normal.X >= 0.0f) ? box.Min.X : box.Max.X;
            vMin.Y = (Normal.Y >= 0.0f) ? box.Min.Y : box.Max.Y;
            vMin.Z = (Normal.Z >= 0.0f) ? box.Min.Z : box.Max.Z;

            Vector3 vMax;
            vMax.X = (Normal.X >= 0.0f) ? box.Max.X : box.Min.X;
            vMax.Y = (Normal.Y >= 0.0f) ? box.Max.Y : box.Min.Y;
            vMax.Z = (Normal.Z >= 0.0f) ? box.Max.Z : box.Min.Z;

            float dot = Normal.X * vMin.X + Normal.Y * vMin.Y + Normal.Z * vMin.Z;
            if (dot + D > 0.0f) return PlaneIntersectionType.Front;

            dot = Normal.X * vMax.X + Normal.Y * vMax.Y + Normal.Z * vMax.Z;
            if ( dot + D < 0.0f) return PlaneIntersectionType.Back;
            return PlaneIntersectionType.Intersecting;
        }

        public void Intersects(ref BoundingBox box, out PlaneIntersectionType result)
        {
            // This algorithm came from Real-Time Rendering Second Edition p.587
            Vector3 vMin;
            vMin.X = (Normal.X >= 0.0f) ? box.Min.X : box.Max.X;
            vMin.Y = (Normal.Y >= 0.0f) ? box.Min.Y : box.Max.Y;
            vMin.Z = (Normal.Z >= 0.0f) ? box.Min.Z : box.Max.Z;

            Vector3 vMax;
            vMax.X = (Normal.X >= 0.0f) ? box.Max.X : box.Min.X;
            vMax.Y = (Normal.Y >= 0.0f) ? box.Max.Y : box.Min.Y;
            vMax.Z = (Normal.Z >= 0.0f) ? box.Max.Z : box.Min.Z;

            float dot = Normal.X * vMin.X + Normal.Y * vMin.Y + Normal.Z * vMin.Z;
            if (dot + D > 0.0f)
            {
                result = PlaneIntersectionType.Front;
                return;
            }

            dot = Normal.X * vMax.X + Normal.Y * vMax.Y + Normal.Z * vMax.Z;
            if (dot + D < 0.0f)
            {
                result = PlaneIntersectionType.Back;
                return;
            }

            result = PlaneIntersectionType.Intersecting;
        }


        /// <summary>
        /// Intersection test against another bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns PlaneIntersectionType that contains value Inside, Outside, or Intersects.</returns>
        public PlaneIntersectionType Intersects(BoundingFrustum frustum)
        {
            if (null == frustum)
            {
                throw new ArgumentNullException("frustum", "FrameworkResources.NullNotAllowed");
            }

            return frustum.Intersects(this);
        }

#if Unused
        /// <summary>
        /// Intersection test against another plane.
        /// </summary>
        /// <param name="plane"></param>
        /// <returns>Returns PlaneIntersectionType that contains value Inside, Outside, or Intersects.</returns>
        public PlaneIntersectionType Intersects(Plane plane)
        {
            // This algorithm came from Real-Time Collision Detection pp.209.

            // Compute direction of intersection line.
            Vector3 d = Vector3.Cross(Normal, plane.Normal);

            // if d is zero, the planes are parallel (and separated)
            // or coincident, so they're not cosidered intersections.
            const float epsilon = 1e-5f;

            if ( d.LengthSquared() < epsilon )
            {
                // Check which side plane exist.
                throw new NotImplementedException();
            }

            // Compute line origin.
            float d11 = Normal.LengthSquared();
            float d12 = Vector3.Dot( Normal, plane.Normal );
            float d22 = plane.Normal.LengthSquared();

            float denom = d11 * d22 - d12 * d12;
            float k1 = (D * d22 - plane.D * d12) / denom;
            float k2 = (plane.D * d11 - D * d12) / denom;

            Vector3 p = Normal * k1 + plane.Normal * k2;

            // p: ray origin, d:ray direction.

            throw new NotImplementedException();
        }

        /// <summary>
        /// Intersection test against another sphere.
        /// </summary>
        /// <param name="plane"></param>
        /// <returns>Returns RayIntersectionResult.</returns>
        public RayIntersectionResult Intersects(Ray ray)
        {
            float nd = Vector3.Dot(Normal, ray.Direction);
            const float epsilon = 1e-5f;

            // Ray never intersected.
            if ( Math.Abs(nd) < epsilon)
            {
                return RayIntersectionResult.Disjoint;
            }

            return new RayIntersectionResult(true, -D - Vector3.Dot(Normal, ray.Position) / nd );
        }
#endif

        /// <summary>
        /// Intersection test against another sphere.
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns>Returns PlaneIntersectionType that contains value Inside, Outside, or Intersects.</returns>
        public PlaneIntersectionType Intersects(BoundingSphere sphere)
        {
            float dot = sphere.Center.X * Normal.X + sphere.Center.Y * Normal.Y + sphere.Center.Z * Normal.Z;
            float dist = dot + D;
            if (dist > sphere.Radius) return PlaneIntersectionType.Front;
            if (dist < -sphere.Radius) return PlaneIntersectionType.Back;
            return PlaneIntersectionType.Intersecting;
        }

        public void Intersects(ref BoundingSphere sphere, out PlaneIntersectionType result)
        {
            float dot = sphere.Center.X * Normal.X + sphere.Center.Y * Normal.Y + sphere.Center.Z * Normal.Z;
            float dist = dot + D;
            if (dist > sphere.Radius)
            {
                result = PlaneIntersectionType.Front;
                return;
            }

            if (dist < -sphere.Radius)
            {
                result = PlaneIntersectionType.Back;
                return;
            }

            result = PlaneIntersectionType.Intersecting;
        }
        #endregion

        #region Operator overloads
        public static bool operator ==(Plane lhs, Plane rhs)
        {
            return lhs.Equals(rhs);
        }

        public static bool operator !=(Plane lhs, Plane rhs)
        {
            return (lhs.Normal.X != rhs.Normal.X || lhs.Normal.Y != rhs.Normal.Y || lhs.Normal.Z != rhs.Normal.Z || lhs.D != rhs.D);
        }
        #endregion

    }
}
