//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Globalization;

using SM = System.Math;

namespace Microsoft.Xna.Framework
{
    public struct Quaternion : IEquatable<Quaternion>
    {
        // XYZ is Vector part and W is scaler part of Quaternion.
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float X;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Y;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Z;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float W;

        #region Static Properties
        private static Quaternion _identity = new Quaternion( 0, 0, 0, 1 );

        /// <summary>
        /// Returns Identity Quaternion(0, 0, 0, 1)
        /// </summary>
        public static Quaternion Identity { get { return _identity; } }
        #endregion

        #region Constructors
        public Quaternion(float x, float y, float z, float w)
        {
            X = x; Y = y; Z = z; W = w;
        }

        public Quaternion(Vector3 vectorPart, float scalarPart)
        {
            X = vectorPart.X; Y = vectorPart.Y; Z = vectorPart.Z; W = scalarPart;
        }
        #endregion

        #region Override methods
        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format( ci, "{{X:{0} Y:{1} Z:{2} W:{3}}}", X.ToString(ci), Y.ToString(ci), Z.ToString(ci), W.ToString(ci));
        }

        // Implement IEquatable<T>
        public bool Equals(Quaternion other)
        {
            return (X == other.X && Y == other.Y && Z == other.Z && W == other.W);
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is Quaternion)
            {
                equal = Equals((Quaternion)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return X.GetHashCode() + Y.GetHashCode() + Z.GetHashCode() + W.GetHashCode();
        }
        #endregion

        #region Operations
        //-------------------------------------
        // Unary Operators
        //=====================================
        /// <summary>
        /// Returns the square of the length of a quaternion.
        /// </summary>
        /// <returns></returns>
        public float LengthSquared()
        {
            return X * X + Y * Y + Z * Z + W * W;
        }

        /// <summary>
        /// Returns the length of a quaternion.
        /// </summary>
        /// <returns></returns>
        public float Length()
        {
            float ls = X * X + Y * Y + Z * Z + W * W;
            return (float)System.Math.Sqrt((double)ls);
        }

        /// <summary>
        /// Computes a unit length quaternion.
        /// </summary>
        /// <returns></returns>
        public void Normalize()
        {
            float ls = X * X + Y * Y + Z * Z + W * W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);
            X *= invNorm;
            Y *= invNorm;
            Z *= invNorm;
            W *= invNorm;
        }

        /// <summary>
        /// Computes a unit length quaternion.
        /// </summary>
        /// <param name="quaternion"></param>
        /// <returns></returns>
        public static Quaternion Normalize( Quaternion quaternion )
        {
            Quaternion ans;

            float ls = quaternion.X * quaternion.X + quaternion.Y * quaternion.Y + quaternion.Z * quaternion.Z + quaternion.W * quaternion.W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            ans.X = quaternion.X * invNorm;
            ans.Y = quaternion.Y * invNorm;
            ans.Z = quaternion.Z * invNorm;
            ans.W = quaternion.W * invNorm;

            return ans;
        }

        public static void Normalize(ref Quaternion quaternion, out Quaternion result)
        {
            float ls = quaternion.X * quaternion.X + quaternion.Y * quaternion.Y + quaternion.Z * quaternion.Z + quaternion.W * quaternion.W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            result.X = quaternion.X * invNorm;
            result.Y = quaternion.Y * invNorm;
            result.Z = quaternion.Z * invNorm;
            result.W = quaternion.W * invNorm;
        }

        /// <summary>
        /// Returns Conjugate of given quaternion.
        /// </summary>
        /// <returns></returns>
        public void Conjugate()
        {
            X = -X;
            Y = -Y;
            Z = -Z;
        }

        public static Quaternion Conjugate(Quaternion value)
        {
            Quaternion ans;
            ans.X = -value.X;
            ans.Y = -value.Y;
            ans.Z = -value.Z;
            ans.W = value.W;

            return ans;
        }

        public static void Conjugate(ref Quaternion value, out Quaternion result)
        {
            result.X = -value.X;
            result.Y = -value.Y;
            result.Z = -value.Z;
            result.W = value.W;
        }


        /// <summary>
        /// Returns inverse of given quaternion.
        /// </summary>
        /// <param name="quaternion"></param>
        /// <returns></returns>
        public static Quaternion Inverse(Quaternion quaternion)
        {
            //  -1   (       a              -v       )
            // q   = ( -------------   ------------- )
            //       (  a^2 + |v|^2  ,  a^2 + |v|^2  )

            Quaternion ans;
            float ls = quaternion.X * quaternion.X + quaternion.Y * quaternion.Y + quaternion.Z * quaternion.Z + quaternion.W * quaternion.W;
            float invNorm = 1.0f / ls;

            ans.X = -quaternion.X * invNorm;
            ans.Y = -quaternion.Y * invNorm;
            ans.Z = -quaternion.Z * invNorm;
            ans.W = quaternion.W * invNorm;

            return ans;
        }

        public static void Inverse(ref Quaternion quaternion, out Quaternion result)
        {
            float ls = quaternion.X * quaternion.X + quaternion.Y * quaternion.Y + quaternion.Z * quaternion.Z + quaternion.W * quaternion.W;
            float invNorm = 1.0f / ls;

            result.X = -quaternion.X * invNorm;
            result.Y = -quaternion.Y * invNorm;
            result.Z = -quaternion.Z * invNorm;
            result.W = quaternion.W * invNorm;
        }


        /// <summary>
        /// Rotates a quaternion about an arbitrary axis.
        /// </summary>
        /// <param name="axis"></param>
        /// <param name="angle"></param>
        /// <returns></returns>
        public static Quaternion CreateFromAxisAngle( Vector3 axis, float angle )
        {
            Quaternion ans;

            float halfAngle = angle * 0.5f;
            float s = (float)SM.Sin( halfAngle );
            float c = (float)SM.Cos(halfAngle);

            ans.X = axis.X * s;
            ans.Y = axis.Y * s;
            ans.Z = axis.Z * s;
            ans.W = c;

            return ans;
        }

        public static void CreateFromAxisAngle(ref Vector3 axis, float angle, out Quaternion result)
        {
            float halfAngle = angle * 0.5f;
            float s = (float)SM.Sin(halfAngle);
            float c = (float)SM.Cos(halfAngle);

            result.X = axis.X * s;
            result.Y = axis.Y * s;
            result.Z = axis.Z * s;
            result.W = c;
        }

        /// <summary>
        /// Creates a quaternion with a specified yaw, pitch, and roll.
        /// </summary>
        /// <param name="yaw"></param>
        /// <param name="pitch"></param>
        /// <param name="roll"></param>
        /// <returns></returns>
        public static Quaternion CreateFromYawPitchRoll(float yaw, float pitch, float roll)
        {
            //  Roll first, about axis the object is facing, then
            //  pitch upward, then yaw to face into the new heading
            float sr, cr, sp, cp, sy, cy;

            float halfRoll = roll * 0.5f;
            sr = (float)SM.Sin(halfRoll); cr = (float)SM.Cos(halfRoll);

            float halfPitch = pitch * 0.5f;
            sp = (float)SM.Sin(halfPitch); cp = (float)SM.Cos(halfPitch);

            float halfYaw = yaw * 0.5f;
            sy = (float)SM.Sin(halfYaw); cy = (float)SM.Cos(halfYaw);

            Quaternion result;
            result.X = cy * sp * cr + sy * cp * sr;
            result.Y = sy * cp * cr - cy * sp * sr;
            result.Z = cy * cp * sr - sy * sp * cr;
            result.W = cy * cp * cr + sy * sp * sr;

            return result;
        }

        public static void CreateFromYawPitchRoll(float yaw, float pitch, float roll, out Quaternion result)
        {
            //  Roll first, about axis the object is facing, then
            //  pitch upward, then yaw to face into the new heading
            float sr, cr, sp, cp, sy, cy;

            float halfRoll = roll * 0.5f;
            sr = (float)SM.Sin(halfRoll); cr = (float)SM.Cos(halfRoll);

            float halfPitch = pitch * 0.5f;
            sp = (float)SM.Sin(halfPitch); cp = (float)SM.Cos(halfPitch);

            float halfYaw = yaw * 0.5f;
            sy = (float)SM.Sin(halfYaw); cy = (float)SM.Cos(halfYaw);

            result.X = cy * sp * cr + sy * cp * sr;
            result.Y = sy * cp * cr - cy * sp * sr;
            result.Z = cy * cp * sr - sy * sp * cr;
            result.W = cy * cp * cr + sy * sp * sr;
        }

        /// <summary>
        /// Create quaternion from rotation matrix.
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Quaternion CreateFromRotationMatrix(Matrix matrix)
        {
            float trace = matrix.M11 + matrix.M22 + matrix.M33;
            Quaternion q = new Quaternion();
            if ( trace > 0.0f )
            {
                float s = (float)SM.Sqrt( trace + 1.0f );
                q.W = s * 0.5f;
                s = 0.5f / s;
                q.X = (matrix.M23 - matrix.M32) * s;
                q.Y = (matrix.M31 - matrix.M13) * s;
                q.Z = (matrix.M12 - matrix.M21) * s;
            }
            else
            {
                if (matrix.M11 >= matrix.M22 && matrix.M11 >= matrix.M33)
                {
                    float s = (float)SM.Sqrt(1.0f + matrix.M11 - matrix.M22 - matrix.M33);
                    float invS = 0.5f / s;
                    q.X = 0.5f * s;
                    q.Y = (matrix.M12 + matrix.M21) * invS;
                    q.Z = (matrix.M13 + matrix.M31) * invS;
                    q.W = (matrix.M23 - matrix.M32) * invS;
                }
                else if (matrix.M22 > matrix.M33)
                {
                    float s = (float)SM.Sqrt(1.0f + matrix.M22 - matrix.M11 - matrix.M33);
                    float invS = 0.5f / s;
                    q.X = (matrix.M21 + matrix.M12) * invS;
                    q.Y = 0.5f * s;
                    q.Z = (matrix.M32 + matrix.M23) * invS;
                    q.W = (matrix.M31 - matrix.M13) * invS;
                }
                else
                {
                    float s = (float)SM.Sqrt(1.0f + matrix.M33 - matrix.M11 - matrix.M22);
                    float invS = 0.5f / s;
                    q.X = (matrix.M31 + matrix.M13) * invS;
                    q.Y = (matrix.M32 + matrix.M23) * invS;
                    q.Z = 0.5f * s;
                    q.W = (matrix.M12 - matrix.M21) * invS;
                }
            }

            return q;
        }

        public static void CreateFromRotationMatrix(ref Matrix matrix, out Quaternion result)
        {
            float trace = matrix.M11 + matrix.M22 + matrix.M33;
            if (trace > 0.0f)
            {
                float s = (float)SM.Sqrt(trace + 1.0f);
                result.W = s * 0.5f;
                s = 0.5f / s;
                result.X = (matrix.M23 - matrix.M32) * s;
                result.Y = (matrix.M31 - matrix.M13) * s;
                result.Z = (matrix.M12 - matrix.M21) * s;
            }
            else
            {
                if (matrix.M11 >= matrix.M22 && matrix.M11 >= matrix.M33)
                {
                    float s = (float)SM.Sqrt(1.0f + matrix.M11 - matrix.M22 - matrix.M33);
                    float invS = 0.5f / s;
                    result.X = 0.5f * s;
                    result.Y = (matrix.M12 + matrix.M21) * invS;
                    result.Z = (matrix.M13 + matrix.M31) * invS;
                    result.W = (matrix.M23 - matrix.M32) * invS;
                }
                else if (matrix.M22 > matrix.M33)
                {
                    float s = (float)SM.Sqrt(1.0f + matrix.M22 - matrix.M11 - matrix.M33);
                    float invS = 0.5f / s;
                    result.X = (matrix.M21 + matrix.M12) * invS;
                    result.Y = 0.5f * s;
                    result.Z = (matrix.M32 + matrix.M23) * invS;
                    result.W = (matrix.M31 - matrix.M13) * invS;
                }
                else
                {
                    float s = (float)SM.Sqrt(1.0f + matrix.M33 - matrix.M11 - matrix.M22);
                    float invS = 0.5f / s;
                    result.X = (matrix.M31 + matrix.M13) * invS;
                    result.Y = (matrix.M32 + matrix.M23) * invS;
                    result.Z = 0.5f * s;
                    result.W = (matrix.M12 - matrix.M21) * invS;
                }
            }
        }

        //-------------------------------------
        // Binary Operators
        //=====================================
        /// <summary>
        /// Returns the dot product of two quaternions.
        /// </summary>
        /// <param name="quaternion1"></param>
        /// <param name="quaternion2"></param>
        /// <returns></returns>
        public static float Dot(Quaternion quaternion1, Quaternion quaternion2)
        {
            return quaternion1.X * quaternion2.X + quaternion1.Y * quaternion2.Y + quaternion1.Z * quaternion2.Z + quaternion1.W * quaternion2.W;
        }

        public static void Dot(ref Quaternion quaternion1, ref Quaternion quaternion2, out float result)
        {
            result = quaternion1.X * quaternion2.X + quaternion1.Y * quaternion2.Y + quaternion1.Z * quaternion2.Z + quaternion1.W * quaternion2.W;
        }

        /// <summary>
        /// Interpolates btween two quaternions, using spherical linear interpolation.
        /// </summary>
        /// <param name="quaternion1"></param>
        /// <param name="quaternion2"></param>
        /// <param name="amount">Paramter that indicates how far to interpolate between the quaternions.</param>
        /// <returns></returns>
        public static Quaternion Slerp(Quaternion quaternion1, Quaternion quaternion2, float amount)
        {
            const float epsilon = 1e-6f;

            float t = amount;
            float cosOmega = quaternion1.X * quaternion2.X + quaternion1.Y * quaternion2.Y +
                                quaternion1.Z * quaternion2.Z + quaternion1.W * quaternion2.W;

            bool flip = false;
            if (cosOmega < 0.0f)
            {
                flip = true;
                cosOmega = -cosOmega;
            }

            float s1, s2;
            if (cosOmega > (1.0f - epsilon))
            {
                // Too close, do straight linear interpolation.
                s1 = 1.0f - t;
                s2 = (flip) ? -t : t;
            }
            else
            {
                float omega = (float)SM.Acos(cosOmega);
                float invSinOmega = (float)(1 / SM.Sin(omega));

                s1 = (float)SM.Sin((1.0f - t) * omega) * invSinOmega;
                s2 = (flip)
                    ? (float)-SM.Sin(t * omega) * invSinOmega
                    : (float)SM.Sin(t * omega) * invSinOmega;
            }

            Quaternion ans;

            ans.X = s1 * quaternion1.X + s2 * quaternion2.X;
            ans.Y = s1 * quaternion1.Y + s2 * quaternion2.Y;
            ans.Z = s1 * quaternion1.Z + s2 * quaternion2.Z;
            ans.W = s1 * quaternion1.W + s2 * quaternion2.W;

            return ans;
        }

        public static void Slerp(ref Quaternion quaternion1, ref Quaternion quaternion2, float amount, out Quaternion result)
        {
            const float epsilon = 1e-6f;

            float t = amount;
            float cosOmega = quaternion1.X * quaternion2.X + quaternion1.Y * quaternion2.Y +
                                quaternion1.Z * quaternion2.Z + quaternion1.W * quaternion2.W;

            bool flip = false;
            if (cosOmega < 0.0f)
            {
                flip = true;
                cosOmega = -cosOmega;
            }

            float s1, s2;
            if (cosOmega > (1.0f - epsilon))
            {
                // Too close, do straight linear interpolation.
                s1 = 1.0f - t;
                s2 = (flip) ? -t : t;
            }
            else
            {
                float omega = (float)SM.Acos(cosOmega);
                float invSinOmega = (float)(1 / SM.Sin(omega));

                s1 = (float)SM.Sin((1.0f - t) * omega) * invSinOmega;
                s2 = (flip)
                    ? (float)-SM.Sin(t * omega) * invSinOmega
                    : (float)SM.Sin(t * omega) * invSinOmega;
            }

            result.X = s1 * quaternion1.X + s2 * quaternion2.X;
            result.Y = s1 * quaternion1.Y + s2 * quaternion2.Y;
            result.Z = s1 * quaternion1.Z + s2 * quaternion2.Z;
            result.W = s1 * quaternion1.W + s2 * quaternion2.W;
        }


        /// <summary>
        /// Linear Interpolation betwee two quaternions.
        /// This works well when two quaternions are less than 90 degrees apart.
        /// </summary>
        /// <param name="quaternion1"></param>
        /// <param name="quaternion2"></param>
        /// <param name="amount">Paramter that indicates how far to interpolate between the quaternions.</param>
        /// <returns></returns>
        public static Quaternion Lerp(Quaternion quaternion1, Quaternion quaternion2, float amount)
        {
            float t = amount;
            float t1 = 1.0f - t;

            Quaternion r = new Quaternion();
            float dot = quaternion1.X * quaternion2.X + quaternion1.Y * quaternion2.Y +
                                quaternion1.Z * quaternion2.Z + quaternion1.W * quaternion2.W;

            if (dot >= 0.0f)
            {
                r.X = t1 * quaternion1.X + t * quaternion2.X;
                r.Y = t1 * quaternion1.Y + t * quaternion2.Y;
                r.Z = t1 * quaternion1.Z + t * quaternion2.Z;
                r.W = t1 * quaternion1.W + t * quaternion2.W;
            }
            else
            {
                r.X = t1 * quaternion1.X - t * quaternion2.X;
                r.Y = t1 * quaternion1.Y - t * quaternion2.Y;
                r.Z = t1 * quaternion1.Z - t * quaternion2.Z;
                r.W = t1 * quaternion1.W - t * quaternion2.W;
            }

            // Normalize it.
            float ls = r.X * r.X + r.Y * r.Y + r.Z * r.Z + r.W * r.W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);
            r.X *= invNorm;
            r.Y *= invNorm;
            r.Z *= invNorm;
            r.W *= invNorm;

            return r;
        }

        public static void Lerp(ref Quaternion quaternion1, ref Quaternion quaternion2, float amount, out Quaternion result)
        {
            float t = amount;
            float t1 = 1.0f - t;

            float dot = quaternion1.X * quaternion2.X + quaternion1.Y * quaternion2.Y +
                                quaternion1.Z * quaternion2.Z + quaternion1.W * quaternion2.W;

            if (dot >= 0.0f)
            {
                result.X = t1 * quaternion1.X + t * quaternion2.X;
                result.Y = t1 * quaternion1.Y + t * quaternion2.Y;
                result.Z = t1 * quaternion1.Z + t * quaternion2.Z;
                result.W = t1 * quaternion1.W + t * quaternion2.W;
            }
            else
            {
                result.X = t1 * quaternion1.X - t * quaternion2.X;
                result.Y = t1 * quaternion1.Y - t * quaternion2.Y;
                result.Z = t1 * quaternion1.Z - t * quaternion2.Z;
                result.W = t1 * quaternion1.W - t * quaternion2.W;
            }

            // Normalize it.
            float ls = result.X * result.X + result.Y * result.Y + result.Z * result.Z + result.W * result.W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);
            result.X *= invNorm;
            result.Y *= invNorm;
            result.Z *= invNorm;
            result.W *= invNorm;
        }

        /// <summary>
        /// Concatenate two quaternions.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <returns>The result represents the rotation value1 followed by the rotation value2.</returns>
        public static Quaternion Concatenate(Quaternion value1, Quaternion value2)
        {
            /*
             * Original code.
                        return new Quaternion(
                            av * quaternion2.W + bv * quaternion1.W + Vector3.Cross(av, bv),  // Vector part
                            quaternion1.W * quaternion2.W - Vector3.Dot(av, bv)               // Scalr part
                        );
             */
            Quaternion ans;

            // Cocatinate rotation is actually q2 * q1 instead of q1 * q2.
            // So that's why value2 goes q1 and value1 goes q2.
            float q1x = value2.X;
            float q1y = value2.Y;
            float q1z = value2.Z;
            float q1w = value2.W;

            float q2x = value1.X;
            float q2y = value1.Y;
            float q2z = value1.Z;
            float q2w = value1.W;

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            ans.X = q1x * q2w + q2x * q1w + cx;
            ans.Y = q1y * q2w + q2y * q1w + cy;
            ans.Z = q1z * q2w + q2z * q1w + cz;
            ans.W = q1w * q2w - dot;

            return ans;
        }

        public static void Concatenate(ref Quaternion value1, ref Quaternion value2, out Quaternion result)
        {
            /*
             * Original code.
                        return new Quaternion(
                            av * quaternion2.W + bv * quaternion1.W + Vector3.Cross(av, bv),  // Vector part
                            quaternion1.W * quaternion2.W - Vector3.Dot(av, bv)               // Scalr part
                        );
             */
            // Cocatinate rotation is actually q2 * q1 instead of q1 * q2.
            // So that's why value2 goes q1 and value1 goes q2.
            float q1x = value2.X;
            float q1y = value2.Y;
            float q1z = value2.Z;
            float q1w = value2.W;

            float q2x = value1.X;
            float q2y = value1.Y;
            float q2z = value1.Z;
            float q2w = value1.W;

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            result.X = q1x * q2w + q2x * q1w + cx;
            result.Y = q1y * q2w + q2y * q1w + cy;
            result.Z = q1z * q2w + q2z * q1w + cz;
            result.W = q1w * q2w - dot;
        }
        #endregion

        #region Operator overloads equivalent methods
        public static Quaternion Negate(Quaternion quaternion)
        {
            Quaternion ans;

            ans.X = -quaternion.X;
            ans.Y = -quaternion.Y;
            ans.Z = -quaternion.Z;
            ans.W = -quaternion.W;

            return ans;
        }

        public static void Negate(ref Quaternion quaternion, out Quaternion result)
        {
            result.X = -quaternion.X;
            result.Y = -quaternion.Y;
            result.Z = -quaternion.Z;
            result.W = -quaternion.W;
        }


        public static Quaternion Add(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            ans.X = quaternion1.X + quaternion2.X;
            ans.Y = quaternion1.Y + quaternion2.Y;
            ans.Z = quaternion1.Z + quaternion2.Z;
            ans.W = quaternion1.W + quaternion2.W;

            return ans;
        }

        public static void Add(ref Quaternion quaternion1, ref Quaternion quaternion2, out Quaternion result)
        {
            result.X = quaternion1.X + quaternion2.X;
            result.Y = quaternion1.Y + quaternion2.Y;
            result.Z = quaternion1.Z + quaternion2.Z;
            result.W = quaternion1.W + quaternion2.W;
        }


        public static Quaternion Subtract(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            ans.X = quaternion1.X - quaternion2.X;
            ans.Y = quaternion1.Y - quaternion2.Y;
            ans.Z = quaternion1.Z - quaternion2.Z;
            ans.W = quaternion1.W - quaternion2.W;

            return ans;
        }

        public static void Subtract(ref Quaternion quaternion1, ref Quaternion quaternion2, out Quaternion result)
        {
            result.X = quaternion1.X - quaternion2.X;
            result.Y = quaternion1.Y - quaternion2.Y;
            result.Z = quaternion1.Z - quaternion2.Z;
            result.W = quaternion1.W - quaternion2.W;
        }


        public static Quaternion Multiply(Quaternion quaternion1, Quaternion quaternion2)
        {
/*
 * Original code.
            return new Quaternion(
                av * quaternion2.W + bv * quaternion1.W + Vector3.Cross(av, bv),  // Vector part
                quaternion1.W * quaternion2.W - Vector3.Dot(av, bv)               // Scalr part
            );
 */ 
            Quaternion ans;

            float q1x = quaternion1.X;
            float q1y = quaternion1.Y;
            float q1z = quaternion1.Z;
            float q1w = quaternion1.W;

            float q2x = quaternion2.X;
            float q2y = quaternion2.Y;
            float q2z = quaternion2.Z;
            float q2w = quaternion2.W;

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            ans.X = q1x * q2w + q2x * q1w + cx;
            ans.Y = q1y * q2w + q2y * q1w + cy;
            ans.Z = q1z * q2w + q2z * q1w + cz;
            ans.W = q1w * q2w - dot;

            return ans;
        }

        public static void Multiply(ref Quaternion quaternion1, ref Quaternion quaternion2, out Quaternion result)
        {
            /*
             * Original code.
                        return new Quaternion(
                            av * quaternion2.W + bv * quaternion1.W + Vector3.Cross(av, bv),  // Vector part
                            quaternion1.W * quaternion2.W - Vector3.Dot(av, bv)               // Scalr part
                        );
             */
            float q1x = quaternion1.X;
            float q1y = quaternion1.Y;
            float q1z = quaternion1.Z;
            float q1w = quaternion1.W;

            float q2x = quaternion2.X;
            float q2y = quaternion2.Y;
            float q2z = quaternion2.Z;
            float q2w = quaternion2.W;

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            result.X = q1x * q2w + q2x * q1w + cx;
            result.Y = q1y * q2w + q2y * q1w + cy;
            result.Z = q1z * q2w + q2z * q1w + cz;
            result.W = q1w * q2w - dot;
        }


        public static Quaternion Multiply(Quaternion quaternion1, float scaleFactor)
        {
            Quaternion ans;

            ans.X = quaternion1.X * scaleFactor;
            ans.Y = quaternion1.Y * scaleFactor;
            ans.Z = quaternion1.Z * scaleFactor;
            ans.W = quaternion1.W * scaleFactor;

            return ans;
        }

        public static void Multiply(ref Quaternion quaternion1, float scaleFactor, out Quaternion result)
        {
            result.X = quaternion1.X * scaleFactor;
            result.Y = quaternion1.Y * scaleFactor;
            result.Z = quaternion1.Z * scaleFactor;
            result.W = quaternion1.W * scaleFactor;
        }


        public static Quaternion Divide(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            float q1x = quaternion1.X;
            float q1y = quaternion1.Y;
            float q1z = quaternion1.Z;
            float q1w = quaternion1.W;

            //-------------------------------------
            // Inverse part.
            float ls = quaternion2.X * quaternion2.X + quaternion2.Y * quaternion2.Y +
                        quaternion2.Z * quaternion2.Z + quaternion2.W * quaternion2.W;
            float invNorm = 1.0f / ls;

            float q2x = -quaternion2.X * invNorm;
            float q2y = -quaternion2.Y * invNorm;
            float q2z = -quaternion2.Z * invNorm;
            float q2w = quaternion2.W * invNorm;

            //-------------------------------------
            // Multiply part.

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            ans.X = q1x * q2w + q2x * q1w + cx;
            ans.Y = q1y * q2w + q2y * q1w + cy;
            ans.Z = q1z * q2w + q2z * q1w + cz;
            ans.W = q1w * q2w - dot;

            return ans;
        }

        public static void Divide(ref Quaternion quaternion1, ref Quaternion quaternion2, out Quaternion result)
        {
            float q1x = quaternion1.X;
            float q1y = quaternion1.Y;
            float q1z = quaternion1.Z;
            float q1w = quaternion1.W;

            //-------------------------------------
            // Inverse part.
            float ls = quaternion2.X * quaternion2.X + quaternion2.Y * quaternion2.Y +
                        quaternion2.Z * quaternion2.Z + quaternion2.W * quaternion2.W;
            float invNorm = 1.0f / ls;

            float q2x = -quaternion2.X * invNorm;
            float q2y = -quaternion2.Y * invNorm;
            float q2z = -quaternion2.Z * invNorm;
            float q2w = quaternion2.W * invNorm;

            //-------------------------------------
            // Multiply part.

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            result.X = q1x * q2w + q2x * q1w + cx;
            result.Y = q1y * q2w + q2y * q1w + cy;
            result.Z = q1z * q2w + q2z * q1w + cz;
            result.W = q1w * q2w - dot;
        }
        #endregion

        #region Operator overloads
        //-------------------------------------
        // Unary operator overloads
        //=====================================
        public static Quaternion operator -(Quaternion quaternion)
        {
            Quaternion ans;

            ans.X = -quaternion.X;
            ans.Y = -quaternion.Y;
            ans.Z = -quaternion.Z;
            ans.W = -quaternion.W;

            return ans;
        }

        //-------------------------------------
        // Binary operator overloads
        //=====================================
        public static bool operator ==(Quaternion quaternion1, Quaternion quaternion2)
        {
            return (quaternion1.X == quaternion2.X && quaternion1.Y == quaternion2.Y && quaternion1.Z == quaternion2.Z && quaternion1.W == quaternion2.W);
        }

        public static bool operator !=(Quaternion quaternion1, Quaternion quaternion2)
        {
            return (quaternion1.X != quaternion2.X || quaternion1.Y != quaternion2.Y || quaternion1.Z != quaternion2.Z || quaternion1.W != quaternion2.W);
        }

        public static Quaternion operator +(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            ans.X = quaternion1.X + quaternion2.X;
            ans.Y = quaternion1.Y + quaternion2.Y;
            ans.Z = quaternion1.Z + quaternion2.Z;
            ans.W = quaternion1.W + quaternion2.W;

            return ans;
        }

        public static Quaternion operator -(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            ans.X = quaternion1.X - quaternion2.X;
            ans.Y = quaternion1.Y - quaternion2.Y;
            ans.Z = quaternion1.Z - quaternion2.Z;
            ans.W = quaternion1.W - quaternion2.W;

            return ans;
        }

        /// <summary>
        /// Multiplies two quaternions.
        /// </summary>
        /// <param name="quaternion1"></param>
        /// <param name="quaternion2"></param>
        /// <returns></returns>
        public static Quaternion operator *(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            float q1x = quaternion1.X;
            float q1y = quaternion1.Y;
            float q1z = quaternion1.Z;
            float q1w = quaternion1.W;

            float q2x = quaternion2.X;
            float q2y = quaternion2.Y;
            float q2z = quaternion2.Z;
            float q2w = quaternion2.W;

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            ans.X = q1x * q2w + q2x * q1w + cx;
            ans.Y = q1y * q2w + q2y * q1w + cy;
            ans.Z = q1z * q2w + q2z * q1w + cz;
            ans.W = q1w * q2w - dot;

            return ans;
        }

        public static Quaternion operator *(Quaternion quaternion1, float scaleFactor)
        {
            Quaternion ans;

            ans.X = quaternion1.X * scaleFactor;
            ans.Y = quaternion1.Y * scaleFactor;
            ans.Z = quaternion1.Z * scaleFactor;
            ans.W = quaternion1.W * scaleFactor;

            return ans;
        }

        public static Quaternion operator /(Quaternion quaternion1, Quaternion quaternion2)
        {
            Quaternion ans;

            float q1x = quaternion1.X;
            float q1y = quaternion1.Y;
            float q1z = quaternion1.Z;
            float q1w = quaternion1.W;

            //-------------------------------------
            // Inverse part.
            float ls = quaternion2.X * quaternion2.X + quaternion2.Y * quaternion2.Y +
                        quaternion2.Z * quaternion2.Z + quaternion2.W * quaternion2.W;
            float invNorm = 1.0f / ls;

            float q2x = -quaternion2.X * invNorm;
            float q2y = -quaternion2.Y * invNorm;
            float q2z = -quaternion2.Z * invNorm;
            float q2w = quaternion2.W * invNorm;

            //-------------------------------------
            // Multiply part.

            // cross(av, bv)
            float cx = q1y * q2z - q1z * q2y;
            float cy = q1z * q2x - q1x * q2z;
            float cz = q1x * q2y - q1y * q2x;

            float dot = q1x * q2x + q1y * q2y + q1z * q2z;

            ans.X = q1x * q2w + q2x * q1w + cx;
            ans.Y = q1y * q2w + q2y * q1w + cy;
            ans.Z = q1z * q2w + q2z * q1w + cz;
            ans.W = q1w * q2w - dot;

            return ans;
        }

        #endregion

    }
}
