//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Globalization;

namespace Microsoft.Xna.Framework
{

    public struct Ray : IEquatable<Ray>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public Vector3 Position;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public Vector3 Direction;

        /// <summary>
        /// Create ray from position and direction.
        /// </summary>
        /// <param name="position">origin of ray.</param>
        /// <param name="direction">direction of ray. this must be unit vector.</param>
        public Ray(Vector3 position, Vector3 direction)
        {
            Position = position;
            Direction = direction;
        }

        #region Override methods

        // Implement IEquatable<T>
        public bool Equals(Ray other)
        {
            return (Position.X == other.Position.X && Position.Y == other.Position.Y && Position.Z == other.Position.Z &&
                        Direction.X == other.Direction.X && Direction.Y == other.Direction.Y && Direction.Z == other.Direction.Z);
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj != null && obj is Ray)
            {
                equal = Equals((Ray)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return Position.GetHashCode() + Direction.GetHashCode();
        }

        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{Position:{0} Direction:{1}}}", Position.ToString(), Direction.ToString());
        }
        #endregion

        #region Intersections
        /// <summary>
        /// Intersection test against another bounding box.
        /// </summary>
        /// <param name="box"></param>
        /// <returns>Returns a nullable float.</returns>
        public Nullable<float> Intersects(BoundingBox box)
        {
            return box.Intersects(this);
        }

        public void Intersects(ref BoundingBox box, out Nullable<float> result)
        {
            box.Intersects(ref this, out result);
        }

        /// <summary>
        /// Intersection test against another bounding frustum.
        /// </summary>
        /// <param name="frustum"></param>
        /// <returns>Returns a nullable float.</returns>
        public Nullable<float> Intersects(BoundingFrustum frustum)
        {
            if (frustum == null)
            {
                throw new ArgumentNullException( "frustum" );
            }
            return frustum.Intersects(this);
        }

        /// <summary>
        /// Intersection test against another plane.
        /// </summary>
        /// <param name="plane"></param>
        /// <returns>Returns a nullable float.</returns>
        public Nullable<float> Intersects(Plane plane)
        {
            float nd = plane.Normal.X * Direction.X + plane.Normal.Y * Direction.Y + plane.Normal.Z * Direction.Z;
            const float epsilon = 1e-5f;

            // Ray and Plane are parallel
            if (Math.Abs(nd) < epsilon)
            {
                return null;
            }

            float dot = plane.Normal.X * Position.X + plane.Normal.Y * Position.Y + plane.Normal.Z * Position.Z;
            float distance = (-plane.D - dot) / nd;
            if (distance < 0.0f)
            {
                if ( distance < -epsilon )
                    return null;
                distance = 0.0f;
            }
            return new Nullable<float>(distance);
        }

        public void Intersects(ref Plane plane, out Nullable<float> result)
        {
            float nd = plane.Normal.X * Direction.X + plane.Normal.Y * Direction.Y + plane.Normal.Z * Direction.Z;
            const float epsilon = 1e-5f;

            // Ray and Plane are parallel
            if (Math.Abs(nd) < epsilon)
            {
                result = null;
                return;
            }

            float dot = plane.Normal.X * Position.X + plane.Normal.Y * Position.Y + plane.Normal.Z * Position.Z;
            float distance = (-plane.D - dot) / nd;
            if (distance < 0.0f)
            {
                if (distance < -epsilon)
                {
                    result = null;
                    return;
                }

                result = 0.0f;
            }

            result = new Nullable<float>(distance);
        }

        /// <summary>
        /// Intersection test against another sphere.
        /// </summary>
        /// <param name="sphere"></param>
        /// <returns>Returns a nullable float.</returns>
        public Nullable<float> Intersects(BoundingSphere sphere)
        {
            // This algorithm base on Real-Time Rendering Second Edition p.571.
            float lx = sphere.Center.X - Position.X;
            float ly = sphere.Center.Y - Position.Y;
            float lz = sphere.Center.Z - Position.Z;

            float l2 = lx * lx + ly * ly + lz * lz;
            float r2 = sphere.Radius * sphere.Radius;

            // Ray origin is inside sphere.
            if (l2 <= r2)
            {
                return 0.0f;
            }

            float s = lx * Direction.X + ly * Direction.Y + lz * Direction.Z; // Dot(l,Direction)
            if (s < 0.0f)
            {
                return null;
            }

            float m2 = l2 - s * s;
            if (m2 > r2)
            {
                return null;
            }

            // Compute nearest intersection point.
            float q = (float)System.Math.Sqrt(r2 - m2);
            return s - q;
        }

        public void Intersects(ref BoundingSphere sphere, out Nullable<float> result)
        {
            // This algorithm base on Real-Time Rendering Second Edition p.571.
            float lx = sphere.Center.X - Position.X;
            float ly = sphere.Center.Y - Position.Y;
            float lz = sphere.Center.Z - Position.Z;

            float l2 = lx * lx + ly * ly + lz * lz;
            float r2 = sphere.Radius * sphere.Radius;

            // Ray origin is inside sphere.
            if (l2 <= r2)
            {
                result = 0.0f;
                return;
            }

            result = null;

            float s = lx * Direction.X + ly * Direction.Y + lz * Direction.Z; // Dot(l,Direction)
            if (s < 0.0f)
            {
                return;
            }

            float m2 = l2 - s * s;
            if (m2 > r2)
            {
                return;
            }

            // Compute nearest intersection point.
            float q = (float)System.Math.Sqrt(r2 - m2);
            result = s - q;
        }
        #endregion

        #region Operator overloads
        public static bool operator ==(Ray a, Ray b)
        {
            return (a.Position.X == b.Position.X && a.Position.Y == b.Position.Y && a.Position.Z == b.Position.Z &&
                        a.Direction.X == b.Direction.X && a.Direction.Y == b.Direction.Y && a.Direction.Z == b.Direction.Z);
        }

        public static bool operator !=(Ray a, Ray b)
        {
            return (a.Position.X != b.Position.X || a.Position.Y != b.Position.Y || a.Position.Z != b.Position.Z ||
                        a.Direction.X != b.Direction.X || a.Direction.Y != b.Direction.Y || a.Direction.Z != b.Direction.Z);
        }
        #endregion
    }
}
