//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Globalization;
using System.Diagnostics.CodeAnalysis;

namespace Microsoft.Xna.Framework
{
    public struct Vector2 : IEquatable<Vector2>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float X;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Y;

        #region Static Properties
        private static Vector2 _zero = new Vector2();
        private static Vector2 _one = new Vector2(1.0f, 1.0f);
        private static Vector2 _unitX = new Vector2(1.0f, 0.0f);
        private static Vector2 _unitY = new Vector2(0.0f, 1.0f);

        /// <summary>
        /// Returns Vector(0,0)
        /// </summary>
        public static Vector2 Zero { get { return _zero; } }
        /// <summary>
        /// Returns Vector(1,1)
        /// </summary>
        public static Vector2 One { get { return _one; } }
        /// <summary>
        /// Returns Vector(1,0)
        /// </summary>
        public static Vector2 UnitX { get { return _unitX; } }
        /// <summary>
        /// Retunrs Vector( 0, 1 )
        /// </summary>
        public static Vector2 UnitY { get { return _unitY; } }
        #endregion

        #region Constructor
        public Vector2(float x, float y)
        {
            X = x; Y = y;
        }

        /// <summary>
        /// Create Vector2 (value, value)
        /// </summary>
        /// <param name="value"></param>
        public Vector2(float value)
        {
            X = Y = value;
        }
        #endregion

        #region Override methods
        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{X:{0} Y:{1}}}", X.ToString(ci), Y.ToString(ci));
        }

        public bool Equals(Vector2 other)
        {
            return X == other.X && Y == other.Y;
        }

       public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is Vector2)
            {
                equal = Equals((Vector2)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return X.GetHashCode() + Y.GetHashCode();
        }
        #endregion

        #region Operations
        public float Length()
        {
            float ls = X * X + Y * Y;
            return (float)System.Math.Sqrt((double)ls);
        }

        public float LengthSquared()
        {
            return X * X + Y * Y;
        }

        public static float Distance(Vector2 value1, Vector2 value2)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float ls = dx * dx + dy * dy;

            return (float)System.Math.Sqrt((double)ls);
        }

        public static void Distance(ref Vector2 value1, ref Vector2 value2, out float result)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float ls = dx * dx + dy * dy;

            result = (float)System.Math.Sqrt((double)ls);
        }


        public static float DistanceSquared(Vector2 value1, Vector2 value2)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            return dx * dx + dy * dy;
        }

        public static void DistanceSquared(ref Vector2 value1, ref Vector2 value2, out float result)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            result = dx * dx + dy * dy;
        }


        public static float Dot(Vector2 value1, Vector2 value2)
        {
            return value1.X * value2.X + value1.Y * value2.Y;
        }

        public static void Dot(ref Vector2 value1, ref Vector2 value2, out float result)
        {
            result = value1.X * value2.X + value1.Y * value2.Y;
        }


        public void Normalize()
        {
            float ls = X * X + Y * Y;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);
            X *= invNorm;
            Y *= invNorm;
        }

        public static Vector2 Normalize(Vector2 value)
        {
            Vector2 ans;
            float ls = value.X * value.X + value.Y * value.Y;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            ans.X = value.X * invNorm;
            ans.Y = value.Y * invNorm;

            return ans;
        }

        public static void Normalize(ref Vector2 value, out Vector2 result)
        {
            float ls = value.X * value.X + value.Y * value.Y;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            result.X = value.X * invNorm;
            result.Y = value.Y * invNorm;
        }


        /// <summary>
        /// Return refelct vector of given vector and normal.
        /// </summary>
        /// <param name="vector"></param>
        /// <param name="normal"></param>
        /// <returns></returns>
        public static Vector2 Reflect(Vector2 vector, Vector2 normal)
        {
            Vector2 ans;

            float dot = vector.X * normal.X + vector.Y * normal.Y;

            ans.X = vector.X - 2.0f * dot * normal.X;
            ans.Y = vector.Y - 2.0f * dot * normal.Y;

            return ans;
        }

        public static void Reflect(ref Vector2 vector, ref Vector2 normal, out Vector2 result)
        {
            float dot = vector.X * normal.X + vector.Y * normal.Y;

            result.X = vector.X - 2.0f * dot * normal.X;
            result.Y = vector.Y - 2.0f * dot * normal.Y;
        }

        public static Vector2 Min(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;
            ans.X = (value1.X < value2.X) ? value1.X : value2.X;
            ans.Y = (value1.Y < value2.Y) ? value1.Y : value2.Y;
            return ans;
        }

        public static void Min(ref Vector2 value1, ref Vector2 value2, out Vector2 result)
        {
            result.X = (value1.X < value2.X) ? value1.X : value2.X;
            result.Y = (value1.Y < value2.Y) ? value1.Y : value2.Y;
        }


        public static Vector2 Max(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;
            ans.X = (value1.X > value2.X) ? value1.X : value2.X;
            ans.Y = (value1.Y > value2.Y) ? value1.Y : value2.Y;
            return ans;
        }

        public static void Max(ref Vector2 value1, ref Vector2 value2, out Vector2 result)
        {
            result.X = (value1.X > value2.X) ? value1.X : value2.X;
            result.Y = (value1.Y > value2.Y) ? value1.Y : value2.Y;
        }


        public static Vector2 Clamp(Vector2 value1, Vector2 min, Vector2 max)
        {
            Vector2 result;

            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.
            float x = value1.X;
            x = (x > max.X) ? max.X : x;
            x = (x < min.X) ? min.X : x;

            float y = value1.Y;
            y = (y > max.Y) ? max.Y : y;
            y = (y < min.Y) ? min.Y : y;

            result.X = x;
            result.Y = y;

            return result;
        }

        public static void Clamp(ref Vector2 value1, ref Vector2 min, ref Vector2 max, out Vector2 result)
        {
            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.
            float x = value1.X;
            x = (x > max.X) ? max.X : x;
            x = (x < min.X) ? min.X : x;

            float y = value1.Y;
            y = (y > max.Y) ? max.Y : y;
            y = (y < min.Y) ? min.Y : y;

            result.X = x;
            result.Y = y;
        }


        public static Vector2 Lerp(Vector2 value1, Vector2 value2, float amount)
        {
            Vector2 ans;

            ans.X = value1.X + (value2.X - value1.X) * amount;
            ans.Y = value1.Y + (value2.Y - value1.Y) * amount;

            return ans;
        }

        public static void Lerp(ref Vector2 value1, ref Vector2 value2, float amount, out Vector2 result)
        {
            result.X = value1.X + (value2.X - value1.X) * amount;
            result.Y = value1.Y + (value2.Y - value1.Y) * amount;
        }


        /// <summary>
        /// Returns a point in Barycentric coordinates, using the specified values.
        /// value1 + amount1( value2 - value1 ) + amount2( value3 - value1 )
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="amount1"></param>
        /// <param name="amount2"></param>
        /// <returns></returns>
        public static Vector2 Barycentric(Vector2 value1, Vector2 value2, Vector2 value3, float amount1, float amount2)
        {
            Vector2 ans;

            ans.X = value1.X + amount1 * (value2.X - value1.X) + amount2 * (value3.X - value1.X);
            ans.Y = value1.Y + amount1 * (value2.Y - value1.Y) + amount2 * (value3.Y - value1.Y);

            return ans;
        }

        public static void Barycentric(ref Vector2 value1, ref Vector2 value2, ref Vector2 value3, float amount1, float amount2, out Vector2 result)
        {
            result.X = value1.X + amount1 * (value2.X - value1.X) + amount2 * (value3.X - value1.X);
            result.Y = value1.Y + amount1 * (value2.Y - value1.Y) + amount2 * (value3.Y - value1.Y);
        }


        public static Vector2 SmoothStep(Vector2 value1, Vector2 value2, float amount)
        {
            Vector2 ans;

            amount = (amount > 1.0f) ? 1.0f : ((amount < 0.0f) ? 0.0f : amount);    // Clamp amount.

            amount = amount * amount * (3.0f - 2.0f * amount);
            ans.X = value1.X + (value2.X - value1.X) * amount;
            ans.Y = value1.Y + (value2.Y - value1.Y) * amount;

            return ans;
        }

        public static void SmoothStep(ref Vector2 value1, ref Vector2 value2, float amount, out Vector2 result)
        {
            amount = (amount > 1.0f) ? 1.0f : ((amount < 0.0f) ? 0.0f : amount);    // Clamp amount.

            amount = amount * amount * (3.0f - 2.0f * amount);
            result.X = value1.X + (value2.X - value1.X) * amount;
            result.Y = value1.Y + (value2.Y - value1.Y) * amount;
        }


        /// <summary>
        /// Perfroms a Catmull-Rom interpolation, using the specfied 2D vectors.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="value4"></param>
        /// <param name="amount">Weighting factor.</param>
        /// <returns></returns>
        public static Vector2 CatmullRom(Vector2 value1, Vector2 value2, Vector2 value3, Vector2 value4, float amount)
        {
            Vector2 ans;

            float t2 = amount * amount;
            float t3 = amount * t2;

            ans.X = 0.5f * (
                        (2.0f * value2.X) +
                        (-value1.X + value3.X) * amount +
                        (2.0f * value1.X - 5.0f * value2.X + 4.0f * value3.X - value4.X) * t2 +
                        (-value1.X + 3.0f * value2.X - 3.0f * value3.X + value4.X) * t3 );

            ans.Y = 0.5f * (
                        (2.0f * value2.Y) +
                        (-value1.Y + value3.Y) * amount +
                        (2.0f * value1.Y - 5.0f * value2.Y + 4.0f * value3.Y - value4.Y) * t2 +
                        (-value1.Y + 3.0f * value2.Y - 3.0f * value3.Y + value4.Y) * t3);

            return ans;
        }

        public static void CatmullRom(ref Vector2 value1, ref Vector2 value2, ref Vector2 value3, ref Vector2 value4, float amount, out Vector2 result)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;

            result.X = 0.5f * (
                        (2.0f * value2.X) +
                        (-value1.X + value3.X) * amount +
                        (2.0f * value1.X - 5.0f * value2.X + 4.0f * value3.X - value4.X) * t2 +
                        (-value1.X + 3.0f * value2.X - 3.0f * value3.X + value4.X) * t3);

            result.Y = 0.5f * (
                        (2.0f * value2.Y) +
                        (-value1.Y + value3.Y) * amount +
                        (2.0f * value1.Y - 5.0f * value2.Y + 4.0f * value3.Y - value4.Y) * t2 +
                        (-value1.Y + 3.0f * value2.Y - 3.0f * value3.Y + value4.Y) * t3);
        }



        /// <summary>
        /// Performs a Hermite spline interpolation, using the specfied 2D values.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="tangent1"></param>
        /// <param name="value2"></param>
        /// <param name="tangent2"></param>
        /// <param name="amount">Weighting factor</param>
        /// <returns></returns>
        public static Vector2 Hermite(Vector2 value1, Vector2 tangent1, Vector2 value2, Vector2 tangent2, float amount)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;
            float h1 = 2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 = t3 - 2.0f * t2 + amount;
            float h4 = t3 - t2;

            Vector2 ans;

            ans.X = value1.X * h1 + value2.X * h2 + tangent1.X * h3 + tangent2.X * h4;
            ans.Y = value1.Y * h1 + value2.Y * h2 + tangent1.Y * h3 + tangent2.Y * h4;

            return ans;
        }

        public static void Hermite(ref Vector2 value1, ref Vector2 tangent1, ref Vector2 value2, ref Vector2 tangent2, float amount, out Vector2 result)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;
            float h1 = 2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 = t3 - 2.0f * t2 + amount;
            float h4 = t3 - t2;

            result.X = value1.X * h1 + value2.X * h2 + tangent1.X * h3 + tangent2.X * h4;
            result.Y = value1.Y * h1 + value2.Y * h2 + tangent1.Y * h3 + tangent2.Y * h4;
        }

        #region Transform by matrix
        /// <summary>
        /// Transforms a 2D vector by a given matrix, projecting the result back into w = 1.
        /// </summary>
        /// <param name="position"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        /// <remarks>This function transforms the vector position(x,y,0,1), by the matrix, projecting the result back into w=1.</remarks>
        public static Vector2 Transform(Vector2 position, Matrix matrix)
        {
            float x = position.X * matrix.M11 + position.Y * matrix.M21 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + matrix.M42;

            Vector2 result;
            result.X = x;
            result.Y = y;
            return result;
        }

        public static void Transform(ref Vector2 position, ref Matrix matrix, out Vector2 result)
        {
            float x = position.X * matrix.M11 + position.Y * matrix.M21 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + matrix.M42;

            result.X = x;
            result.Y = y;
        }

        // Transform Normal
        /// <summary>
        /// Transforms the 2D vector normal by the given matrix.
        /// </summary>
        /// <param name="normal"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        /// <remarks>This method transforms the vector normal (x, y, 0, 0 ) of the vector, v by the matrix.</remarks>
        public static Vector2 TransformNormal(Vector2 normal, Matrix matrix)
        {
            float x = normal.X * matrix.M11 + normal.Y * matrix.M21;
            float y = normal.X * matrix.M12 + normal.Y * matrix.M22;

            Vector2 result;
            result.X = x;
            result.Y = y;

            return result;
        }

        public static void TransformNormal(ref Vector2 normal, ref Matrix matrix, out Vector2 result)
        {
            float x = normal.X * matrix.M11 + normal.Y * matrix.M21;
            float y = normal.X * matrix.M12 + normal.Y * matrix.M22;

            result.X = x;
            result.Y = y;
        }
        #endregion

        #region Transform by Quaternion
        /// <summary>
        /// Transform vector (x, y, 0, 0) by given quaternion.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        /// <remarks>This method transforms the vector normal (x, y, 0, 0 ) of the vector, v by the quaternion.</remarks>
        public static Vector2 Transform(Vector2 value, Quaternion rotation)
        {
            Vector2 result;

            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float yy2 = rotation.Y * y2;
            float zz2 = rotation.Z * z2;

            float x = value.X * (1.0f - yy2 - zz2) + value.Y * (xy2 - wz2);
            float y = value.X * (xy2 + wz2) + value.Y * (1.0f - xx2 - zz2);

            result.X = x;
            result.Y = y;

            return result;
        }

        public static void Transform(ref Vector2 value, ref Quaternion rotation, out Vector2 result)
        {
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float yy2 = rotation.Y * y2;
            float zz2 = rotation.Z * z2;

            float x = value.X * (1.0f - yy2 - zz2) + value.Y * (xy2 - wz2);
            float y = value.X * (xy2 + wz2) + value.Y * (1.0f - xx2 - zz2);

            result.X = x;
            result.Y = y;
        }
        #endregion

        #region Trasform by Array
        /// <summary>
        /// Transfrom by array of vectors by given matrix.
        /// </summary>
        /// <param name="sourceArray"></param>
        /// <param name="matrix"></param>
        /// <param name="destinationArray"></param>
        /// <remarks>This method transforms the vector normal (x, y, 0, 1 ) of the vector, v by the matrix.</remarks>
        public static void Transform(Vector2[] sourceArray, ref Matrix matrix, Vector2[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y;

                destinationArray[idx].X = sx * matrix.M11 + sy * matrix.M21 + matrix.M41;
                destinationArray[idx].Y = sx * matrix.M12 + sy * matrix.M22 + matrix.M42;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void Transform(Vector2[] sourceArray, int sourceIndex, ref Matrix matrix, Vector2[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y;

                destinationArray[destinationIndex].X = sx * matrix.M11 + sy * matrix.M21 + matrix.M41;
                destinationArray[destinationIndex].Y = sx * matrix.M12 + sy * matrix.M22 + matrix.M42;

                sourceIndex++; destinationIndex++;
                length--;
            }
        }

        /// <summary>
        /// Transfrom by array of vectors by given matrix.
        /// </summary>
        /// <param name="sourceArray"></param>
        /// <param name="matrix"></param>
        /// <param name="destinationArray"></param>
        /// <remarks>This method transforms the vector normal (x, y, 0, 0 ) of the vector, v by the matrix.</remarks>
        public static void TransformNormal(Vector2[] sourceArray, ref Matrix matrix, Vector2[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y;

                destinationArray[idx].X = sx * matrix.M11 + sy * matrix.M21;
                destinationArray[idx].Y = sx * matrix.M12 + sy * matrix.M22;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void TransformNormal(Vector2[] sourceArray, int sourceIndex, ref Matrix matrix, Vector2[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y;

                destinationArray[destinationIndex].X = sx * matrix.M11 + sy * matrix.M21;
                destinationArray[destinationIndex].Y = sx * matrix.M12 + sy * matrix.M22;

                sourceIndex++; destinationIndex++;
                length--;
            }
        }

        /// <summary>
        /// Transform array of vectors by given quaternion.
        /// </summary>
        /// <param name="sourceArray"></param>
        /// <param name="rotation"></param>
        /// <param name="destinationArray"></param>
        /// <remarks>This method transforms the vector normal (x, y, 0, 0 ) of the vector, v by the quaternion.</remarks>
        public static void Transform(Vector2[] sourceArray, ref Quaternion rotation, Vector2[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float yy2 = rotation.Y * y2;
            float zz2 = rotation.Z * z2;

            float m11 = (1.0f - yy2 - zz2);
            float m21 = (xy2 - wz2);

            float m12 = (xy2 + wz2);
            float m22 = (1.0f - xx2 - zz2);

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y;

                destinationArray[idx].X = sx * m11 + sy * m21;
                destinationArray[idx].Y = sx * m12 + sy * m22;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void Transform(Vector2[] sourceArray, int sourceIndex, ref Quaternion rotation, Vector2[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float yy2 = rotation.Y * y2;
            float zz2 = rotation.Z * z2;

            float m11 = (1.0f - yy2 - zz2);
            float m21 = (xy2 - wz2);

            float m12 = (xy2 + wz2);
            float m22 = (1.0f - xx2 - zz2);

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y;

                destinationArray[destinationIndex].X = sx * m11 + sy * m21;
                destinationArray[destinationIndex].Y = sx * m12 + sy * m22;

                sourceIndex++; destinationIndex++;
                length--;
            }

        }
        #endregion

        #endregion

        #region Operator overloads equivalent methods
        public static Vector2 Negate(Vector2 value)
        {
            Vector2 ans;

            ans.X = -value.X;
            ans.Y = -value.Y;

            return ans;
        }

        public static void Negate(ref Vector2 value, out Vector2 result)
        {
            result.X = -value.X;
            result.Y = -value.Y;
        }


        public static Vector2 Add(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X + value2.X;
            ans.Y = value1.Y + value2.Y;

            return ans;
        }

        public static void Add(ref Vector2 value1, ref Vector2 value2, out Vector2 result)
        {
            result.X = value1.X + value2.X;
            result.Y = value1.Y + value2.Y;
        }


        public static Vector2 Subtract(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X - value2.X;
            ans.Y = value1.Y - value2.Y;

            return ans;
        }

        public static void Subtract(ref Vector2 value1, ref Vector2 value2, out Vector2 result)
        {
            result.X = value1.X - value2.X;
            result.Y = value1.Y - value2.Y;
        }


        public static Vector2 Multiply(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X * value2.X;
            ans.Y = value1.Y * value2.Y;

            return ans;
        }

        public static void Multiply(ref Vector2 value1, ref Vector2 value2, out Vector2 result)
        {
            result.X = value1.X * value2.X;
            result.Y = value1.Y * value2.Y;
        }


        public static Vector2 Multiply(Vector2 value1, float scaleFactor)
        {
            Vector2 ans;

            ans.X = value1.X * scaleFactor;
            ans.Y = value1.Y * scaleFactor;

            return ans;
        }

        public static void Multiply(ref Vector2 value1, float scaleFactor, out Vector2 result)
        {
            result.X = value1.X * scaleFactor;
            result.Y = value1.Y * scaleFactor;
        }


        public static Vector2 Divide(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X / value2.X;
            ans.Y = value1.Y / value2.Y;

            return ans;
        }

        public static void Divide(ref Vector2 value1, ref Vector2 value2, out Vector2 result)
        {
            result.X = value1.X / value2.X;
            result.Y = value1.Y / value2.Y;
        }


        public static Vector2 Divide(Vector2 value1, float divider)
        {
            Vector2 ans;

            float invDiv = 1.0f / divider;
            ans.X = value1.X * invDiv;
            ans.Y = value1.Y * invDiv;

            return ans;
        }

        public static void Divide(ref Vector2 value1, float divider, out Vector2 result)
        {
            float invDiv = 1.0f / divider;
            result.X = value1.X * invDiv;
            result.Y = value1.Y * invDiv;
        }

        #endregion

        #region Operator overloads
        // Unary operator overloads
        public static Vector2 operator -(Vector2 value)
        {
            Vector2 ans;

            ans.X = -value.X;
            ans.Y = -value.Y;

            return ans;
        }

        // Binary operator overloads
        public static bool operator ==(Vector2 value1, Vector2 value2)
        {
            return (value1.X == value2.X && value1.Y == value2.Y);
        }

        public static bool operator !=(Vector2 value1, Vector2 value2)
        {
            return (value1.X != value2.X || value1.Y != value2.Y);
        }

        public static Vector2 operator +(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X + value2.X;
            ans.Y = value1.Y + value2.Y;

            return ans;
        }

        public static Vector2 operator -(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X - value2.X;
            ans.Y = value1.Y - value2.Y;

            return ans;
        }

        public static Vector2 operator * (Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X * value2.X;
            ans.Y = value1.Y * value2.Y;

            return ans;
        }

        public static Vector2 operator *(Vector2 value, float scaleFactor )
        {
            Vector2 ans;

            ans.X = value.X * scaleFactor;
            ans.Y = value.Y * scaleFactor;

            return ans;
        }

        public static Vector2 operator *(float scaleFactor, Vector2 value)
        {
            Vector2 ans;

            ans.X = value.X * scaleFactor;
            ans.Y = value.Y * scaleFactor;

            return ans;
        }

        public static Vector2 operator /(Vector2 value1, Vector2 value2)
        {
            Vector2 ans;

            ans.X = value1.X / value2.X;
            ans.Y = value1.Y / value2.Y;

            return ans;
        }

        public static Vector2 operator /(Vector2 value1, float divider )
        {
            Vector2 ans;

            float invDiv = 1.0f / divider;
            ans.X = value1.X * invDiv;
            ans.Y = value1.Y * invDiv;

            return ans;
        }
        #endregion
    }



}
