//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Globalization;
using System.Diagnostics.CodeAnalysis;

namespace Microsoft.Xna.Framework
{
    public struct Vector3 : IEquatable<Vector3>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float X;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Y;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Z;

        #region Static Properties
        private static Vector3 _zero    = new Vector3();
        private static Vector3 _one     = new Vector3( 1.0f, 1.0f, 1.0f);
        private static Vector3 _unitX   = new Vector3( 1.0f, 0.0f, 0.0f);
        private static Vector3 _unitY   = new Vector3( 0.0f, 1.0f, 0.0f);
        private static Vector3 _unitZ   = new Vector3( 0.0f, 0.0f, 1.0f);
        private static Vector3 _up      = new Vector3( 0.0f, 1.0f, 0.0f);
        private static Vector3 _down    = new Vector3( 0.0f,-1.0f, 0.0f);
        private static Vector3 _right   = new Vector3( 1.0f, 0.0f, 0.0f);
        private static Vector3 _left    = new Vector3(-1.0f, 0.0f, 0.0f);
        private static Vector3 _forward = new Vector3(0.0f, 0.0f, -1.0f);
        private static Vector3 _backward= new Vector3( 0.0f, 0.0f, 1.0f);

        /// <summary>
        /// Retunrs Vector( 0, 0, 0 )
        /// </summary>
        public static Vector3 Zero      { get { return _zero; } }
        /// <summary>
        /// Retunrs Vector( 1, 1, 1 )
        /// </summary>
        public static Vector3 One       { get { return _one; } }
        /// <summary>
        /// Retunrs Vector( 1, 0, 0 )
        /// </summary>
        public static Vector3 UnitX     { get { return _unitX; } }
        /// <summary>
        /// Retunrs Vector( 0, 1, 0 )
        /// </summary>
        public static Vector3 UnitY     { get { return _unitY; } }
        /// <summary>
        /// Returns Vector( 0, 0, 1 )
        /// </summary>
        public static Vector3 UnitZ     { get { return _unitZ; } }
        /// <summary>
        /// Returns Up direction unit vector ( 0,1, 0)
        /// </summary>
        public static Vector3 Up        { get { return _up; } }
        /// <summary>
        /// Returns Down direction unit vector ( 0,-1, 0)
        /// </summary>
        public static Vector3 Down { get { return _down; } }
        /// <summary>
        /// Returns Right direction unit vector ( 1, 0, 0)
        /// </summary>
        public static Vector3 Right { get { return _right; } }
        /// <summary>
        /// Returns Left direction unit vector ( -1, 0, 0)
        /// </summary>
        public static Vector3 Left { get { return _left; } }
        /// <summary>
        /// Returns Forward direction unit vector ( 0, 0, -1)
        /// </summary>
        public static Vector3 Forward { get { return _forward; } }
        /// <summary>
        /// Returns Backward direction unit vector ( 0, 0, 1)
        /// </summary>
        public static Vector3 Backward { get { return _backward; } }
        #endregion

        #region Constructor
        public Vector3(float x, float y, float z)
        {
            X = x; Y = y; Z = z;
        }

        /// <summary>
        /// Create Vector3 (value, value, value)
        /// </summary>
        /// <param name="value"></param>
        public Vector3(float value)
        {
            X = Y = Z = value;
        }

        public Vector3(Vector2 value, float z)
        {
            X = value.X; Y = value.Y; Z = z;
        }
        #endregion

        #region Override methods
        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{X:{0} Y:{1} Z:{2}}}", X.ToString(ci), Y.ToString(ci), Z.ToString(ci));
        }

        public bool Equals(Vector3 other)
        {
            return X == other.X && Y == other.Y && Z == other.Z;
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is Vector3)
            {
                equal = Equals((Vector3)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return X.GetHashCode() + Y.GetHashCode() + Z.GetHashCode();
        }
        #endregion

        #region Operations
        // Unary Operators.
        public float Length()
        {
            float ls = X * X + Y * Y + Z * Z;
            return (float)System.Math.Sqrt((double)ls);
        }

        public float LengthSquared()
        {
            return X * X + Y * Y + Z * Z;
        }

        public static float Distance(Vector3 value1, Vector3 value2)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            float ls = dx * dx + dy * dy + dz * dz;

            return (float)System.Math.Sqrt((double)ls);
        }

        public static void Distance(ref Vector3 value1, ref Vector3 value2, out float result)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            float ls = dx * dx + dy * dy + dz * dz;

            result = (float)System.Math.Sqrt((double)ls);
        }


        public static float DistanceSquared(Vector3 value1, Vector3 value2)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            return dx * dx + dy * dy + dz * dz;
        }

        public static void DistanceSquared(ref Vector3 value1, ref Vector3 value2, out float result)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            result = dx * dx + dy * dy + dz * dz;
        }


        public static float Dot(Vector3 vector1, Vector3 vector2)
        {
            return vector1.X * vector2.X + vector1.Y * vector2.Y + vector1.Z * vector2.Z;
        }

        public static void Dot(ref Vector3 vector1, ref Vector3 vector2, out float result)
        {
            result = vector1.X * vector2.X + vector1.Y * vector2.Y + vector1.Z * vector2.Z;
        }


        public void Normalize()
        {
            float ls = X * X + Y * Y + Z * Z;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);
            X *= invNorm;
            Y *= invNorm;
            Z *= invNorm;
        }

        public static Vector3 Normalize(Vector3 value)
        {
            Vector3 ans;
            float ls = value.X * value.X + value.Y * value.Y + value.Z * value.Z;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            ans.X = value.X * invNorm;
            ans.Y = value.Y * invNorm;
            ans.Z = value.Z * invNorm;

            return ans;
        }

        public static void Normalize(ref Vector3 value, out Vector3 result)
        {
            float ls = value.X * value.X + value.Y * value.Y + value.Z * value.Z;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            result.X = value.X * invNorm;
            result.Y = value.Y * invNorm;
            result.Z = value.Z * invNorm;
        }


        public static Vector3 Cross(Vector3 vector1, Vector3 vector2)
        {
            Vector3 ans;

            ans.X = vector1.Y * vector2.Z - vector1.Z * vector2.Y;
            ans.Y = vector1.Z * vector2.X - vector1.X * vector2.Z;
            ans.Z = vector1.X * vector2.Y - vector1.Y * vector2.X;

            return ans;
        }

        public static void Cross(ref Vector3 vector1, ref Vector3 vector2, out Vector3 result)
        {
            float x = vector1.Y * vector2.Z - vector1.Z * vector2.Y;
            float y = vector1.Z * vector2.X - vector1.X * vector2.Z;
            float z = vector1.X * vector2.Y - vector1.Y * vector2.X;

            result.X = x;
            result.Y = y;
            result.Z = z;
        }


        /// <summary>
        /// Return refelct vector of given vector and normal.
        /// </summary>
        /// <param name="vector"></param>
        /// <param name="normal"></param>
        /// <returns></returns>
        public static Vector3 Reflect(Vector3 vector, Vector3 normal)
        {
            Vector3 ans;

            float dot = vector.X * normal.X + vector.Y * normal.Y + vector.Z * normal.Z;

            ans.X = vector.X - 2.0f * dot * normal.X;
            ans.Y = vector.Y - 2.0f * dot * normal.Y;
            ans.Z = vector.Z - 2.0f * dot * normal.Z;

            return ans;
        }

        public static void Reflect(ref Vector3 vector, ref Vector3 normal, out Vector3 result)
        {
            float dot = vector.X * normal.X + vector.Y * normal.Y + vector.Z * normal.Z;

            result.X = vector.X - 2.0f * dot * normal.X;
            result.Y = vector.Y - 2.0f * dot * normal.Y;
            result.Z = vector.Z - 2.0f * dot * normal.Z;
        }


        public static Vector3 Min(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = (value1.X < value2.X) ? value1.X : value2.X;
            ans.Y = (value1.Y < value2.Y) ? value1.Y : value2.Y;
            ans.Z = (value1.Z < value2.Z) ? value1.Z : value2.Z;

            return ans;
        }

        public static void Min(ref Vector3 value1, ref Vector3 value2, out Vector3 result)
        {
            result.X = (value1.X < value2.X) ? value1.X : value2.X;
            result.Y = (value1.Y < value2.Y) ? value1.Y : value2.Y;
            result.Z = (value1.Z < value2.Z) ? value1.Z : value2.Z;
        }


        public static Vector3 Max(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = (value1.X > value2.X) ? value1.X : value2.X;
            ans.Y = (value1.Y > value2.Y) ? value1.Y : value2.Y;
            ans.Z = (value1.Z > value2.Z) ? value1.Z : value2.Z;

            return ans;
        }

        public static void Max(ref Vector3 value1, ref Vector3 value2, out Vector3 result)
        {
            result.X = (value1.X > value2.X) ? value1.X : value2.X;
            result.Y = (value1.Y > value2.Y) ? value1.Y : value2.Y;
            result.Z = (value1.Z > value2.Z) ? value1.Z : value2.Z;
        }


        public static Vector3 Clamp(Vector3 value1, Vector3 min, Vector3 max)
        {
            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.
            Vector3 result;

            float x = value1.X;
            x = (x > max.X) ? max.X : x;
            x = (x < min.X) ? min.X : x;

            float y = value1.Y;
            y = (y > max.Y) ? max.Y : y;
            y = (y < min.Y) ? min.Y : y;

            float z = value1.Z;
            z = (z > max.Z) ? max.Z : z;
            z = (z < min.Z) ? min.Z : z;

            result.X = x;
            result.Y = y;
            result.Z = z;

            return result;
        }

        public static void Clamp(ref Vector3 value1, ref Vector3 min, ref Vector3 max, out Vector3 result)
        {
            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.

            float x = value1.X;
            x = (x > max.X) ? max.X : x;
            x = (x < min.X) ? min.X : x;

            float y = value1.Y;
            y = (y > max.Y) ? max.Y : y;
            y = (y < min.Y) ? min.Y : y;

            float z = value1.Z;
            z = (z > max.Z) ? max.Z : z;
            z = (z < min.Z) ? min.Z : z;

            result.X = x;
            result.Y = y;
            result.Z = z;
        }


        public static Vector3 Lerp(Vector3 value1, Vector3 value2, float amount)
        {
            Vector3 ans;

            ans.X = value1.X + (value2.X - value1.X) * amount;
            ans.Y = value1.Y + (value2.Y - value1.Y) * amount;
            ans.Z = value1.Z + (value2.Z - value1.Z) * amount;

            return ans;
        }

        public static void Lerp(ref Vector3 value1, ref Vector3 value2, float amount, out Vector3 result)
        {
            result.X = value1.X + (value2.X - value1.X) * amount;
            result.Y = value1.Y + (value2.Y - value1.Y) * amount;
            result.Z = value1.Z + (value2.Z - value1.Z) * amount;
        }


        /// <summary>
        /// Returns a point in Barycentric coordinates, using the specified values.
        /// value1 + amount1( value2 - value1 ) + amount2( value3 - value1 )
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="amount1"></param>
        /// <param name="amount2"></param>
        /// <returns></returns>
        public static Vector3 Barycentric(Vector3 value1, Vector3 value2, Vector3 value3, float amount1, float amount2)
        {
            Vector3 ans;

            ans.X = value1.X + amount1 * (value2.X - value1.X) + amount2 * (value3.X - value1.X);
            ans.Y = value1.Y + amount1 * (value2.Y - value1.Y) + amount2 * (value3.Y - value1.Y);
            ans.Z = value1.Z + amount1 * (value2.Z - value1.Z) + amount2 * (value3.Z - value1.Z);

            return ans;
        }

        public static void Barycentric(ref Vector3 value1, ref Vector3 value2, ref Vector3 value3, float amount1, float amount2, out Vector3 result)
        {
            result.X = value1.X + amount1 * (value2.X - value1.X) + amount2 * (value3.X - value1.X);
            result.Y = value1.Y + amount1 * (value2.Y - value1.Y) + amount2 * (value3.Y - value1.Y);
            result.Z = value1.Z + amount1 * (value2.Z - value1.Z) + amount2 * (value3.Z - value1.Z);
        }


        public static Vector3 SmoothStep(Vector3 value1, Vector3 value2, float amount)
        {
            Vector3 ans;

            amount = (amount > 1.0f) ? 1.0f : ((amount < 0.0f) ? 0.0f : amount);    // Clamp amount.

            amount = amount * amount * (3.0f - 2.0f * amount);
            ans.X = value1.X + (value2.X - value1.X) * amount;
            ans.Y = value1.Y + (value2.Y - value1.Y) * amount;
            ans.Z = value1.Z + (value2.Z - value1.Z) * amount;

            return ans;
        }

        public static void SmoothStep(ref Vector3 value1, ref Vector3 value2, float amount, out Vector3 result)
        {
            amount = (amount > 1.0f) ? 1.0f : ((amount < 0.0f) ? 0.0f : amount);    // Clamp amount.

            amount = amount * amount * (3.0f - 2.0f * amount);
            result.X = value1.X + (value2.X - value1.X) * amount;
            result.Y = value1.Y + (value2.Y - value1.Y) * amount;
            result.Z = value1.Z + (value2.Z - value1.Z) * amount;
        }


        /// <summary>
        /// Perfroms a Catmull-Rom interpolation, using the specfied 3D vectors.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="value4"></param>
        /// <param name="amount">Weighting factor.</param>
        /// <returns></returns>
        public static Vector3 CatmullRom(Vector3 value1, Vector3 value2, Vector3 value3, Vector3 value4, float amount)
        {
            Vector3 ans;

            float t2 = amount * amount;
            float t3 = amount * t2;

            ans.X = 0.5f * (
                        (2.0f * value2.X) +
                        (-value1.X + value3.X) * amount +
                        (2.0f * value1.X - 5.0f * value2.X + 4.0f * value3.X - value4.X) * t2 +
                        (-value1.X + 3.0f * value2.X - 3.0f * value3.X + value4.X) * t3);

            ans.Y = 0.5f * (
                        (2.0f * value2.Y) +
                        (-value1.Y + value3.Y) * amount +
                        (2.0f * value1.Y - 5.0f * value2.Y + 4.0f * value3.Y - value4.Y) * t2 +
                        (-value1.Y + 3.0f * value2.Y - 3.0f * value3.Y + value4.Y) * t3);

            ans.Z = 0.5f * (
                        (2.0f * value2.Z) +
                        (-value1.Z + value3.Z) * amount +
                        (2.0f * value1.Z - 5.0f * value2.Z + 4.0f * value3.Z - value4.Z) * t2 +
                        (-value1.Z + 3.0f * value2.Z - 3.0f * value3.Z + value4.Z) * t3);

            return ans;
        }

        public static void CatmullRom(ref Vector3 value1, ref Vector3 value2, ref Vector3 value3, ref Vector3 value4, float amount, out Vector3 result)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;

            result.X = 0.5f * (
                        (2.0f * value2.X) +
                        (-value1.X + value3.X) * amount +
                        (2.0f * value1.X - 5.0f * value2.X + 4.0f * value3.X - value4.X) * t2 +
                        (-value1.X + 3.0f * value2.X - 3.0f * value3.X + value4.X) * t3);

            result.Y = 0.5f * (
                        (2.0f * value2.Y) +
                        (-value1.Y + value3.Y) * amount +
                        (2.0f * value1.Y - 5.0f * value2.Y + 4.0f * value3.Y - value4.Y) * t2 +
                        (-value1.Y + 3.0f * value2.Y - 3.0f * value3.Y + value4.Y) * t3);

            result.Z = 0.5f * (
                        (2.0f * value2.Z) +
                        (-value1.Z + value3.Z) * amount +
                        (2.0f * value1.Z - 5.0f * value2.Z + 4.0f * value3.Z - value4.Z) * t2 +
                        (-value1.Z + 3.0f * value2.Z - 3.0f * value3.Z + value4.Z) * t3);
        }


        /// <summary>
        /// Performs a Hermite spline interpolation, using the specfied 3D values.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="tangent1"></param>
        /// <param name="value2"></param>
        /// <param name="tangent2"></param>
        /// <param name="amount">Weighting factor</param>
        /// <returns></returns>
        public static Vector3 Hermite(Vector3 value1, Vector3 tangent1, Vector3 value2, Vector3 tangent2, float amount)
        {
            Vector3 ans;

            float t2 = amount * amount;
            float t3 = amount * t2;
            float h1 = 2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 = t3 - 2.0f * t2 + amount;
            float h4 = t3 - t2;

            ans.X = value1.X * h1 + value2.X * h2 + tangent1.X * h3 + tangent2.X * h4;
            ans.Y = value1.Y * h1 + value2.Y * h2 + tangent1.Y * h3 + tangent2.Y * h4;
            ans.Z = value1.Z * h1 + value2.Z * h2 + tangent1.Z * h3 + tangent2.Z * h4;

            return ans;
        }

        public static void Hermite(ref Vector3 value1, ref Vector3 tangent1, ref Vector3 value2, ref Vector3 tangent2, float amount, out Vector3 result)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;
            float h1 = 2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 = t3 - 2.0f * t2 + amount;
            float h4 = t3 - t2;

            result.X = value1.X * h1 + value2.X * h2 + tangent1.X * h3 + tangent2.X * h4;
            result.Y = value1.Y * h1 + value2.Y * h2 + tangent1.Y * h3 + tangent2.Y * h4;
            result.Z = value1.Z * h1 + value2.Z * h2 + tangent1.Z * h3 + tangent2.Z * h4;
        }

        #region Transform by Matrix

        /// <summary>
        /// Transforms a 3D vector by a given matrix, projecting the result back into w = 1.
        /// </summary>
        /// <param name="position"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        /// <remarks>This function transforms the vector position(x,y,z,1), by the matrix, projecting the result back into w=1.</remarks>
        public static Vector3 Transform(Vector3 position, Matrix matrix)
        {
            Vector3 result;

            float x = position.X * matrix.M11 + position.Y * matrix.M21 + position.Z * matrix.M31 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + position.Z * matrix.M32 + matrix.M42;
            float z = position.X * matrix.M13 + position.Y * matrix.M23 + position.Z * matrix.M33 + matrix.M43;

            result.X = x;
            result.Y = y;
            result.Z = z;

            return result;
        }

        public static void Transform(ref Vector3 position, ref Matrix matrix, out Vector3 result)
        {
            float x = position.X * matrix.M11 + position.Y * matrix.M21 + position.Z * matrix.M31 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + position.Z * matrix.M32 + matrix.M42;
            float z = position.X * matrix.M13 + position.Y * matrix.M23 + position.Z * matrix.M33 + matrix.M43;

            result.X = x;
            result.Y = y;
            result.Z = z;
        }


        /// <summary>
        /// Transforms the 3D vector normal by the given matrix.
        /// </summary>
        /// <param name="normal"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        /// <remarks>This method transforms the vector normal (x, y, z, 0 ) of the vector, v by the matrix.</remarks>
        public static Vector3 TransformNormal(Vector3 normal, Matrix matrix)
        {
            Vector3 result;

            float x = normal.X * matrix.M11 + normal.Y * matrix.M21 + normal.Z * matrix.M31;
            float y = normal.X * matrix.M12 + normal.Y * matrix.M22 + normal.Z * matrix.M32;
            float z = normal.X * matrix.M13 + normal.Y * matrix.M23 + normal.Z * matrix.M33;

            result.X = x;
            result.Y = y;
            result.Z = z;

            return result;
        }

        public static void TransformNormal(ref Vector3 normal, ref Matrix matrix, out Vector3 result)
        {
            float x = normal.X * matrix.M11 + normal.Y * matrix.M21 + normal.Z * matrix.M31;
            float y = normal.X * matrix.M12 + normal.Y * matrix.M22 + normal.Z * matrix.M32;
            float z = normal.X * matrix.M13 + normal.Y * matrix.M23 + normal.Z * matrix.M33;

            result.X = x;
            result.Y = y;
            result.Z = z;
        }
        #endregion


        #region Transform by Quaternion
        /// <summary>
        /// Transform vector (x, y, z, 1) by given quaternion.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        /// <remarks>This method transforms the vector normal (x, y, z, 0 ) of the vector, v by the quaternion.</remarks>
        public static Vector3 Transform(Vector3 value, Quaternion rotation)
        {
            Vector3 result;

            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * (1.0f - yy2 - zz2) + value.Y * (xy2 - wz2) + value.Z * (xz2 + wy2);
            float y = value.X * (xy2 + wz2) + value.Y * (1.0f - xx2 - zz2) + value.Z * (yz2 - wx2);
            float z = value.X * (xz2 - wy2) + value.Y * (yz2 + wx2) + value.Z * (1.0f - xx2 - yy2);

            result.X = x;
            result.Y = y;
            result.Z = z;

            return result;
        }

        public static void Transform(ref Vector3 value, ref Quaternion rotation, out Vector3 result)
        {
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * (1.0f - yy2 - zz2) + value.Y * (xy2 - wz2) + value.Z * (xz2 + wy2);
            float y = value.X * (xy2 + wz2) + value.Y * (1.0f - xx2 - zz2) + value.Z * (yz2 - wx2);
            float z = value.X * (xz2 - wy2) + value.Y * (yz2 + wx2) + value.Z * (1.0f - xx2 - yy2);

            result.X = x;
            result.Y = y;
            result.Z = z;
        }
        #endregion

        #region Trasform by Array
        /// <summary>
        /// Transform array of vectors (x, y, z, 1) by given matrix.
        /// </summary>
        /// <param name="sourceArray"></param>
        /// <param name="matrix"></param>
        /// <param name="destinationArray"></param>
        /// <returns></returns>
        /// <remarks>This method transforms the vector (x, y, z, 1 ) of the vector, v by the matrix.</remarks>
        public static void Transform(Vector3[] sourceArray, ref Matrix matrix, Vector3[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y, sz = sourceArray[idx].Z;

                destinationArray[idx].X = sx * matrix.M11 + sy * matrix.M21 + sz * matrix.M31 + matrix.M41;
                destinationArray[idx].Y = sx * matrix.M12 + sy * matrix.M22 + sz * matrix.M32 + matrix.M42;
                destinationArray[idx].Z = sx * matrix.M13 + sy * matrix.M23 + sz * matrix.M33 + matrix.M43;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void Transform(Vector3[] sourceArray, int sourceIndex, ref Matrix matrix, Vector3[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y, sz = sourceArray[sourceIndex].Z;

                destinationArray[destinationIndex].X = sx * matrix.M11 + sy * matrix.M21 + sz * matrix.M31 + matrix.M41;
                destinationArray[destinationIndex].Y = sx * matrix.M12 + sy * matrix.M22 + sz * matrix.M32 + matrix.M42;
                destinationArray[destinationIndex].Z = sx * matrix.M13 + sy * matrix.M23 + sz * matrix.M33 + matrix.M43;

                sourceIndex++; destinationIndex++;
                length--;
            }
        }

        /// <summary>
        /// Transfrom by array of vectors by given matrix.
        /// This method treat vector3 as (x, y, z, 0)
        /// </summary>
        /// <param name="sourceArray"></param>
        /// <param name="matrix"></param>
        /// <param name="destinationArray"></param>
        /// <remarks>This method transforms the vector (x, y, z, 0 ) of the vector, v by the matrix.</remarks>
        public static void TransformNormal(Vector3[] sourceArray, ref Matrix matrix, Vector3[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y, sz = sourceArray[idx].Z;

                destinationArray[idx].X = sx * matrix.M11 + sy * matrix.M21 + sz * matrix.M31;
                destinationArray[idx].Y = sx * matrix.M12 + sy * matrix.M22 + sz * matrix.M32;
                destinationArray[idx].Z = sx * matrix.M13 + sy * matrix.M23 + sz * matrix.M33;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void TransformNormal(Vector3[] sourceArray, int sourceIndex, ref Matrix matrix, Vector3[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y, sz = sourceArray[sourceIndex].Z;

                destinationArray[destinationIndex].X = sx * matrix.M11 + sy * matrix.M21 + sz * matrix.M31;
                destinationArray[destinationIndex].Y = sx * matrix.M12 + sy * matrix.M22 + sz * matrix.M32;
                destinationArray[destinationIndex].Z = sx * matrix.M13 + sy * matrix.M23 + sz * matrix.M33;

                sourceIndex++; destinationIndex++;
                length--;
            }
        }

        /// <summary>
        /// Transform array of vectors (x, y, z, 1) by given quaternion.
        /// </summary>
        /// <param name="sourceArray"></param>
        /// <param name="rotation"></param>
        /// <param name="destinationArray"></param>
        /// <returns></returns>
        /// <remarks>This method transforms the vector (x, y, z, 1 ) of the vector, v by the quaternion.</remarks>
        public static void Transform(Vector3[] sourceArray, ref Quaternion rotation, Vector3[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float m11 = 1.0f - yy2 - zz2;
            float m21 = xy2 - wz2;
            float m31 = xz2 + wy2;

            float m12 = xy2 + wz2;
            float m22 = 1.0f - xx2 - zz2;
            float m32 = yz2 - wx2;

            float m13 = xz2 - wy2;
            float m23 = yz2 + wx2;
            float m33 = 1.0f - xx2 - yy2;

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y, sz = sourceArray[idx].Z;

                destinationArray[idx].X = sx * m11 + sy * m21 + sz * m31;
                destinationArray[idx].Y = sx * m12 + sy * m22 + sz * m32;
                destinationArray[idx].Z = sx * m13 + sy * m23 + sz * m33;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void Transform(Vector3[] sourceArray, int sourceIndex, ref Quaternion rotation, Vector3[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float m11 = 1.0f - yy2 - zz2;
            float m21 = xy2 - wz2;
            float m31 = xz2 + wy2;

            float m12 = xy2 + wz2;
            float m22 = 1.0f - xx2 - zz2;
            float m32 = yz2 - wx2;

            float m13 = xz2 - wy2;
            float m23 = yz2 + wx2;
            float m33 = 1.0f - xx2 - yy2;

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y, sz = sourceArray[sourceIndex].Z;

                destinationArray[destinationIndex].X = sx * m11 + sy * m21 + sz * m31;
                destinationArray[destinationIndex].Y = sx * m12 + sy * m22 + sz * m32;
                destinationArray[destinationIndex].Z = sx * m13 + sy * m23 + sz * m33;

                sourceIndex++; destinationIndex++;
                length--;
            }

        }
        #endregion

        #endregion

        #region Operator overloads equivalent methods
        public static Vector3 Negate(Vector3 value)
        {
            Vector3 ans;

            ans.X = -value.X;
            ans.Y = -value.Y;
            ans.Z = -value.Z;

            return ans;
        }

        public static void Negate(ref Vector3 value, out Vector3 result)
        {
            result.X = -value.X;
            result.Y = -value.Y;
            result.Z = -value.Z;
        }


        public static Vector3 Add(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X + value2.X;
            ans.Y = value1.Y + value2.Y;
            ans.Z = value1.Z + value2.Z;

            return ans;
        }

        public static void Add(ref Vector3 value1, ref Vector3 value2, out Vector3 result)
        {
            result.X = value1.X + value2.X;
            result.Y = value1.Y + value2.Y;
            result.Z = value1.Z + value2.Z;
        }


        public static Vector3 Subtract(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X - value2.X;
            ans.Y = value1.Y - value2.Y;
            ans.Z = value1.Z - value2.Z;

            return ans;
        }

        public static void Subtract(ref Vector3 value1, ref Vector3 value2, out Vector3 result)
        {
            result.X = value1.X - value2.X;
            result.Y = value1.Y - value2.Y;
            result.Z = value1.Z - value2.Z;
        }


        public static Vector3 Multiply(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X * value2.X;
            ans.Y = value1.Y * value2.Y;
            ans.Z = value1.Z * value2.Z;

            return ans;
        }

        public static void Multiply(ref Vector3 value1, ref Vector3 value2, out Vector3 result)
        {
            result.X = value1.X * value2.X;
            result.Y = value1.Y * value2.Y;
            result.Z = value1.Z * value2.Z;
        }


        public static Vector3 Multiply(Vector3 value1, float scaleFactor)
        {
            Vector3 ans;

            ans.X = value1.X * scaleFactor;
            ans.Y = value1.Y * scaleFactor;
            ans.Z = value1.Z * scaleFactor;

            return ans;
        }

        public static void Multiply(ref Vector3 value1, float scaleFactor, out Vector3 result)
        {
            result.X = value1.X * scaleFactor;
            result.Y = value1.Y * scaleFactor;
            result.Z = value1.Z * scaleFactor;
        }


        public static Vector3 Divide(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X / value2.X;
            ans.Y = value1.Y / value2.Y;
            ans.Z = value1.Z / value2.Z;

            return ans;
        }

        public static void Divide(ref Vector3 value1, ref Vector3 value2, out Vector3 result)
        {
            result.X = value1.X / value2.X;
            result.Y = value1.Y / value2.Y;
            result.Z = value1.Z / value2.Z;
        }


        public static Vector3 Divide(Vector3 value1, float value2)
        {
            Vector3 ans;

            float invDiv = 1.0f / value2;
            ans.X = value1.X * invDiv;
            ans.Y = value1.Y * invDiv;
            ans.Z = value1.Z * invDiv;

            return ans;
        }

        public static void Divide(ref Vector3 value1, float value2, out Vector3 result)
        {
            float invDiv = 1.0f / value2;
            result.X = value1.X * invDiv;
            result.Y = value1.Y * invDiv;
            result.Z = value1.Z * invDiv;
        }

        #endregion

        #region Operator overloads
        // Unary operator overloads
        public static Vector3 operator -(Vector3 value)
        {
            Vector3 ans;

            ans.X = -value.X;
            ans.Y = -value.Y;
            ans.Z = -value.Z;

            return ans;
        }

        // Binary operator overloads
        public static bool operator ==(Vector3 value1, Vector3 value2)
        {
            return (value1.X == value2.X && value1.Y == value2.Y && value1.Z == value2.Z);
        }

        public static bool operator !=(Vector3 value1, Vector3 value2)
        {
            return (value1.X != value2.X || value1.Y != value2.Y || value1.Z != value2.Z);
        }

        public static Vector3 operator +(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X + value2.X;
            ans.Y = value1.Y + value2.Y;
            ans.Z = value1.Z + value2.Z;

            return ans;
        }

        public static Vector3 operator -(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X - value2.X;
            ans.Y = value1.Y - value2.Y;
            ans.Z = value1.Z - value2.Z;

            return ans;
        }

        public static Vector3 operator * (Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X * value2.X;
            ans.Y = value1.Y * value2.Y;
            ans.Z = value1.Z * value2.Z;

            return ans;
        }

        public static Vector3 operator *(Vector3 value, float scaleFactor )
        {
            Vector3 ans;

            ans.X = value.X * scaleFactor;
            ans.Y = value.Y * scaleFactor;
            ans.Z = value.Z * scaleFactor;

            return ans;
        }

        public static Vector3 operator *(float scaleFactor, Vector3 value)
        {
            Vector3 ans;

            ans.X = value.X * scaleFactor;
            ans.Y = value.Y * scaleFactor;
            ans.Z = value.Z * scaleFactor;

            return ans;
        }

        public static Vector3 operator /(Vector3 value1, Vector3 value2)
        {
            Vector3 ans;

            ans.X = value1.X / value2.X;
            ans.Y = value1.Y / value2.Y;
            ans.Z = value1.Z / value2.Z;

            return ans;
        }

        public static Vector3 operator /(Vector3 value, float divider )
        {
            Vector3 ans;

            float invDiv = 1.0f / divider;
            ans.X = value.X * invDiv;
            ans.Y = value.Y * invDiv;
            ans.Z = value.Z * invDiv;

            return ans;
        }
        #endregion

    }
}
