//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"


using System;
using System.Globalization;
using System.Diagnostics.CodeAnalysis;

namespace Microsoft.Xna.Framework
{
    public struct Vector4 : IEquatable<Vector4>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float X;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Y;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float Z;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float W;

        #region Static Properties
        private static Vector4 _zero = new Vector4();
        private static Vector4 _one = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
        private static Vector4 _unitX = new Vector4(1.0f, 0.0f, 0.0f, 0.0f);
        private static Vector4 _unitY = new Vector4(0.0f, 1.0f, 0.0f, 0.0f);
        private static Vector4 _unitZ = new Vector4(0.0f, 0.0f, 1.0f, 0.0f);
        private static Vector4 _unitW = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);

        /// <summary>
        /// Returns Vector (0, 0, 0, 0)
        /// </summary>
        public static Vector4 Zero { get { return _zero; } }
        /// <summary>
        /// Returns Vector (1, 1, 1, 1)
        /// </summary>
        public static Vector4 One { get { return _one; } }
        /// <summary>
        /// Returns Vector (1, 0, 0, 0)
        /// </summary>
        public static Vector4 UnitX { get { return _unitX; } }
        /// <summary>
        /// Returns Vector (0, 1, 0, 0)
        /// </summary>
        public static Vector4 UnitY { get { return _unitY; } }
        /// <summary>
        /// Returns Vector (0, 0, 1, 0)
        /// </summary>
        public static Vector4 UnitZ { get { return _unitZ; } }
        /// <summary>
        /// Returns Vector (0, 0, 0, 1)
        /// </summary>
        public static Vector4 UnitW { get { return _unitW; } }
        #endregion

        #region Constructors
        public Vector4(float x, float y, float z, float w)
        {
            X = x; Y = y; Z = z; W = w;
        }

        public Vector4(Vector2 value, float z, float w)
        {
            X = value.X; Y = value.Y; Z = z; W = w;
        }

        public Vector4(Vector3 value, float w)
        {
            X = value.X; Y = value.Y; Z = value.Z; W = w;
        }

        /// <summary>
        /// Create Vector4 (value, value, value, value)
        /// </summary>
        /// <param name="value"></param>
        public Vector4(float value)
        {
            X = Y = Z = W = value;
        }
        #endregion

        #region Override methods
        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return String.Format(ci, "{{X:{0} Y:{1} Z:{2} W:{3}}}", X.ToString(ci), Y.ToString(ci), Z.ToString(ci), W.ToString(ci));
        }

        public bool Equals(Vector4 other)
        {
            return X == other.X && Y == other.Y && Z == other.Z && W == other.W;
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is Vector4)
            {
                equal = Equals((Vector4)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return X.GetHashCode() + Y.GetHashCode() + Z.GetHashCode() + W.GetHashCode();
        }
        #endregion

        #region Operations
        // Unary Operators.
        public float Length()
        {
            float ls = X * X + Y * Y + Z * Z + W * W;
            return (float)System.Math.Sqrt((double)ls);
        }

        public float LengthSquared()
        {
            return X * X + Y * Y + Z * Z + W * W;
        }

        public static float Distance(Vector4 value1, Vector4 value2)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            float dw = value1.W - value2.W;
            float ls = dx * dx + dy * dy + dz * dz + dw * dw;

            return (float)System.Math.Sqrt((double)ls);
        }

        public static void Distance(ref Vector4 value1, ref Vector4 value2, out float result)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            float dw = value1.W - value2.W;
            float ls = dx * dx + dy * dy + dz * dz + dw * dw;

            result = (float)System.Math.Sqrt((double)ls);
        }

        public static float DistanceSquared(Vector4 value1, Vector4 value2)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            float dw = value1.W - value2.W;
            return dx * dx + dy * dy + dz * dz + dw * dw;
        }

        public static void DistanceSquared(ref Vector4 value1, ref Vector4 value2, out float result)
        {
            float dx = value1.X - value2.X;
            float dy = value1.Y - value2.Y;
            float dz = value1.Z - value2.Z;
            float dw = value1.W - value2.W;
            result = dx * dx + dy * dy + dz * dz + dw * dw;
        }

        public static float Dot(Vector4 vector1, Vector4 vector2)
        {
            return vector1.X * vector2.X + vector1.Y * vector2.Y + vector1.Z * vector2.Z + vector1.W * vector2.W;
        }

        public static void Dot(ref Vector4 vector1, ref Vector4 vector2, out float result)
        {
            result = vector1.X * vector2.X + vector1.Y * vector2.Y + vector1.Z * vector2.Z + vector1.W * vector2.W;
        }

        public void Normalize()
        {
            float ls = X * X + Y * Y + Z * Z + W * W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);
            X *= invNorm;
            Y *= invNorm;
            Z *= invNorm;
            W *= invNorm;
        }

        public static Vector4 Normalize(Vector4 vector)
        {
            Vector4 ans;

            float ls = vector.X * vector.X + vector.Y * vector.Y + vector.Z * vector.Z + vector.W * vector.W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            ans.X = vector.X * invNorm;
            ans.Y = vector.Y * invNorm;
            ans.Z = vector.Z * invNorm;
            ans.W = vector.W * invNorm;

            return ans;
        }

        public static void Normalize(ref Vector4 vector, out Vector4 result)
        {
            float ls = vector.X * vector.X + vector.Y * vector.Y + vector.Z * vector.Z + vector.W * vector.W;
            float invNorm = 1.0f / (float)System.Math.Sqrt((double)ls);

            result.X = vector.X * invNorm;
            result.Y = vector.Y * invNorm;
            result.Z = vector.Z * invNorm;
            result.W = vector.W * invNorm;
        }

        // Binary Operators
        public static Vector4 Min(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = (value1.X < value2.X) ? value1.X : value2.X;
            ans.Y = (value1.Y < value2.Y) ? value1.Y : value2.Y;
            ans.Z = (value1.Z < value2.Z) ? value1.Z : value2.Z;
            ans.W = (value1.W < value2.W) ? value1.W : value2.W;

            return ans;
        }

        public static void Min(ref Vector4 value1, ref Vector4 value2, out Vector4 result)
        {
            result.X = (value1.X < value2.X) ? value1.X : value2.X;
            result.Y = (value1.Y < value2.Y) ? value1.Y : value2.Y;
            result.Z = (value1.Z < value2.Z) ? value1.Z : value2.Z;
            result.W = (value1.W < value2.W) ? value1.W : value2.W;
        }


        public static Vector4 Max(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = (value1.X > value2.X) ? value1.X : value2.X;
            ans.Y = (value1.Y > value2.Y) ? value1.Y : value2.Y;
            ans.Z = (value1.Z > value2.Z) ? value1.Z : value2.Z;
            ans.W = (value1.W > value2.W) ? value1.W : value2.W;

            return ans;
        }

        public static void  Max(ref Vector4 value1, ref Vector4 value2, out Vector4 result)
        {
            result.X = (value1.X > value2.X) ? value1.X : value2.X;
            result.Y = (value1.Y > value2.Y) ? value1.Y : value2.Y;
            result.Z = (value1.Z > value2.Z) ? value1.Z : value2.Z;
            result.W = (value1.W > value2.W) ? value1.W : value2.W;
        }

        public static Vector4 Clamp(Vector4 value1, Vector4 min, Vector4 max)
        {
            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.
            Vector4 result;

            float x = value1.X;
            x = (x> max.X) ? max.X : x;
            x = (x< min.X) ? min.X : x;

            float y = value1.Y;
            y = (y > max.Y) ? max.Y : y;
            y = (y < min.Y) ? min.Y : y;

            float z = value1.Z;
            z = (z > max.Z) ? max.Z : z;
            z = (z < min.Z) ? min.Z : z;

            float w = value1.W;
            w = (w > max.W) ? max.W : w;
            w = (w < min.W) ? min.W : w;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;

            return result;
        }

        public static void Clamp(ref Vector4 value1, ref Vector4 min, ref Vector4 max, out Vector4 result)
        {
            // This compare order is very important!!!
            // We must follow HLSL behavior in the case user specfied min value is bigger than max value.

            float x = value1.X;
            x = (x > max.X) ? max.X : x;
            x = (x < min.X) ? min.X : x;

            float y = value1.Y;
            y = (y > max.Y) ? max.Y : y;
            y = (y < min.Y) ? min.Y : y;

            float z = value1.Z;
            z = (z > max.Z) ? max.Z : z;
            z = (z < min.Z) ? min.Z : z;

            float w = value1.W;
            w = (w > max.W) ? max.W : w;
            w = (w < min.W) ? min.W : w;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;
        }


        public static Vector4 Lerp(Vector4 value1, Vector4 value2, float amount)
        {
            Vector4 ans;

            ans.X = value1.X + (value2.X - value1.X) * amount;
            ans.Y = value1.Y + (value2.Y - value1.Y) * amount;
            ans.Z = value1.Z + (value2.Z - value1.Z) * amount;
            ans.W = value1.W + (value2.W - value1.W) * amount;

            return ans;
        }

        public static void Lerp(ref Vector4 value1, ref Vector4 value2, float amount, out Vector4 result)
        {
            result.X = value1.X + (value2.X - value1.X) * amount;
            result.Y = value1.Y + (value2.Y - value1.Y) * amount;
            result.Z = value1.Z + (value2.Z - value1.Z) * amount;
            result.W = value1.W + (value2.W - value1.W) * amount;
        }


        /// <summary>
        /// Returns a point in Barycentric coordinates, using the specified values.
        /// value1 + amount1( value2 - value1 ) + amount2( value3 - value1 )
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="amount1"></param>
        /// <param name="amount2"></param>
        /// <returns></returns>
        public static Vector4 Barycentric(Vector4 value1, Vector4 value2, Vector4 value3, float amount1, float amount2)
        {
            Vector4 ans;

            ans.X = value1.X + amount1 * (value2.X - value1.X) + amount2 * (value3.X - value1.X);
            ans.Y = value1.Y + amount1 * (value2.Y - value1.Y) + amount2 * (value3.Y - value1.Y);
            ans.Z = value1.Z + amount1 * (value2.Z - value1.Z) + amount2 * (value3.Z - value1.Z);
            ans.W = value1.W + amount1 * (value2.W - value1.W) + amount2 * (value3.W - value1.W);

            return ans;
        }

        public static void Barycentric(ref Vector4 value1, ref Vector4 value2, ref Vector4 value3, float amount1, float amount2, out Vector4 result)
        {
            result.X = value1.X + amount1 * (value2.X - value1.X) + amount2 * (value3.X - value1.X);
            result.Y = value1.Y + amount1 * (value2.Y - value1.Y) + amount2 * (value3.Y - value1.Y);
            result.Z = value1.Z + amount1 * (value2.Z - value1.Z) + amount2 * (value3.Z - value1.Z);
            result.W = value1.W + amount1 * (value2.W - value1.W) + amount2 * (value3.W - value1.W);
        }


        public static Vector4 SmoothStep(Vector4 value1, Vector4 value2, float amount)
        {
            Vector4 ans;

            amount = (amount > 1.0f) ? 1.0f : ((amount < 0.0f) ? 0.0f : amount);    // Clamp amount.

            amount = amount * amount * (3.0f - 2.0f * amount);
            ans.X = value1.X + (value2.X - value1.X) * amount;
            ans.Y = value1.Y + (value2.Y - value1.Y) * amount;
            ans.Z = value1.Z + (value2.Z - value1.Z) * amount;
            ans.W = value1.W + (value2.W - value1.W) * amount;

            return ans;
        }

        public static void SmoothStep(ref Vector4 value1, ref Vector4 value2, float amount, out Vector4 result)
        {
            amount = (amount > 1.0f) ? 1.0f : ((amount < 0.0f) ? 0.0f : amount);    // Clamp amount.

            amount = amount * amount * (3.0f - 2.0f * amount);
            result.X = value1.X + (value2.X - value1.X) * amount;
            result.Y = value1.Y + (value2.Y - value1.Y) * amount;
            result.Z = value1.Z + (value2.Z - value1.Z) * amount;
            result.W = value1.W + (value2.W - value1.W) * amount;
        }


        /// <summary>
        /// Perfroms a Catmull-Rom interpolation, using the specfied 4D vectors.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="value2"></param>
        /// <param name="value3"></param>
        /// <param name="value4"></param>
        /// <param name="amount">Weighting factor.</param>
        /// <returns></returns>
        public static Vector4 CatmullRom(Vector4 value1, Vector4 value2, Vector4 value3, Vector4 value4, float amount)
        {
            Vector4 ans;

            float t2 = amount * amount;
            float t3 = amount * t2;

            ans.X = 0.5f * (
                        (2.0f * value2.X) +
                        (-value1.X + value3.X) * amount +
                        (2.0f * value1.X - 5.0f * value2.X + 4.0f * value3.X - value4.X) * t2 +
                        (-value1.X + 3.0f * value2.X - 3.0f * value3.X + value4.X) * t3);

            ans.Y = 0.5f * (
                        (2.0f * value2.Y) +
                        (-value1.Y + value3.Y) * amount +
                        (2.0f * value1.Y - 5.0f * value2.Y + 4.0f * value3.Y - value4.Y) * t2 +
                        (-value1.Y + 3.0f * value2.Y - 3.0f * value3.Y + value4.Y) * t3);

            ans.Z = 0.5f * (
                        (2.0f * value2.Z) +
                        (-value1.Z + value3.Z) * amount +
                        (2.0f * value1.Z - 5.0f * value2.Z + 4.0f * value3.Z - value4.Z) * t2 +
                        (-value1.Z + 3.0f * value2.Z - 3.0f * value3.Z + value4.Z) * t3);

            ans.W = 0.5f * (
                        (2.0f * value2.W) +
                        (-value1.W + value3.W) * amount +
                        (2.0f * value1.W - 5.0f * value2.W + 4.0f * value3.W - value4.W) * t2 +
                        (-value1.W + 3.0f * value2.W - 3.0f * value3.W + value4.W) * t3);

            return ans;
        }

        public static void CatmullRom(ref Vector4 value1, ref Vector4 value2, ref Vector4 value3, ref Vector4 value4, float amount, out Vector4 result)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;

            result.X = 0.5f * (
                        (2.0f * value2.X) +
                        (-value1.X + value3.X) * amount +
                        (2.0f * value1.X - 5.0f * value2.X + 4.0f * value3.X - value4.X) * t2 +
                        (-value1.X + 3.0f * value2.X - 3.0f * value3.X + value4.X) * t3);

            result.Y = 0.5f * (
                        (2.0f * value2.Y) +
                        (-value1.Y + value3.Y) * amount +
                        (2.0f * value1.Y - 5.0f * value2.Y + 4.0f * value3.Y - value4.Y) * t2 +
                        (-value1.Y + 3.0f * value2.Y - 3.0f * value3.Y + value4.Y) * t3);

            result.Z = 0.5f * (
                        (2.0f * value2.Z) +
                        (-value1.Z + value3.Z) * amount +
                        (2.0f * value1.Z - 5.0f * value2.Z + 4.0f * value3.Z - value4.Z) * t2 +
                        (-value1.Z + 3.0f * value2.Z - 3.0f * value3.Z + value4.Z) * t3);

            result.W = 0.5f * (
                        (2.0f * value2.W) +
                        (-value1.W + value3.W) * amount +
                        (2.0f * value1.W - 5.0f * value2.W + 4.0f * value3.W - value4.W) * t2 +
                        (-value1.W + 3.0f * value2.W - 3.0f * value3.W + value4.W) * t3);
        }


        /// <summary>
        /// Performs a Hermite spline interpolation, using the specfied 4D values.
        /// </summary>
        /// <param name="value1"></param>
        /// <param name="tangent1"></param>
        /// <param name="value2"></param>
        /// <param name="tangent2"></param>
        /// <param name="amount">Weighting factor</param>
        /// <returns></returns>
        public static Vector4 Hermite(Vector4 value1, Vector4 tangent1, Vector4 value2, Vector4 tangent2, float amount)
        {
            Vector4 ans;

            float t2 = amount * amount;
            float t3 = amount * t2;
            float h1 = 2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 = t3 - 2.0f * t2 + amount;
            float h4 = t3 - t2;

            ans.X = value1.X * h1 + value2.X * h2 + tangent1.X * h3 + tangent2.X * h4;
            ans.Y = value1.Y * h1 + value2.Y * h2 + tangent1.Y * h3 + tangent2.Y * h4;
            ans.Z = value1.Z * h1 + value2.Z * h2 + tangent1.Z * h3 + tangent2.Z * h4;
            ans.W = value1.W * h1 + value2.W * h2 + tangent1.W * h3 + tangent2.W * h4;

            return ans;
        }

        public static void Hermite(ref Vector4 value1, ref Vector4 tangent1, ref Vector4 value2, ref Vector4 tangent2, float amount, out Vector4 result)
        {
            float t2 = amount * amount;
            float t3 = amount * t2;
            float h1 = 2.0f * t3 - 3.0f * t2 + 1.0f;
            float h2 = -2.0f * t3 + 3.0f * t2;
            float h3 = t3 - 2.0f * t2 + amount;
            float h4 = t3 - t2;

            result.X = value1.X * h1 + value2.X * h2 + tangent1.X * h3 + tangent2.X * h4;
            result.Y = value1.Y * h1 + value2.Y * h2 + tangent1.Y * h3 + tangent2.Y * h4;
            result.Z = value1.Z * h1 + value2.Z * h2 + tangent1.Z * h3 + tangent2.Z * h4;
            result.W = value1.W * h1 + value2.W * h2 + tangent1.W * h3 + tangent2.W * h4;
        }

        #region Transform by Matrix
        /// <summary>
        /// Transform vector (x, y, 0, 1) by given matrix.
        /// </summary>
        /// <param name="position"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Vector4 Transform(Vector2 position, Matrix matrix)
        {
            Vector4 result;

            float x = position.X * matrix.M11 + position.Y * matrix.M21 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + matrix.M42;
            float z = position.X * matrix.M13 + position.Y * matrix.M23 + matrix.M43;
            float w = position.X * matrix.M14 + position.Y * matrix.M24 + matrix.M44;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;

            return result;
        }

        public static void Transform(ref Vector2 position, ref Matrix matrix, out Vector4 result)
        {
            float x = position.X * matrix.M11 + position.Y * matrix.M21 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + matrix.M42;
            float z = position.X * matrix.M13 + position.Y * matrix.M23 + matrix.M43;
            float w = position.X * matrix.M14 + position.Y * matrix.M24 + matrix.M44;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;
        }


        /// <summary>
        /// Transform vector (x, y, z, 1) by given matrix.
        /// </summary>
        /// <param name="position"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Vector4 Transform(Vector3 position, Matrix matrix)
        {
            Vector4 result;

            float x = position.X * matrix.M11 + position.Y * matrix.M21 + position.Z * matrix.M31 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + position.Z * matrix.M32 + matrix.M42;
            float z = position.X * matrix.M13 + position.Y * matrix.M23 + position.Z * matrix.M33 + matrix.M43;
            float w = position.X * matrix.M14 + position.Y * matrix.M24 + position.Z * matrix.M34 + matrix.M44;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;

            return result;
        }

        public static void Transform(ref Vector3 position, ref Matrix matrix, out Vector4 result)
        {
            float x = position.X * matrix.M11 + position.Y * matrix.M21 + position.Z * matrix.M31 + matrix.M41;
            float y = position.X * matrix.M12 + position.Y * matrix.M22 + position.Z * matrix.M32 + matrix.M42;
            float z = position.X * matrix.M13 + position.Y * matrix.M23 + position.Z * matrix.M33 + matrix.M43;
            float w = position.X * matrix.M14 + position.Y * matrix.M24 + position.Z * matrix.M34 + matrix.M44;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;
        }

        /// <summary>
        /// Transform vector (x, y, z, w) by given matrix.
        /// </summary>
        /// <param name="vector"></param>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Vector4 Transform(Vector4 vector, Matrix matrix)
        {
            Vector4 result;

            float x = vector.X * matrix.M11 + vector.Y * matrix.M21 + vector.Z * matrix.M31 + vector.W * matrix.M41;
            float y = vector.X * matrix.M12 + vector.Y * matrix.M22 + vector.Z * matrix.M32 + vector.W * matrix.M42;
            float z = vector.X * matrix.M13 + vector.Y * matrix.M23 + vector.Z * matrix.M33 + vector.W * matrix.M43;
            float w = vector.X * matrix.M14 + vector.Y * matrix.M24 + vector.Z * matrix.M34 + vector.W * matrix.M44;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;

            return result;
        }

        public static void Transform(ref Vector4 vector, ref Matrix matrix, out Vector4 result)
        {
            float x = vector.X * matrix.M11 + vector.Y * matrix.M21 + vector.Z * matrix.M31 + vector.W * matrix.M41;
            float y = vector.X * matrix.M12 + vector.Y * matrix.M22 + vector.Z * matrix.M32 + vector.W * matrix.M42;
            float z = vector.X * matrix.M13 + vector.Y * matrix.M23 + vector.Z * matrix.M33 + vector.W * matrix.M43;
            float w = vector.X * matrix.M14 + vector.Y * matrix.M24 + vector.Z * matrix.M34 + vector.W * matrix.M44;

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = w;
        }
        #endregion

        #region Transform by Quaternion
        /// <summary>
        /// Transform vector (x, y, 0, 1) by given quaternion.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        public static Vector4 Transform(Vector2 value, Quaternion rotation)
        {
            Vector4 result;

            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * ( 1.0f - yy2 - zz2 ) + value.Y * (        xy2 - wz2 );
            float y = value.X * (        xy2 + wz2 ) + value.Y * ( 1.0f - xx2 - zz2 );
            float z = value.X * (        xz2 - wy2 ) + value.Y * (        yz2 + wx2 );

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = 1.0f;

            return result;
        }

        public static void Transform(ref Vector2 value, ref Quaternion rotation, out Vector4 result)
        {
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * ( 1.0f - yy2 - zz2 ) + value.Y * (        xy2 - wz2 );
            float y = value.X * (        xy2 + wz2 ) + value.Y * ( 1.0f - xx2 - zz2 );
            float z = value.X * (        xz2 - wy2 ) + value.Y * (        yz2 + wx2 );

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = 1.0f;
        }


        /// <summary>
        /// Transform vector (x, y, z, 1) by given quaternion.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        public static Vector4 Transform(Vector3 value, Quaternion rotation)
        {
            Vector4 result;

            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * ( 1.0f - yy2 - zz2 ) + value.Y * (        xy2 - wz2 ) + value.Z * (        xz2 + wy2 );
            float y = value.X * (        xy2 + wz2 ) + value.Y * ( 1.0f - xx2 - zz2 ) + value.Z * (        yz2 - wx2 );
            float z = value.X * (        xz2 - wy2 ) + value.Y * (        yz2 + wx2 ) + value.Z * ( 1.0f - xx2 - yy2 );

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = 1.0f;

            return result;
        }

        public static void Transform(ref Vector3 value, ref Quaternion rotation, out Vector4 result)
        {
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * ( 1.0f - yy2 - zz2 ) + value.Y * (        xy2 - wz2 ) + value.Z * (        xz2 + wy2 );
            float y = value.X * (        xy2 + wz2 ) + value.Y * ( 1.0f - xx2 - zz2 ) + value.Z * (        yz2 - wx2 );
            float z = value.X * (        xz2 - wy2 ) + value.Y * (        yz2 + wx2 ) + value.Z * ( 1.0f - xx2 - yy2 );

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = 1.0f;
        }

        /// <summary>
        /// Transform vector (x, y, z, w) by given quaternion.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        public static Vector4 Transform(Vector4 value, Quaternion rotation)
        {
            Vector4 result;

            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * ( 1.0f - yy2 - zz2 ) + value.Y * (        xy2 - wz2 ) + value.Z * (        xz2 + wy2 );
            float y = value.X * (        xy2 + wz2 ) + value.Y * ( 1.0f - xx2 - zz2 ) + value.Z * (        yz2 - wx2 );
            float z = value.X * (        xz2 - wy2 ) + value.Y * (        yz2 + wx2 ) + value.Z * ( 1.0f - xx2 - yy2 );

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = value.W;

            return result;
        }

        public static void Transform(ref Vector4 value, ref Quaternion rotation, out Vector4 result)
        {
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float x = value.X * ( 1.0f - yy2 - zz2 ) + value.Y * (        xy2 - wz2 ) + value.Z * (        xz2 + wy2 );
            float y = value.X * (        xy2 + wz2 ) + value.Y * ( 1.0f - xx2 - zz2 ) + value.Z * (        yz2 - wx2 );
            float z = value.X * (        xz2 - wy2 ) + value.Y * (        yz2 + wx2 ) + value.Z * ( 1.0f - xx2 - yy2 );

            result.X = x;
            result.Y = y;
            result.Z = z;
            result.W = value.W;
        }
        #endregion

        #region Trasform by Array
        public static void Transform(Vector4[] sourceArray, ref Matrix matrix, Vector4[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y, sz = sourceArray[idx].Z, sw = sourceArray[idx].W;

                destinationArray[idx].X = sx * matrix.M11 + sy * matrix.M21 + sz * matrix.M31 + sw * matrix.M41;
                destinationArray[idx].Y = sx * matrix.M12 + sy * matrix.M22 + sz * matrix.M32 + sw * matrix.M42;
                destinationArray[idx].Z = sx * matrix.M13 + sy * matrix.M23 + sz * matrix.M33 + sw * matrix.M43;
                destinationArray[idx].W = sx * matrix.M14 + sy * matrix.M24 + sz * matrix.M34 + sw * matrix.M44;
            }
        }

        [SuppressMessage("Microsoft.Usage","CA2233")]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void Transform(Vector4[] sourceArray, int sourceIndex, ref Matrix matrix, Vector4[] destinationArray, int destinationIndex, int length )
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");


            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y, sz = sourceArray[sourceIndex].Z, sw = sourceArray[sourceIndex].W;

                destinationArray[destinationIndex].X = sx * matrix.M11 + sy * matrix.M21 + sz * matrix.M31 + sw * matrix.M41;
                destinationArray[destinationIndex].Y = sx * matrix.M12 + sy * matrix.M22 + sz * matrix.M32 + sw * matrix.M42;
                destinationArray[destinationIndex].Z = sx * matrix.M13 + sy * matrix.M23 + sz * matrix.M33 + sw * matrix.M43;
                destinationArray[destinationIndex].W = sx * matrix.M14 + sy * matrix.M24 + sz * matrix.M34 + sw * matrix.M44;

                sourceIndex++; destinationIndex++;
                length--;
            }
        }

        public static void Transform(Vector4[] sourceArray, ref Quaternion rotation, Vector4[] destinationArray)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (destinationArray.Length < sourceArray.Length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float m11 = 1.0f - yy2 - zz2;
            float m21 = xy2 - wz2;
            float m31 = xz2 + wy2;

            float m12 = xy2 + wz2;
            float m22 = 1.0f - xx2 - zz2;
            float m32 = yz2 - wx2;

            float m13 = xz2 - wy2;
            float m23 = yz2 + wx2;
            float m33 = 1.0f - xx2 - yy2;

            for (int idx = 0; idx < sourceArray.Length; ++idx)
            {
                float sx = sourceArray[idx].X, sy = sourceArray[idx].Y, sz = sourceArray[idx].Z;

                destinationArray[idx].X = sx * m11 + sy * m21 + sz * m31;
                destinationArray[idx].Y = sx * m12 + sy * m22 + sz * m32;
                destinationArray[idx].Z = sx * m13 + sy * m23 + sz * m33;
                destinationArray[idx].W = sourceArray[idx].W;
            }
        }

        [SuppressMessage( "Microsoft.Usage", "CA2233" )]// suppressing warning about potential overflow in sourceIndex and destinationIndex. we check for this.
        public static void Transform(Vector4[] sourceArray, int sourceIndex, ref Quaternion rotation, Vector4[] destinationArray, int destinationIndex, int length)
        {
            if (sourceArray == null)
                throw new ArgumentNullException("sourceArray");

            if (destinationArray == null)
                throw new ArgumentNullException("destinationArray");

            if (sourceArray.Length < (long)sourceIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughSourceSize");

            if (destinationArray.Length < (long)destinationIndex + (long)length)
                throw new ArgumentException("FrameworkResources.NotEnoughTargetSize");

            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float m11 = 1.0f - yy2 - zz2;
            float m21 = xy2 - wz2;
            float m31 = xz2 + wy2;

            float m12 = xy2 + wz2;
            float m22 = 1.0f - xx2 - zz2;
            float m32 = yz2 - wx2;

            float m13 = xz2 - wy2;
            float m23 = yz2 + wx2;
            float m33 = 1.0f - xx2 - yy2;

            while (length > 0)
            {
                float sx = sourceArray[sourceIndex].X, sy = sourceArray[sourceIndex].Y, sz = sourceArray[sourceIndex].Z, sw = sourceArray[sourceIndex].W;

                destinationArray[destinationIndex].X = sx * m11 + sy * m21 + sz * m31;
                destinationArray[destinationIndex].Y = sx * m12 + sy * m22 + sz * m32;
                destinationArray[destinationIndex].Z = sx * m13 + sy * m23 + sz * m33;
                destinationArray[destinationIndex].W = sw;

                sourceIndex++; destinationIndex++;
                length--;
            }

        }
        #endregion

        #endregion

        #region Operator overloads equivalent methods
        public static Vector4 Negate(Vector4 value)
        {
            Vector4 ans;

            ans.X = -value.X;
            ans.Y = -value.Y;
            ans.Z = -value.Z;
            ans.W = -value.W;

            return ans;
        }

        public static void Negate(ref Vector4 value, out Vector4 result)
        {
            result.X = -value.X;
            result.Y = -value.Y;
            result.Z = -value.Z;
            result.W = -value.W;
        }


        public static Vector4 Add(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X + value2.X;
            ans.Y = value1.Y + value2.Y;
            ans.Z = value1.Z + value2.Z;
            ans.W = value1.W + value2.W;

            return ans;
        }

        public static void Add(ref Vector4 value1, ref Vector4 value2, out Vector4 result)
        {
            result.X = value1.X + value2.X;
            result.Y = value1.Y + value2.Y;
            result.Z = value1.Z + value2.Z;
            result.W = value1.W + value2.W;
        }


        public static Vector4 Subtract(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X - value2.X;
            ans.Y = value1.Y - value2.Y;
            ans.Z = value1.Z - value2.Z;
            ans.W = value1.W - value2.W;

            return ans;
        }

        public static void Subtract(ref Vector4 value1, ref Vector4 value2, out Vector4 result)
        {
            result.X = value1.X - value2.X;
            result.Y = value1.Y - value2.Y;
            result.Z = value1.Z - value2.Z;
            result.W = value1.W - value2.W;
        }


        public static Vector4 Multiply(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X * value2.X;
            ans.Y = value1.Y * value2.Y;
            ans.Z = value1.Z * value2.Z;
            ans.W = value1.W * value2.W;

            return ans;
        }

        public static void Multiply(ref Vector4 value1, ref Vector4 value2, out Vector4 result)
        {
            result.X = value1.X * value2.X;
            result.Y = value1.Y * value2.Y;
            result.Z = value1.Z * value2.Z;
            result.W = value1.W * value2.W;
        }


        public static Vector4 Multiply(Vector4 value1, float scaleFactor)
        {
            Vector4 ans;

            ans.X = value1.X * scaleFactor;
            ans.Y = value1.Y * scaleFactor;
            ans.Z = value1.Z * scaleFactor;
            ans.W = value1.W * scaleFactor;

            return ans;
        }

        public static void Multiply(ref Vector4 value1, float scaleFactor, out Vector4 result)
        {
            result.X = value1.X * scaleFactor;
            result.Y = value1.Y * scaleFactor;
            result.Z = value1.Z * scaleFactor;
            result.W = value1.W * scaleFactor;
        }


        public static Vector4 Divide(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X / value2.X;
            ans.Y = value1.Y / value2.Y;
            ans.Z = value1.Z / value2.Z;
            ans.W = value1.W / value2.W;

            return ans;
        }

        public static void Divide(ref Vector4 value1, ref Vector4 value2, out Vector4 result)
        {
            result.X = value1.X / value2.X;
            result.Y = value1.Y / value2.Y;
            result.Z = value1.Z / value2.Z;
            result.W = value1.W / value2.W;
        }


        public static Vector4 Divide(Vector4 value1, float divider)
        {
            Vector4 ans;

            float invDiv = 1.0f / divider;
            ans.X = value1.X * invDiv;
            ans.Y = value1.Y * invDiv;
            ans.Z = value1.Z * invDiv;
            ans.W = value1.W * invDiv;

            return ans;
        }

        public static void Divide(ref Vector4 value1, float divider, out Vector4 result)
        {
            float invDiv = 1.0f / divider;
            result.X = value1.X * invDiv;
            result.Y = value1.Y * invDiv;
            result.Z = value1.Z * invDiv;
            result.W = value1.W * invDiv;
        }
        #endregion


        #region Operator overloads
        // Unary operator overloads
        public static Vector4 operator -(Vector4 value)
        {
            Vector4 ans;

            ans.X = -value.X;
            ans.Y = -value.Y;
            ans.Z = -value.Z;
            ans.W = -value.W;

            return ans;
        }

        // Binary operator overloads
        public static bool operator ==(Vector4 value1, Vector4 value2)
        {
            return (value1.X == value2.X && value1.Y == value2.Y && value1.Z == value2.Z && value1.W == value2.W);
        }

        public static bool operator !=(Vector4 value1, Vector4 value2)
        {
            return (value1.X != value2.X || value1.Y != value2.Y || value1.Z != value2.Z || value1.W != value2.W);
        }

        public static Vector4 operator +(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X + value2.X;
            ans.Y = value1.Y + value2.Y;
            ans.Z = value1.Z + value2.Z;
            ans.W = value1.W + value2.W;

            return ans;
        }

        public static Vector4 operator -(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X - value2.X;
            ans.Y = value1.Y - value2.Y;
            ans.Z = value1.Z - value2.Z;
            ans.W = value1.W - value2.W;

            return ans;
        }

        public static Vector4 operator *(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X * value2.X;
            ans.Y = value1.Y * value2.Y;
            ans.Z = value1.Z * value2.Z;
            ans.W = value1.W * value2.W;

            return ans;
        }

        public static Vector4 operator *(Vector4 value1, float scaleFactor)
        {
            Vector4 ans;

            ans.X = value1.X * scaleFactor;
            ans.Y = value1.Y * scaleFactor;
            ans.Z = value1.Z * scaleFactor;
            ans.W = value1.W * scaleFactor;

            return ans;
        }

        public static Vector4 operator *(float scaleFactor, Vector4 value1)
        {
            Vector4 ans;

            ans.X = value1.X * scaleFactor;
            ans.Y = value1.Y * scaleFactor;
            ans.Z = value1.Z * scaleFactor;
            ans.W = value1.W * scaleFactor;

            return ans;
        }

        public static Vector4 operator /(Vector4 value1, Vector4 value2)
        {
            Vector4 ans;

            ans.X = value1.X / value2.X;
            ans.Y = value1.Y / value2.Y;
            ans.Z = value1.Z / value2.Z;
            ans.W = value1.W / value2.W;

            return ans;
        }

        public static Vector4 operator /(Vector4 value1, float divider)
        {
            Vector4 ans;

            float invDiv = 1.0f / divider;
            ans.X = value1.X * invDiv;
            ans.Y = value1.Y * invDiv;
            ans.Z = value1.Z * invDiv;
            ans.W = value1.W * invDiv;

            return ans;
        }
        #endregion

    }
}
